(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module H = Hashtbl
module L = List

module T = Type

open Vcg

module ESet = Set.Make (
  struct
    type t = Vcg.expr
    let compare t1 t2 = compare t1 t2
  end)

module RSet = Set.Make (
  struct
    type t = Vcg.expr * Vcg.expr
    let compare t1 t2 = compare t1 t2
  end)

(* --- array properties --- *)

exception Done of bool

let sub emap expr =
  let rec app1 expr =
    try L.assoc expr emap with
    | Not_found -> app2 expr
  and app2 = function
    | Array (upd, arr) -> 
	Array (L.map (fun (e1, e2) -> (app1 e1, app1 e2)) upd, app1 arr)
    | Length e -> Length (app1 e)
    | ArrayAcc (e1, e2) -> ArrayAcc (app1 e1, app1 e2)
    | Struct (upd, st) ->
	Struct (L.map (fun (fld, e) -> (fld, app1 e)) upd, app1 st)
    | StructAcc (e, fld) -> StructAcc (app1 e, fld)
    | FnCall (fn, args) -> FnCall (fn, L.map app1 args)
    | Plus (e1, e2) -> Plus (app1 e1, app1 e2)
    | Minus (e1, e2) -> Minus (app1 e1, app1 e2)
    | Times (e1, e2) -> Times (app1 e1, app1 e2)
    | Div (e1, e2) -> Div (app1 e1, app1 e2)
    | IDiv (e1, e2) -> IDiv (app1 e1, app1 e2)
    | Mod (e1, e2) -> Mod (app1 e1, app1 e2)
    | UMinus e -> UMinus (app1 e)
    | LT (e1, e2) -> LT (app1 e1, app1 e2)
    | LE (e1, e2) -> LE (app1 e1, app1 e2)
    | NE (e1, e2) -> NE (app1 e1, app1 e2)
    | EQ (e1, e2) -> 
	let e1' = app1 e1 in
	let e2' = app1 e2 in
	if e1' = e2' then Bool true else EQ (e1', e2')
    | GE (e1, e2) -> GE (app1 e1, app1 e2)
    | GT (e1, e2) -> GT (app1 e1, app1 e2)
    | BExpr e -> BExpr (app1 e)
    | And (e1, e2) -> 
	let e1' = app1 e1 in
	let e2' = app1 e2 in
	if e1' = Bool true then e2'
	else if e2' = Bool true then e1'
	else And (e1', e2')
    | Or (e1, e2) -> Or (app1 e1, app1 e2)
    | Not e -> Not (app1 e)
    | x -> x
  in
  app1 expr

let rewrite expr =
  let rltn = ref RSet.empty in
  let relate x y =
    let rec propagate () =
      let n = ref RSet.empty in
      RSet.iter (fun (w, x) ->
	RSet.iter (fun (y, z) ->
	  if x = y then n := RSet.add (w, z) !n)
	  !rltn)
	!rltn;
      let rltn' = RSet.union !rltn !n in
      if not (RSet.equal rltn' !rltn) then begin
	rltn := rltn';
	propagate ()
      end
    in
    rltn := RSet.add (x, y) !rltn;
    propagate ()
  in
  let related x y = RSet.mem (x, y) !rltn in

  let first1 = ref ESet.empty in
  let first2 = ref RSet.empty in

  let aprop = ref false in

  let nvars = ref [] in
  let id = let id = ref 0 in fun () -> incr id; string_of_int !id in
  let nvar bnm =
    let v = bnm ^ "_" ^ (id ()) in
    nvars := v :: !nvars;
    v
  in
  let rewrite iset good phase expr =
    let appor = function
      | [] -> Bool false
      | [h] -> h
      | h :: t -> L.fold_left (fun e1 e2 -> Or (e1, e2)) h t
    in
    let rec rewrite phase neg = function
      | FnCall ("exists", [ix; l; u; phi]) when phase = 1 && neg ->
	  aprop := true;
	  let i = Var (nvar "ix") in
	  first1 := ESet.add ix !first1;
	  Or (Not (And (LE (l, i), LE (i, u))),
	      (sub [(ix, i)] (rewrite phase neg phi)))
      | FnCall ("forall", [ix; l; u; phi]) when phase = 1 && not neg ->
	  aprop := true;
	  let i = Var (nvar "ix") in
	  first1 := ESet.add ix !first1;
	  Or (Not (And (LE (l, i), LE (i, u))),
	      sub [(ix, i)] (rewrite phase neg phi))
      | FnCall ("sorted", [arr; l; u]) when phase = 1 && not neg ->
          aprop := true;
	  let i = Var (nvar "i") in
	  let j = Var (nvar "j") in
	  first2 := RSet.add (i, j) !first2;
	  Or (Not (And (LE (l, i), And (LE (i, j), LE (j, u)))),
	      LE (ArrayAcc (arr, i), ArrayAcc (arr, j)))
      | FnCall ("partitioned", [arr; l1; u1; l2; u2])
	when phase = 1 && not neg ->
	  aprop := true;
	  let i = Var (nvar "i") in
	  let j = Var (nvar "j") in
	  first2 := RSet.add (i, j) !first2;
	  Or (Not (And (And (And (And (LE (l1, i), LE (i, u1)),
				  LT (u1, l2)), 
			     LE (l2, j)), 
			LE (j, u2))),
	      LE (ArrayAcc (arr, i), ArrayAcc (arr, j)))
      | FnCall ("eq", [a1; a2; l; u]) when phase = 1 && not neg ->
	  aprop := true;
	  let i = Var (nvar "i") in
	  first1 := ESet.add i !first1;
	  Or (Not (And (LE (l, i), LE (i, u))),
	      EQ (ArrayAcc (a1, i), ArrayAcc (a2, i)))
      (* sound but NOT COMPLETE *)
      | FnCall ("iEqualOffset", [a1; b1; a2; b2; len]) 
	when phase = 1 && not neg ->
	  aprop := true;
	  let i = Var (nvar "i") in
	  first1 := ESet.add i !first1;
	  Or (Not (And (LE (Int 0, i), LT (i, len))),
	      EQ (ArrayAcc (a1, Plus (b1, i)), ArrayAcc (a2, (Plus (b2, i)))))

      | FnCall ("exists", [ix; l; u; phi]) when phase = 2 && not neg -> begin
	  let disj = ref [] in
	  ESet.iter (fun i ->
	    if good [i] then
	    disj := 
	      (And (And (LE (l, i), LE (i, u)),
		    sub [(ix, i)] (rewrite phase neg phi)))
	      :: !disj)
	    iset;
	  appor !disj	  
      end
      | FnCall ("forall", [ix; l; u; phi]) when phase = 2 && neg -> begin
	  let disj = ref [] in
	  ESet.iter (fun i ->
	    if good [i] then
	    disj := 
	      (And (And (LE (l, i), LE (i, u)),
		    sub [(ix, i)] (rewrite phase neg phi)))
	      :: !disj)
	    iset;
	  appor !disj
      end
      | FnCall ("sorted", [arr; l; u]) when phase = 2 && neg -> begin
	  let disj = ref [] in
	  ESet.iter (fun i ->
	    ESet.iter (fun j ->
	      if good [i; j] then
	      disj := 
		(And (And (LE (l, i), And (LE (i, j), LE (j, u))),
		      GT (ArrayAcc (arr, i), ArrayAcc (arr, j))))
		  :: !disj)
	      iset)
	    iset;
	  appor !disj
      end
      | FnCall ("partitioned", [arr; l1; u1; l2; u2])
	when phase = 2 && neg ->
	  let disj = ref [] in
	  ESet.iter (fun i ->
	    ESet.iter (fun j ->
	      if good [i; j] then
	      disj := 
		And (And (And (And (And (LE (l1, i), LE (i, u1)),
				    LT (u1, l2)), 
			       LE (l2, j)), 
			  LE (j, u2)),
		     GT (ArrayAcc (arr, i), ArrayAcc (arr, j)))
		:: !disj)
	      iset)
	    iset;
	  appor !disj
      | FnCall ("eq", [a1; a2; l; u]) when phase = 2 && neg -> begin
	  let disj = ref [] in
	  ESet.iter (fun i ->
	    if good [i] then
	    disj :=
	      (And (And (LE (l, i), LE (i, u)),
		    NE (ArrayAcc (a1, i), ArrayAcc (a2, i))))
	      :: !disj)
	    iset;
	  appor !disj
      end
      | FnCall ("iEqualOffset", [a1; b1; a2; b2; len])
	when phase = 2 && neg -> begin
	  let disj = ref [] in
	  ESet.iter (fun i ->
	    if good [i] then
	    disj :=
	      (And (And (LE (Int 0, i), LT (i, len)),
		    NE (ArrayAcc (a1, Plus (b1, i)), 
			ArrayAcc (a2, Plus (b2, i)))))
	      :: !disj)
	    iset;
	  appor !disj
      end

      (* record negative occurrences during phase 1 *)
      | FnCall ("exists", [ix; l; u; phi]) as x -> aprop := true; x
      | FnCall ("forall", [ix; l; u; phi]) as x -> aprop := true; Not x
      | FnCall ("sorted", _) as x -> aprop := true; Not x
      | FnCall ("partitioned", _) as x -> aprop := true; Not x
      | FnCall ("eq", _) as x -> aprop := true; Not x
      | FnCall ("iEqualOffset", _) as x -> aprop := true; Not x

      | BExpr b -> BExpr (rewrite phase neg b)
      | And (e1, e2) -> 
	  let e1, e2 = rewrite phase neg e1, rewrite phase neg e2 in
	  if neg then Or (e1, e2) else And (e1, e2)
      | Or (e1, e2) ->
	  let e1, e2 = rewrite phase neg e1, rewrite phase neg e2 in
	  if neg then And (e1, e2) else Or (e1, e2)
      | Not e -> rewrite phase (not neg) e
      | Forall (t, v, e) -> Forall (t, v, rewrite phase neg e)
      | x -> if neg then Not x else x
    in
    rewrite phase false expr
  in

  let expr = rewrite ESet.empty (fun x -> true) 1 expr in

  let isets =
    (* syntactic index sets (read/write) *)
    let rsset = ref ESet.empty in
    let wsset = ref ESet.empty in
    (* syntactic bounds set *)
    let bset = ref ESet.empty in
    (* syntactic offset set *)
    let oset = ref ESet.empty in

    let rec is = function
      | Array (upd, arr) ->
	  L.iter 
	    (fun (i, v) -> 
	      wsset := ESet.add i !wsset;
	      is i; is v)
	    upd;
	  is arr
      | Length e -> is e
      | ArrayAcc (arr, i) -> 
	  rsset := ESet.add i !rsset;
	  is arr;
	  is i
      | Struct (upd, st) ->
	  L.iter (fun (_, e) -> is e) upd; is st
      | StructAcc (st, _) -> is st
      | FnCall ("forall", [ix; l; u; phi]) ->
	  bset := ESet.add l (ESet.add u !bset);
	  is l; is u; is phi; 
          (* remove place-holder ix from rsset *)
	  rsset := ESet.remove ix !rsset
      | FnCall ("exists", [ix; l; u; phi]) ->
	  bset := ESet.add l (ESet.add u !bset);
	  is l; is u; is phi; 
          (* remove place-holder ix from rsset *)
	  rsset := ESet.remove ix !rsset
      | FnCall ("sorted", [arr; l; u]) ->
	  bset := ESet.add l (ESet.add u !bset);
	  is arr; is l; is u
      | FnCall ("partitioned", [arr; l1; u1; l2; u2]) ->
	  L.iter
	    (fun x -> bset := ESet.add x !bset)
	    [l1; u1; l2; u2];
	  is arr; is l1; is u1; is l2; is u2
      | FnCall ("eq", [a1; a2; l; u]) ->
	  bset := ESet.add l (ESet.add u !bset);
	  is a1; is a2; is l; is u
      | FnCall ("iEqualOffset", [a1; b1; a2; b2; len]) ->
	  oset := ESet.add b1 (ESet.add b2 !oset);
	  bset := ESet.add (Int 0) (ESet.add (Minus (len, Int 1)) !bset);
	  (* should descend into a1, a2; 
           * add writes as usual, reads offset by b1, b2 *)
	  is b1; is b2; is len
      | FnCall (_, args) -> L.iter is args
      | Plus (e1, e2)
      | Minus (e1, e2)
      | Times (e1, e2)
      | Div (e1, e2)
      | IDiv (e1, e2)
      | Mod (e1, e2)
      | LT (e1, e2)
      | LE (e1, e2)
      | NE (e1, e2)
      | EQ (e1, e2)
      | GE (e1, e2)
      | GT (e1, e2)
      | And (e1, e2)
      | Or (e1, e2) -> is e1; is e2
      | UMinus e
      | BExpr e
      | Not e -> is e
      | Forall (_, _, e) -> is e
      | _ -> ()
    in
    is expr;
    let shift = ref ESet.empty in
    ESet.iter 
      (fun i -> 
	shift := ESet.add (Plus (i, Int 1)) 
	    (ESet.add (Minus (i, Int 1)) !shift);
	relate (Plus (i, Int 1)) i;
	relate i (Minus (i, Int 1)))
      !wsset;
    let sset = ESet.union !rsset !wsset in
    if ESet.is_empty !oset then begin
      let iset = ESet.union sset (ESet.union !shift !bset) in
      [sset; ESet.union sset !bset; iset]
    end else begin
      let iset = ESet.union sset (ESet.union !shift !bset) in
      let orig = [sset; ESet.union sset !bset; iset] in
      let offsets set =
	let shift set' = function
	  | Int 0 -> ()
	  | offset ->
	      ESet.iter 
		(fun i -> set' := ESet.add (Minus (i, offset)) !set') 
		set
	in
	let set' = ref set in
	ESet.iter (fun offset -> shift set' offset) !oset;
	!set'
      in
      let rec thread l1 l2 =
	match l1, l2 with
	| h1 :: t1, h2 :: t2 -> h1 :: (h2 :: (thread t1 t2))
	| l, [] | [], l -> l
      in
      thread orig (L.map offsets orig)
    end
  in

  let related1 = function
    | [i] -> ESet.mem i !first1
    | [i; j] -> RSet.mem (i, j) !first2
    | _ -> assert false
  in
  let related2 = function
    | [i] -> true
    | [i; j] -> i <> j && not (related i j)
    | _ -> assert false
  in

  if not !aprop then
    let first = ref true in
    fun () -> if !first then (first := false; expr) else raise (Done false)
  else
    let attempt = ref 1 in
    fun () ->
      let expr =
	match !attempt with
	| 1 -> rewrite (L.nth isets 0) related1 2 expr
	| n when n >= 2 + (L.length isets) -> raise (Done false)
	| n -> rewrite (L.nth isets (n - 2)) related2 2 expr
      in
      incr attempt;
      let rec declare e = function
	| v :: t -> Forall (T.Int, v, declare e t)
	| [] -> e
      in
      declare expr !nvars

let rec sresolution = function
  | Forall (t, v, e) -> Forall (t, v, sresolution e)
  | Or (Not ante, cons) ->
      let conjuncts expr =
	let conj = ref ESet.empty in
	let rec conjuncts = function
	  | And (e1, e2) -> conjuncts e1; conjuncts e2
	  | BExpr e -> conjuncts e
	  | x -> conj := ESet.add x !conj
	in
	conjuncts expr;
	!conj
      in
      let totrue expr tt =
	let rec totrue = function
	  | And (e1, e2) -> And (totrue e1, totrue e2)
	  | BExpr e -> BExpr (totrue e)
	  | x -> if ESet.mem x tt then (Bool true) else x
	in
	totrue expr
      in
      let apply eqns expr = sub eqns expr in
      let sapply eqns exprs =
	let eset = ref ESet.empty in
	ESet.iter (fun e -> eset := ESet.add (apply eqns e) !eset) exprs;
	!eset
      in
      let equations exprs =
	let update eqns e1 e2 =
	  let e2 = apply eqns e2 in
	  (e1, e2) :: (L.map (fun (l, r) -> (l, apply [(e1, e2)] r)) eqns)
	in
	let eqns = ref [] in
	let equations = function
	  | EQ (e1, e2) -> eqns := update !eqns e1 e2
	  | x -> ()
	in
	ESet.iter (fun e -> equations e) exprs;
	!eqns
      in
      let antec = conjuncts ante in
      let eqns = equations antec in
      let antec = sapply eqns antec in
      let cons = apply eqns cons in
      let consc = conjuncts cons in
      let tt = ESet.inter antec consc in
      Or (Not ante, totrue cons tt)
  | x -> x

let rec simplify = function
  | Array (upd, arr) ->
      Array (L.map (fun (e1, e2) -> (simplify e1, simplify e2)) upd, 
	     simplify arr)
  | Length e -> begin
      match simplify e with
      | Array (upd, arr) -> Length arr
      | e -> Length e
  end
  | ArrayAcc (arr, i) -> begin
      let arr = simplify arr in
      let i = simplify i in
      let dflt = ArrayAcc (arr, i) in
      match arr with
      | Array (upd, init) as x -> begin
	  try L.assoc i upd with
	  | Not_found -> dflt
      end
      | _ -> dflt
  end
  | Struct (upd, st) -> 
      Struct (L.map (fun (fld, e) -> (fld, simplify e)) upd, simplify st)
  | StructAcc (e, fld) -> StructAcc (simplify e, fld)
  | FnCall (fn, args) -> FnCall (fn, L.map simplify args)
  | Plus (e1, e2) -> Plus (simplify e1, simplify e2)
  | Minus (e1, e2) -> Minus (simplify e1, simplify e2)
  | Times (e1, e2) -> Times (simplify e1, simplify e2)
  | Div (e1, e2) -> Div (simplify e1, simplify e2)
  | IDiv (e1, e2) -> IDiv (simplify e1, simplify e2)
  | Mod (e1, e2) -> Mod (simplify e1, simplify e2)
  | UMinus e -> UMinus (simplify e)
  | LT (e1, e2) -> LT (simplify e1, simplify e2)
  | LE (e1, e2) -> LE (simplify e1, simplify e2)
  | NE (e1, e2) -> NE (simplify e1, simplify e2)
  | EQ (e1, e2) -> EQ (simplify e1, simplify e2)
  | GE (e1, e2) -> GE (simplify e1, simplify e2)
  | GT (e1, e2) -> GT (simplify e1, simplify e2)
  | BExpr e -> simplify e
  | And (e1, e2) -> begin
      match simplify e1, simplify e2 with
      | Bool b1, Bool b2 -> Bool (b1 && b2)
      | Bool b1, x2 -> if b1 then x2 else Bool false
      | x1, Bool b2 -> if b2 then x1 else Bool false
      | x1, x2 -> And (x1, x2)
  end
  | Or (e1, e2) -> begin
      match simplify e1, simplify e2 with
      | Bool b1, Bool b2 -> Bool (b1 or b2)
      | Bool b1, x2 -> if b1 then Bool true else x2
      | x1, Bool b2 -> if b2 then Bool true else x1
      | x1, x2 -> Or (x1, x2)
  end
  | Not e -> begin
      match simplify e with
      | Bool b -> Bool (not b)
      | x -> Not x
  end
  | Forall (typ, vnm, e) -> begin
      match simplify e with
      | Bool b as x -> x
      | e -> Forall (typ, vnm, e)
  end
  | x -> x

let valid prgm fn expr =
  let valid e =
    let tr = Transform.smt_string (Transform.value_of_expr prgm fn e) in
    if !Cvc.cvc_exec <> "" then not (Cvc.sat tr) else
    if !Yices.yices_exec <> "" then not (Yices.sat tr) else
    raise Not_found
  in

  let expr = sresolution expr in
  print_endline (string_of_expr expr);
  let exprs = rewrite expr in
  try
    let a = ref 1 in
    while true do
      print_endline ("Attempt: " ^ (string_of_int !a));
      incr a;
      match simplify (exprs ()) with
      | Bool b -> if b then raise (Done true)
      | expr -> if valid expr then raise (Done true)
    done;
    (false, expr)
  with
  | Done b -> (b, expr)

let smt_of prgm fn expr =
  let iid = ref 0 in
  let ivar () = incr iid; "V_" ^ (string_of_int !iid) in
  let qivar () = incr iid; "?V_" ^ (string_of_int !iid) in
  let lengths = H.create 13 in
  let vars = ref [] in
  let rec expand = function
    | FnCall ("sorted", [arr; l; u]) ->	
	let ix = Var (qivar ()) in
	let jx = Var (qivar ()) in
	FnCall ("forall", [jx; l; u; 
			   FnCall ("forall", [ix; l; jx; 
					      LE (ArrayAcc (arr, ix), ArrayAcc (arr, jx))])])
    | FnCall ("partitioned", [arr; l1; u1; l2; u2]) ->
	let ix = Var (qivar ()) in
	let jx = Var (qivar ()) in
	Or (GE (u1, l2),
	    FnCall ("forall", [ix; l1; u1;
			       FnCall ("forall", [jx; l2; u2;
						  LE (ArrayAcc (arr, ix), ArrayAcc (arr, jx))])]))
    | FnCall ("eq", [a1; a2; l; u]) ->
	let ix = Var (qivar ()) in
	FnCall ("forall", [ix; l; u; EQ (ArrayAcc (a1, ix), ArrayAcc (a2, ix))])
    | BExpr e -> BExpr (expand e)
    | And (e1, e2) -> And (expand e1, expand e2)
    | Or (e1, e2) -> Or (expand e1, expand e2)
    | Not e -> Not (expand e)
    | Forall (t, s, e) -> Forall (t, s, expand e)
    | x -> x
  in
  let rec nnf = function
    | Bool b -> Bool b
    | FnCall ("forall", [ix; l; u; phi]) -> 
	FnCall ("forall", [ix; l; u; nnf phi])
    | FnCall ("exists", [ix; l; u; phi]) -> 
	FnCall ("exists", [ix; l; u; nnf phi])
    | BExpr e -> BExpr (nnf e)
    | And (e1, e2) -> And (nnf e1, nnf e2)
    | Or (e1, e2) -> Or (nnf e1, nnf e2)
    | Not e -> negate e
    | x -> x
  and negate = function
    | Bool b -> Bool (not b)
    | Var s -> Not (Var s)
    | FnCall ("forall", [ix; l; u; phi]) -> 
	FnCall ("exists", [ix; l; u; negate phi])
    | FnCall ("exists", [ix; l; u; phi]) -> 
	FnCall ("forall", [ix; l; u; negate phi])
    | BExpr e -> BExpr (negate e)
    | And (e1, e2) -> Or (negate e1, negate e2)
    | Or (e1, e2) -> And (negate e1, negate e2)
    | Not e -> nnf e
    | LT (e1, e2) -> GE (e1, e2)
    | LE (e1, e2) -> GT (e1, e2)
    | NE (e1, e2) -> EQ (e1, e2)
    | EQ (e1, e2) -> NE (e1, e2)
    | GE (e1, e2) -> LT (e1, e2)
    | GT (e1, e2) -> LE (e1, e2)
    | x -> print_endline (string_of_expr x); assert false
  in
  let rec smt_of = function
    | Bool b -> if b then "true" else "false"
    | Int n -> string_of_int n
    | Var s -> s
    | Length e as x -> begin
	match simplify x with
	| Length e -> begin
	    try H.find lengths e with
	    | Not_found ->
		let vnm = ivar () in
		vars := (T.Int, vnm) :: !vars;
		H.add lengths e vnm;
		vnm
	end
	| _ -> assert false
    end
    | Array (upd, e) ->
	let rec store = function
	  | [] -> smt_of e
	  | (i, e) :: t ->
	      "(store " ^ (store t) ^ " " ^ (smt_of i) ^ " " ^ (smt_of e) ^ ")"
	in
	store upd
    | ArrayAcc (e, i) -> 
	"(select " ^ (smt_of e) ^ " " ^ (smt_of i) ^ ")"
    | FnCall ("forall", [ix; l; u; phi]) ->
	let ixnm = qivar () in
	let ix' = Var ixnm in
	"(forall (" ^ ixnm ^ 
	" Int) (implies (and (<= " ^ (smt_of l) ^ " " ^ ixnm ^ 
	") (<= " ^ ixnm ^ " " ^ (smt_of u) ^ ")) " ^ 
	(smt_of (sub [(ix, ix')] phi)) ^ "))"
    | FnCall ("exists", [ix; l; u; phi]) ->
	let ixnm = qivar () in
	let ix' = Var ixnm in
	"(exists (" ^ ixnm ^ 
	" Int) (and (and (<= " ^ (smt_of l) ^ " " ^ ixnm ^ 
	") (<= " ^ ixnm ^ " " ^ (smt_of u) ^ ")) " ^ 
	(smt_of (sub [(ix, ix')] phi)) ^ "))"
    | Plus (e1, e2) -> "(+ " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | Minus (e1, e2) -> "(- " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | Times (e1, e2) -> "(* " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | UMinus e -> "(~ " ^ (smt_of e) ^ ")"
    | LT (e1, e2) -> "(< " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | LE (e1, e2) -> "(<= " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | NE (e1, e2) -> "(not (= " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ "))"
    | EQ (e1, e2) -> "(= " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | GE (e1, e2) -> "(>= " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | GT (e1, e2) -> "(> " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | BExpr e -> smt_of e
    | And (e1, e2) -> "(and " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | Or (e1, e2) -> "(or " ^ (smt_of e1) ^ " " ^ (smt_of e2) ^ ")"
    | Not e -> "(not " ^ (smt_of e) ^ ")"
    | x -> print_endline (string_of_expr x); assert false
  in
  let rec smt = function
    | Forall (t, v, e) ->
	vars := (t, v) :: !vars;
	smt e
    | x -> 
	let sdecls () =
          String.concat "\n"
            (L.map 
               (fun (typ, vnm) -> 
		 match T.expand typ with
		 | T.Bool -> ":extrapreds ((" ^ vnm ^ "))"
		 | T.Int -> ":extrafuns ((" ^ vnm ^ " Int))"
		 | T.Array _ -> ":extrafuns ((" ^ vnm ^ " Array))"
		 | _ -> assert false)
               !vars)
	in
	let x' = smt_of (negate (expand x)) in
	"(benchmark tmp\n:source {piVC}\n:status unknown\n:logic QF_AUFLIA\n"
	^ (sdecls ()) ^ "\n:formula\n" ^ x' ^ "\n)"
  in
  smt expr

(*
(* --- read-over-write --- *)

type state = 
    { expr : Vcg.expr;  (* current expression *)
      rows : ESet.t;    (* read-over-write *)
    }

let rows expr =
  let r = ref ESet.empty in
  let rec rows = function
    | Array (upd, arr) ->
	L.iter (fun (e1, e2) -> rows e1; rows e2) upd; 
	rows arr
    | Length e -> rows e
    (* nontrivial read-over-write *)
    | ArrayAcc (Array (h :: t, b) as arr, i) as x -> 
	r := ESet.add x !r;
	rows arr;
	rows i
    | Struct (upd, b) -> L.iter (fun (_, e) -> rows e) upd; rows b
    | StructAcc (st, _) -> rows st
    | FnCall (_, args) -> L.iter rows args
    | Plus (e1, e2)
    | Minus (e1, e2)
    | Times (e1, e2)
    | Div (e1, e2)
    | IDiv (e1, e2)
    | Mod (e1, e2)
    | LT (e1, e2)
    | LE (e1, e2)
    | NE (e1, e2)
    | EQ (e1, e2)
    | GE (e1, e2)
    | GT (e1, e2)
    | And (e1, e2)
    | Or (e1, e2) -> rows e1; rows e2
    | UMinus e
    | BExpr e
    | Not e
    | Forall (_, _, e) -> rows e
    | _ -> ()
  in
  rows expr;
  !r

let sub_row expr l r =
  let rec sub = function
    | Array (upd, arr) ->
	Array (L.map (fun (e1, e2) -> (sub e1, sub e2)) upd, sub arr)
    | Length e -> Length (sub e)
    | ArrayAcc (arr, i) as x -> if x = l then r else ArrayAcc (sub arr, sub i)
    | Struct (upd, st) -> 
	Struct (L.map (fun (fld, e) -> (fld, sub e)) upd, sub st)
    | StructAcc (st, fld) -> StructAcc (sub st, fld)
    | FnCall (f, args) -> FnCall (f, L.map sub args)
    | Plus (e1, e2) -> Plus (sub e1, sub e2)
    | Minus (e1, e2) -> Minus (sub e1, sub e2)
    | Times (e1, e2) -> Times (sub e1, sub e2)
    | Div (e1, e2) -> Div (sub e1, sub e2)
    | IDiv (e1, e2) -> IDiv (sub e1, sub e2)
    | Mod (e1, e2) -> Mod (sub e1, sub e2)
    | UMinus e -> UMinus (sub e)
    | LT (e1, e2) -> LT (sub e1, sub e2)
    | LE (e1, e2) -> LE (sub e1, sub e2)
    | NE (e1, e2) -> NE (sub e1, sub e2)
    | EQ (e1, e2) -> EQ (sub e1, sub e2)
    | GE (e1, e2) -> GE (sub e1, sub e2)
    | GT (e1, e2) -> GT (sub e1, sub e2)
    | BExpr e -> BExpr (sub e)
    | And (e1, e2) -> And (sub e1, sub e2)
    | Or (e1, e2) -> Or (sub e1, sub e2)
    | Not e -> Not (sub e)
    | Forall (t, v, e) -> Forall (t, v, sub e)
    | x -> x
  in
  sub expr

let sub_arr expr map =
  let rec sub = function
    | Array (upd, arr) as x -> begin
	try H.find map x with
	| Not_found -> 
	    Array (L.map (fun (e1, e2) -> (sub e1, sub e2)) upd, sub arr)
    end
    (* don't substitute directly for Array, as length is same *)
    | Length (Array (upd, arr)) -> 
	Length (Array (L.map (fun (e1, e2) -> (sub e1, sub e2)) upd, sub arr))
    | Length e -> Length (sub e)
    | ArrayAcc (arr, i) -> ArrayAcc (sub arr, sub i)
    | Struct (upd, st) -> 
	Struct (L.map (fun (fld, e) -> (fld, sub e)) upd, sub st)
    | StructAcc (st, fld) -> StructAcc (sub st, fld)
    | FnCall (f, args) -> FnCall (f, L.map sub args)
    | Plus (e1, e2) -> Plus (sub e1, sub e2)
    | Minus (e1, e2) -> Minus (sub e1, sub e2)
    | Times (e1, e2) -> Times (sub e1, sub e2)
    | Div (e1, e2) -> Div (sub e1, sub e2)
    | IDiv (e1, e2) -> IDiv (sub e1, sub e2)
    | Mod (e1, e2) -> Mod (sub e1, sub e2)
    | UMinus e -> UMinus (sub e)
    | LT (e1, e2) -> LT (sub e1, sub e2)
    | LE (e1, e2) -> LE (sub e1, sub e2)
    | NE (e1, e2) -> NE (sub e1, sub e2)
    | EQ (e1, e2) -> EQ (sub e1, sub e2)
    | GE (e1, e2) -> GE (sub e1, sub e2)
    | GT (e1, e2) -> GT (sub e1, sub e2)
    | BExpr e -> BExpr (sub e)
    | And (e1, e2) -> And (sub e1, sub e2)
    | Or (e1, e2) -> Or (sub e1, sub e2)
    | Not e -> Not (sub e)
    | Forall (t, v, e) -> Forall (t, v, sub e)
    | x -> x
  in
  sub expr

let var_type expr v =
  let rec vtype = function
    | Forall (typ, v', e) -> if v' = v then typ else vtype e
    | _ -> assert false
  in
  vtype expr

let array_type expr arr =
  let rec atype = function
    | Var v -> var_type expr v
    | Array (_, arr) -> atype arr
    | _ -> assert false
  in
  atype arr

let partial state =
  let arrays = ref ESet.empty in
  ESet.iter
    (function
      | ArrayAcc (arr, _) -> arrays := ESet.add arr !arrays
      | _ -> assert false)
    state.rows;

  let nvars = ref [] in
  let id = 
    let id = ref 0 in
    fun () -> incr id; string_of_int !id
  in
  let nvar arr =
    let v = "arr_" ^ (id ()) in
    nvars := (array_type state.expr arr, v) :: !nvars;
    v
  in
  let map = H.create 13 in
  ESet.iter (fun arr -> let v = nvar arr in H.add map arr (Var v)) !arrays;

  let sub = sub_arr state.expr map in
  let rec declare = function
    | (typ, v) :: t -> Forall (typ, v, (declare t))
    | [] -> sub
  in
  declare !nvars

let valid prgm fn expr =
  let implies ante cons =
    let rec implies = function
      | Forall (t, s, e) -> Forall (t, s, implies e)
      | x -> Or (Not ante, x)
    in
    implies cons
  in
  let esmap f es =
    let es' = ref ESet.empty in
    ESet.iter (fun e -> es' := ESet.add (f e) !es') es;
    !es'
  in
  let split_left state = function
    | ArrayAcc (Array ((j, v) :: t, arr), i) as x ->
	{ expr = implies (EQ (i, j)) (sub_row state.expr x v);
	  rows = esmap (fun e -> sub_row e x v) (ESet.remove x state.rows);
	}
    | _ -> assert false
  in
  let split_right state = function
    | ArrayAcc (Array ((j, v) :: t, arr), i) as x -> begin
	match t with
	| [] -> 
	    let r = ArrayAcc (arr, i) in
	    { expr = implies (NE (i, j)) (sub_row state.expr x r);
	      rows = esmap (fun e -> sub_row e x r) (ESet.remove x state.rows);
	    }
	| _ -> 
	    let r = ArrayAcc (Array (t, arr), i) in
	    { expr = implies (NE (i, j)) (sub_row state.expr x r);
	      rows = esmap (fun e -> sub_row e x r) 
		(ESet.add r (ESet.remove x state.rows));
	    }
    end
    | _ -> assert false
  in

  let valid e =
    let tr = Transform.annoi_string (Transform.value_of_expr prgm fn e) in
    print_endline tr;
    not (Iannoi.sat tr)
  in

  let rec search state =
    if valid (partial state) then true else
    if ESet.is_empty state.rows then false else
    let row = ESet.choose state.rows in
    if Random.bool () then
      (search (split_left state row)) 
	&& (search (split_right state row))
    else
      (search (split_right state row)) 
	&& (search (split_left state row))
  in

  let expr = sresolution expr in
  print_endline (string_of_expr expr);
  let exprs = rewrite expr in
  try
    while true do
      let expr = exprs () in
      if search { expr = expr; rows = rows expr } then raise (Done true)
    done;
    false
  with
  | Done b -> b
*)
