(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module T = Type
module V = Vcg

module A = Array
module H = Hashtbl
module L = List

(* Failure to translate *)
exception Fail

type sent =
  | True
  | False
  | Pred of pred
  | EQ of value * value
  | EQZ of nval
  | GEZ of nval
  | GTZ of nval
  | Not of sent
  | And of sent * sent
  | Or of sent * sent
  | Exists of Type.typ * string * sent

and read =
  | Var of string
  | ArrayAcc of array * nval
  | StructAcc of read * string
  | FnCall of string * value list

and pred =
  | PVar of read
  | PPred of string * value list

and array = (nval * value) list * read

and nval =
  | Constant of int
  | Term of int * read
  | Sum of nval list

and value =
  | Bool of sent
  | Read of read
  | Array of array
  | Number of nval

(* --- for annoi --- *)

let rec string_of_sent sent =
  let typs = ref T.TSet.empty in
  let vars = ref [] in
  let rec sos = function
    | True -> "true"
    | False -> "false"
    | Pred p -> (string_of_pred p)
    | EQ (v1, v2) -> (string_of_value v1) ^ " == " ^ (string_of_value v2)
    | EQZ nv -> (string_of_nval nv) ^ " == 0"
    | GEZ nv -> (string_of_nval nv) ^ " >= 0"
    | GTZ nv -> (string_of_nval nv) ^ " > 0"
    | Not s -> "!(" ^ (sos s) ^ ")"
    | And (s1, s2) -> "(" ^ (sos s1) ^ ") & (" ^ (sos s2) ^ ")"
    | Or (s1, s2) -> "(" ^ (sos s1) ^ ") | (" ^ (sos s2) ^ ")"
    | Exists (typ, vnm, s) ->
        typs := T.TSet.add (T.expand typ) !typs;
        vars := (typ, vnm) :: !vars;
        sos s
  in
  let fix s =
    if String.length s > 6 && String.sub s 0 6 = "struct" then
    String.sub s 7 (String.length s - 7) else
    s
  in
  let rec short_sot = function
    | T.Int -> "int"
    | T.Float -> "real"
    | T.Bool -> "bool"
    | T.Alias nm -> fix nm
    | T.UIT _ -> raise Fail
    | T.Struct (nm, _) -> fix nm
    | T.Array t -> (short_sot t) ^ "_AR"
  in
  let rec sot = function
    | T.Int -> "int"
    | T.Float -> "real"
    | T.Bool -> "bool"
    | T.Alias nm -> fix nm
    | T.UIT _ -> raise Fail
    | T.Struct (nm, flds) ->
        "type " ^ nm ^ "(" 
        ^ (String.concat ", " 
             (L.map 
                (fun (typ, fld) -> fld ^ " [" ^ (short_sot typ) ^ "]") 
                flds))
        ^ ")"
    | T.Array t -> 
        "type " ^ (short_sot t) ^ "_AR = [int -> " ^ (short_sot t) ^ "]"
  in
  match sent with
  | Exists _ -> begin
      let stypes () =
        (* TODO: dependency ordering *)
        let s = ref "" in
        T.TSet.iter
          (function 
            | T.Int | T.Float -> ()  (* nothing to declare *)
            | T.Alias _ -> assert false
            | T.UIT _ -> raise Fail
            | x -> s := (sot x) ^ ";\n" ^ !s)
          !typs;
        !s
      in
      let sdecls () =
        String.concat "\n"
          (L.map 
             (fun (typ, vnm) -> (short_sot typ) ^ " " ^ vnm ^ ";") 
             !vars)
      in
      let sos = sos sent in
      (stypes ()) ^ "\n" ^ (sdecls ()) ^ "\n\nassert\n\n" ^ sos
  end
  | _ -> sos sent

and string_of_read = function
  | Var s -> s
  | ArrayAcc (arr, i) -> (string_of_array arr) ^ "(" ^ (string_of_nval i) ^ ")"
  | StructAcc (st, fld) -> fld ^ "(" ^ (string_of_read st) ^ ")"
  | FnCall (fnm, args) -> 
      fnm ^ "(" ^ (String.concat ", " (L.map string_of_value args)) ^ ")"

and string_of_pred = function
  | PVar rd -> string_of_read rd
  | PPred (pnm, args) -> 
      pnm ^ "(" ^ (String.concat ", " (L.map string_of_value args)) ^ ")"

and string_of_array (upd, init) =
  (string_of_read init)
  ^ (String.concat "" 
       (L.map 
          (fun (i, v) -> 
            "{" ^ (string_of_nval i) ^ " <- " ^ (string_of_value v) ^ "}")
          upd))

and string_of_nval = function
  | Constant i -> string_of_int i
  | Term (1, rd) -> string_of_read rd
  | Term (c, rd) -> (string_of_int c) ^ " " ^ (string_of_read rd)
  | Sum nvs -> String.concat " + " (L.map string_of_nval nvs)

and string_of_value = function
  | Bool s -> string_of_sent s
  | Read rd -> string_of_read rd
  | Array arr -> string_of_array arr
  | Number nv -> string_of_nval nv

let annoi_string = function
  | Bool (Exists _) as x -> string_of_value x
  | x -> "assert\n\n" ^ (string_of_value x)

(* --- SMT-LIB format --- *)

let rec smt_of_sent sent = smt_of_sent1 true sent
and smt_of_sent1 top sent =
  let typs = ref T.TSet.empty in
  let vars = ref [] in
  let rec sos = function
    | True -> "true"
    | False -> "false"
    | Pred p -> (smt_of_pred p)
    | EQ (v1, v2) -> 
        "(= " ^ (smt_of_value v1) ^ " " ^ (smt_of_value v2) ^ ")"
    | EQZ nv -> "(= 0 " ^ (smt_of_nval nv) ^ ")"
    | GEZ nv -> "(<= 0 " ^ (smt_of_nval nv) ^ ")"
    | GTZ nv -> "(< 0 " ^ (smt_of_nval nv) ^ ")"
    | Not s -> "(not " ^ (sos s) ^ ")"
    | And (s1, s2) -> "(and " ^ (sos s1) ^ " " ^ (sos s2) ^ ")"
    | Or (s1, s2) -> "(or " ^ (sos s1) ^ " " ^ (sos s2) ^ ")"
    | Exists (typ, vnm, s) ->
        typs := T.TSet.add (T.expand typ) !typs;
        vars := (typ, vnm) :: !vars;
        sos s
  in
  let fix s =
    if String.length s > 6 && String.sub s 0 6 = "struct" then
    String.sub s 7 (String.length s - 7) else
    s
  in
  match sent with
  | Exists _ -> begin
      let sdecls () =
        String.concat "\n"
          (L.map 
             (fun (typ, vnm) -> 
               match T.expand typ with
               | T.Bool -> ":extrapreds ((" ^ vnm ^ "))"
               | T.Int -> ":extrafuns ((" ^ vnm ^ " Int))"
               | T.Array _ -> ":extrafuns ((" ^ vnm ^ " Array))"
	       | T.Struct (_, flds) -> begin
                 (* temporary solution *)
                 let field (typ, fld) =
		   let vnm = vnm ^ "." ^ fld in
		   match typ with
		   | T.Bool -> ":extrapreds (( " ^ vnm ^ "))"
		   | T.Int -> ":extrafuns ((" ^ vnm ^ " Int))"
		   | T.Array _ -> ":extrafuns ((" ^ vnm ^ " Array))"
		   | _ -> raise Fail
		 in
		 String.concat "\n" (L.map field flds)
	       end
               | _ -> raise Fail)
             !vars)
      in
      let sos = sos sent in
      "(benchmark tmp\n:source {pvc}\n:status unknown\n:logic QF_AUFLIA\n"
      ^ (sdecls ()) ^ "\n:formula\n" ^ sos ^ "\n)"
  end
  | _ -> 
      let sos = sos sent in
      if not top then sos else
      "(benchmark tmp\n:source {pvc}\n:status unknown\n:logic QF_AUFLIA\n"
      ^ ":formula\n" ^ sos ^ "\n)"

and smt_of_read = function
  | Var s -> s
  | ArrayAcc (arr, i) -> 
      "(select " ^ (smt_of_array arr) ^ " " ^ (smt_of_nval i) ^ ")"
  | StructAcc (st, fld) -> (smt_of_read st) ^ "." ^ fld
  | FnCall (fnm, args) -> 
      "(" ^ fnm ^ " " ^ (String.concat " " (L.map smt_of_value args)) ^ ")"

and smt_of_pred = function
  | PVar rd -> smt_of_read rd
  | PPred (pnm, args) -> 
      "(" ^ pnm ^ " " ^ (String.concat " " (L.map smt_of_value args)) ^ ")"

and smt_of_array (upd, init) =
  let rec stores = function
    | (i, v) :: t -> 
        "(store " ^ (stores t) ^ " " ^ (smt_of_nval i) ^ " "
        ^ (smt_of_value v) ^ ")"
    | [] -> smt_of_read init
  in
  stores upd

and smt_of_nval = function
  | Constant i -> 
      if i < 0 then "(~ " ^ (smt_of_nval (Constant (-i))) ^ ")" else
      string_of_int i
  | Term (1, rd) -> smt_of_read rd
  | Term (c, rd) -> 
      "(* " ^ (smt_of_nval (Constant c)) ^ " " ^ (smt_of_read rd) ^ ")"
  | Sum nvs -> 
      "(+ " ^ (String.concat " " (L.map smt_of_nval nvs)) ^ ")"

and smt_of_value = function
  | Bool s -> smt_of_sent1 false s
  | Read rd -> smt_of_read rd
  | Array arr -> smt_of_array arr
  | Number nv -> smt_of_nval nv

let smt_string = function
  | Bool x -> smt_of_sent x
  | x -> smt_of_value x

(* --- end SMT-LIB format --- *)

module SSet = Set.Make (
  struct
    type t = string
    let compare t1 t2 = compare t1 t2
  end)

let variables v =
  let vars = ref SSet.empty in
  let rec sentv = function
    | Pred p -> predv p
    | EQ (v1, v2) -> valuev v1; valuev v2
    | EQZ nv | GEZ nv | GTZ nv -> nvalv nv
    | Not s -> sentv s
    | And (s1, s2) | Or (s1, s2) -> sentv s1; sentv s2
    | _ -> ()
  and readv = function
    | Var s -> vars := SSet.add s !vars
    | ArrayAcc (arr, i) -> arrayv arr; nvalv i
    | StructAcc (st, _) -> readv st
    | FnCall (_, args) -> L.iter valuev args
  and predv = function
    | PVar rd -> readv rd
    | PPred (_, args) -> L.iter valuev args
  and arrayv (upd, init) =
    L.iter (fun (i, v) -> nvalv i; valuev v) upd;
    readv init
  and nvalv = function
    | Term (_, rd) -> readv rd
    | Sum ts -> L.iter nvalv ts
    | _ -> ()
  and valuev = function
    | Bool s -> sentv s
    | Read rd -> readv rd
    | Array arr -> arrayv arr
    | Number nv -> nvalv nv
  in
  valuev v;
  !vars

let value_of_expr prgm fn expr =
  let fvars = ref [] in
  let vid = ref 0 in
  let fresh typ =
    incr vid;
    let vnm = "tr" ^ (string_of_int !vid) in
    fvars := (typ, vnm) :: !fvars;
    vnm
  in
  let lengths = H.create 13 in
  let lid = ref 0 in
  let length_of arr =
    try H.find lengths arr with
    | Not_found ->
        incr lid;
        let vnm = "length" ^ (string_of_int !lid) in
        fvars := (T.Int, vnm) :: !fvars;
        let v = Var vnm in
        H.add lengths arr v;
        v
  in
  let fresh_vars () = !fvars in

  let idivvars = ref [] in
  let idiv_var expr =
    try (false, L.assoc expr !idivvars) with
    | Not_found ->
        let fv = Var (fresh T.Int) in
        idivvars := (expr, fv) :: !idivvars;
        (true, fv)
  in
  let imodvars = ref [] in
  let imod_vars expr =
    try (false, L.assoc expr !imodvars) with
    | Not_found ->
        let fv = Var (fresh T.Int) in
        let c = Var (fresh T.Int) in
        imodvars := (expr, (fv, c)) :: !imodvars;
        (true, (fv, c))
  in

  (* type of Vcg.expr *)
  let svars = 
    let vars = ref [] in
    let rec v = function
      | V.Forall (typ, vnm, e) -> vars := (typ, vnm) :: !vars; v e
      | _ -> ()
    in
    v expr;
    !vars
  in
  let type_of expr' = V.type_of_expr (svars @ !fvars) prgm fn expr' in
  let fields typ =
    match T.expand typ with
    | T.Struct (_, flds) -> L.map (fun (_, fld) -> fld) flds
    | _ -> assert false
  in

  let req = ref [] in
  let require sent = req := sent :: !req in
  let required () = 
    match !req with
    | [] -> True
    | h :: [] -> h
    | h :: t -> L.fold_left (fun e1 e2 -> And (e1, e2)) h t
  in

  (* if not exact, then abstract ints to reals *)
  let exact = ref false in

  let rec sent_of expr =
    let rec nnf = function
      | V.BExpr e -> V.BExpr (nnf e)
      | V.And (e1, e2) -> V.And (nnf e1, nnf e2)
      | V.Or (e1, e2) -> V.Or (nnf e1, nnf e2)
      | V.Not e -> negate e
      | x -> x
    and negate = function
      | V.BExpr e -> V.BExpr (negate e)
      | V.And (e1, e2) -> V.Or (negate e1, negate e2)
      | V.Or (e1, e2) -> V.And (negate e1, negate e2)
      | V.Not e -> nnf e
      | V.LT (e1, e2) -> V.GE (e1, e2)
      | V.LE (e1, e2) -> V.GT (e1, e2)
      | V.GE (e1, e2) -> V.LT (e1, e2)
      | V.GT (e1, e2) -> V.LE (e1, e2)
      | V.EQ (e1, e2) -> V.NE (e1, e2)
      | V.NE (e1, e2) -> V.EQ (e1, e2)
      | x -> V.Not x
    in
    let rec so = function
      | V.Bool b -> if b then True else False
      | V.Var _ as x -> Pred (pred_of x)
      | V.ArrayAcc _ as x -> Pred (pred_of x)
      | V.StructAcc _ as x -> Pred (pred_of x)
      | V.FnCall _ as x -> Pred (pred_of x)
      | V.BExpr e -> sent_of e
      | V.LT (e1, e2) -> GTZ (nval_of true (V.Minus (e2, e1)))
      | V.LE (e1, e2) -> GEZ (nval_of true (V.Minus (e2, e1)))
      | V.GE (e1, e2) -> GEZ (nval_of true (V.Minus (e1, e2)))
      | V.GT (e1, e2) -> GTZ (nval_of true (V.Minus (e1, e2)))
      | V.EQ (e, V.Anything) -> sent_of (V.EQ (e, V.Var (fresh (type_of e))))
      | V.EQ (e1, e2) -> begin
          match type_of e1 with
          | T.Int
          | T.Float -> EQZ (nval_of true (V.Minus (e1, e2)))
          | _ -> EQ (value_of e1, value_of e2)
      end
      | V.NE (e1, e2) -> Not (sent_of (V.EQ (e1, e2)))
      (* PREVIOUSLY: != --> < or >
      | V.NE (e1, e2) -> begin
          match type_of e1 with
          | T.Int
          | T.Float -> sent_of (V.Or (V.LT (e1, e2), V.GT (e1, e2)))
          | _ -> sent_of (V.Not (V.EQ (e1, e2)))
      end
      *)
      | V.And (e1, e2) -> And (sent_of e1, sent_of e2)
      | V.Or (e1, e2) -> Or (sent_of e1, sent_of e2)
      | V.Not e -> Not (sent_of e)
      | _ -> raise Fail
    in
    so (nnf expr)
            
  and read_of expr =
    let rec base_array = function
      | V.Array (_, init) -> base_array init
      | x -> x
    in
    let rec rd = function
      | V.Var s -> Var s
      | V.ArrayAcc (arr, i) -> ArrayAcc (array_of arr, nval_of false i)
      | V.StructAcc (st, fld) -> StructAcc (struct_of st, fld)
      | V.Length e -> 
          (* length of base array (same as updated array) *)
          let base = base_array e in
          length_of base
      | V.FnCall (fnm, args) -> FnCall (fnm, L.map value_of args)
      | _ -> raise Fail
    in
    rd expr

  and pred_of = function
    | V.Var _ as x -> PVar (read_of x)
    | V.ArrayAcc _ as x -> PVar (read_of x)
    | V.StructAcc _ as x -> PVar (read_of x)
    | V.FnCall (fnm, args) -> PPred (fnm, L.map value_of args)
    | _ -> raise Fail

  and array_of = function
    | V.Var _ as x -> ([], read_of x)
    | V.ArrayAcc _ as x -> ([], read_of x)
    | V.StructAcc _ as x -> ([], read_of x)
    | V.FnCall _ as x -> ([], read_of x)
    | V.Array (upd, init) ->
        let upd' = L.map (fun (i, v) -> (nval_of false i, value_of v)) upd in
        (upd', read_of init)
    | _ -> raise Fail

  and struct_of = function
    | V.Var _ as x -> read_of x
    | V.ArrayAcc _ as x -> read_of x
    | V.StructAcc _ as x -> read_of x
    | V.FnCall _ as x -> read_of x
    | V.Struct (upd, init) ->
        (* introduce new var, equate fields with upd and init's fields *)
        let typ = type_of init in
        let fv = fresh typ in
        let v = Var fv in
        let init = read_of init in
        let asrt =
          let con =
            L.map
              (fun fld -> 
                let expr =
                  try value_of (L.assoc fld upd) with
                  | Not_found -> Read (StructAcc (init, fld))
                in
                EQ (Read (StructAcc (v, fld)), expr))
              (fields typ)
          in
          match con with
          | [] -> True
          | h :: [] -> h
          | h :: t -> L.fold_left (fun e1 e2 -> And (e1, e2)) h t
        in
        require asrt;
        v
    | _ -> raise Fail

  and nval_of reduce expr =
    let iof f =
      (* TODO *)
      let iof = int_of_float f in
      if abs_float (float_of_int iof -. f) < 0.0001 then (iof, 1) else
      raise Fail
    in

    (* good ol' Euclid *)
    let rec gcd m n =
      assert (n > 0);
      let n' = m mod n in
      if n' == 0 then n else gcd n n'
    in
    let gcds = function
      | h :: t -> L.fold_left gcd h t
      | [] -> 1
    in

    (* put in normal form of (num, den, read option) list *)
    let sum terms =
      let mults : (read option * (int * int)) list ref = ref [] in
      let get_coeff x =
        try L.assoc x !mults with
        | Not_found -> (0, 1)
      in
      let set_coeff x n d =
        let m = L.remove_assoc x !mults in
        mults := if n = 0 then m else (x, (n, d)) :: m
      in
      let rec sum = function
        | (n, d, x) :: t ->
            let n', d' = get_coeff x in
            let n = (n' * d) + (n * d') in
            let d = d' * d in
            let gcd = gcd n d in
            let n = n / gcd in
            let d = d / gcd in
            set_coeff x n d;
            sum t
        | [] -> ()
      in
      sum terms;
      L.map (fun (x, (n, d)) -> (n, d, x)) !mults
    in
    let times s n d =
      L.map 
        (fun (n', d', x) ->
          let n = n * n' in
          let d = d * d' in
          let gcd = gcd n d in
          (n / gcd, d / gcd, x))
        s
    in
    (* convert normal form to nval *)
    let nval sum =
      let s =
	if reduce then begin
	  let mult = ref 1 in
	  L.iter (fun (_, d, _) -> mult := !mult * d) sum;
	  let sum =
            if !mult != 1 then
              L.map (fun (n, d, x) -> (n * (!mult / d), 1, x)) sum else
              sum
	  in
	  let acoeffs = L.map (fun (n, _, _) -> abs n) sum in
	  let gcd = gcds (L.filter (fun n -> n > 0) acoeffs) in
          L.map (function
            | (n, 1, None) -> Constant (n / gcd)
            | (n, 1, Some x) -> Term (n / gcd, x)
            | _ -> assert false)
            sum
	end else
          L.map (function
            | (n, 1, None) -> Constant n
            | (n, 1, Some x) -> Term (n, x)
            | _ -> raise Fail)
            sum
      in
      match s with
      | [] -> Constant 0
      | h :: [] -> h
      | x -> Sum x
    in

    let rec nv = function
      | V.Int i -> [(i, 1, None)]
      | V.Float f -> 
          let n, d = iof f in
          [(n, d, None)]
      | V.Var _ as x -> [(1, 1, Some (read_of x))]
      | V.ArrayAcc _ as x -> [(1, 1, Some (read_of x))]
      | V.StructAcc _ as x -> [(1, 1, Some (read_of x))]
      | V.Length _ as x -> [(1, 1, Some (read_of x))]
      | V.FnCall _ as x -> [(1, 1, Some (read_of x))]
      | V.Plus (e1, e2) -> sum ((nv e1) @ (nv e2))
      | V.Minus (e1, e2) -> sum ((nv e1) @ (nv (V.UMinus e2)))
      | V.UMinus e -> L.map (fun (n, d, x) -> (-n, d, x)) (nv e)
      | V.Times (e1, e2) -> begin
          match sum (nv e1), sum (nv e2) with
          | [(n, d, None)], s
          | s, [(n, d, None)] -> times s n d
          | _ -> raise Fail
      end
      | V.Div (e1, e2) -> begin
          match sum (nv e1), sum (nv e2) with
          | s, [(n, d, None)] -> times s d n
          | _ -> raise Fail
      end
      | V.IDiv (e1, e2) as x ->
          exact := true;
          let isnew, fv = idiv_var x in
          if not isnew then [(1, 1, Some fv)] else begin
            match sum (nv e1), sum (nv e2) with
            | s, [(n, d, None)] ->
                let div = n / d in  (* integer division of constants *)
                (* s = fv * div + c, for some c in [-div + 1, div - 1] *)
                let asrt =
                  let dis = ref [] in
                  for c = -div + 1 to div - 1 do
                    let s' = nval s in
                    (* constraint on sign of s *)
                    let gzc = 
                      if c > 0 then GEZ s' else 
                      if c < 0 then GEZ (nval (times s (-1) 1)) else 
                      True
                    in
                    dis := 
                      (And (gzc,
                            (EQZ (Sum (s' :: 
                                       [Term (-div, fv); Constant (-c)])))))
                      :: !dis
                  done;
                  match !dis with
                  | [] -> False
                  | h :: [] -> h
                  | h :: t -> L.fold_left (fun e1 e2 -> Or (e1, e2)) h t
                in
                require asrt;
                [(1, 1, Some fv)]
            | _ -> raise Fail
          end
      | V.Mod (e1, e2) as x ->
          exact := true;
          let isnew, (fv, c) = imod_vars x in
          if not isnew then [(1, 1, Some c)] else begin
            match sum (nv e1), sum (nv e2) with
            | s, [(n, d, None)] ->
                let md = n / d in  (* integer division of constants *)
                (* s = fv * md + c, for some c in [-div + 1, div - 1] *)
                let s = nval s in
                let asrt =
                  let main = EQZ (Sum (s :: [Term (-md, fv); Term (-1, c)])) in
                  (* c <= 0 or c >= 0 *)
                  Or (And (And (main, GEZ (Term (1, c))), 
                           GTZ (Sum [Term (-1, c); Constant md])),
                      And (And (main, GEZ (Term (-1, c))),
                           GTZ (Sum [Term (1, c); Constant md])))
                in
                require asrt;
                [(1, 1, Some c)]
            | _ -> raise Fail
          end
      | x -> 
          print_endline (V.string_of_expr x);
          raise Fail
    in
    nval (nv expr)

  and value_of expr =
    let lval x = 
      match type_of x with
      | T.Bool -> Bool (sent_of x)
      | T.Int
      | T.Float -> Number (nval_of false x)
      | T.Array _ -> Array (array_of x)
      | _ -> Read (read_of x)
    in
    match expr with
    | V.Bool _ as x -> Bool (sent_of x)
    | V.Int _ as x -> Number (nval_of false x)
    | V.Float _ as x -> Number (nval_of false x)
    | V.Var _ as x -> lval x
    | V.Array _ as x -> Array (array_of x)
    | V.Length _ as x -> Read (read_of x)
    | V.ArrayAcc _ as x -> lval x
    | V.Struct _ as x -> Read (struct_of x)
    | V.StructAcc _ as x -> lval x
    | V.FnCall _ as x -> begin
        match type_of x with
        | T.Bool -> Bool (sent_of x)
        | _ -> Read (read_of x)
    end
    | V.Plus _ as x -> Number (nval_of false x)
    | V.Minus _ as x -> Number (nval_of false x)
    | V.Times _ as x -> Number (nval_of false x)
    | V.Div _ as x -> Number (nval_of false x)
    | V.IDiv _ as x -> Number (nval_of false x)
    | V.Mod _ as x -> Number (nval_of false x)
    | V.UMinus _ as x -> Number (nval_of false x)
    | V.BExpr _ as x -> Bool (sent_of x)
    | V.And _ as x -> Bool (sent_of x)
    | V.Or _ as x -> Bool (sent_of x)
    | V.Not _ as x -> Bool (sent_of x)
    | _ -> raise Fail
  in

  (* need to push this into the DP *)
  let int_to_float expr =
    let rec sent_itf = function
      | Pred p -> Pred (pred_itf p)
      | EQZ nv ->
          let rec negate = function
            | Constant i -> Constant (-i)
            | Term (i, r) -> Term (-i, r)
            | Sum ts -> Sum (L.map negate ts)
          in
          And (GEZ nv, GEZ (negate nv))
      | GTZ nv ->
          let nv' = match nv with
          | Constant i -> Constant (i - 1)
          | Term _ as x -> Sum [x; Constant (-1)]
          | Sum ts ->
              let rec decr = function
                | (Constant i) :: t -> (Constant (i - 1)) :: t
                | h :: t -> h :: (decr t)
                | [] -> [Constant (-1)]
              in
              Sum (decr ts)
          in
          GEZ nv'
      | Not s -> Not (sent_itf s)
      | And (s1, s2) -> And (sent_itf s1, sent_itf s2)
      | Or (s1, s2) -> Or (sent_itf s1, sent_itf s2)
      | x -> x
    and read_itf = function
      | FnCall (fnm, args) -> FnCall (fnm, L.map value_itf args)
      | x -> x
    and pred_itf = function
      | PPred (pnm, args) -> PPred (pnm, L.map value_itf args)
      | x -> x
    and array_itf (upd, init) =
      (L.map (fun (i, v) -> (i, value_itf v)) upd, init)
    and value_itf = function
      | Bool s -> Bool (sent_itf s)
      | Read r -> Read (read_itf r)
      | Array a -> Array (array_itf a)
      | x -> x
    in
    sent_itf expr
  in

  (* need to push int-as-real behavior into DP: new type, rint *)
  exact := true;

  (* strip away Foralls, sent_of negation of what's left *)
  let vars = ref SSet.empty in
  let rec transform = function
    | V.Forall (typ, vnm, expr) -> begin
        let expr' = transform expr in
        if SSet.mem vnm !vars then
        Exists ((if not !exact && typ = T.Int then T.Float else typ), 
                vnm, expr')
        else expr'
    end
    | x ->
        let tr = sent_of (V.Not x) in
        let tr = match required () with
        | True -> tr
        | x -> And (tr, x)
        in
        let tr = if not !exact then int_to_float tr else tr in
        (* record variables *)
        vars := variables (Bool tr);
        L.fold_left 
          (fun s (typ, vnm) ->
            Exists ((if not !exact && typ = T.Int then T.Float else typ),
                    vnm, s))
          tr (fresh_vars ())
  in
  Bool (transform expr)
