import java.io.*;
import java.util.*;

public class Pi implements Runnable {

    final static int PI_COMPILE = 1;
    final static int PI_EVAL = 2;

    int action;

    ArrayList listeners = new ArrayList();

    boolean proveVCs = true;
    boolean generateRuntimeAsserts = false;
    boolean generateInvariants = false;

    HashSet functions = new HashSet();
    HashMap f2vcns = new HashMap();
    HashMap vcn2vc = new HashMap();
    HashMap vcn2v = new HashMap();

    String piexec, cvcexec, cvctimeout;
    Process piProc;

    PiDebugger debugger;

    private volatile Thread piThread = null;

    public Pi(String piexec, String cvcexec, 
	      String cvctimeout) {
	this.piexec = piexec;
	this.cvcexec = cvcexec;
	this.cvctimeout = cvctimeout;
    }

    public void setDebugger(PiDebugger d) {
	debugger = d;
	d.setPi(this);
    }

    public void addListener(PiListener l) {
	listeners.add(l);
    }

    public void firePiOutputLine(String s) {
	for (Iterator it = listeners.iterator(); it.hasNext(); )
	    ((PiListener) it.next()).piOutputLine(s);
    }

    public void firePiNewVC(String fname, String vcname, boolean valid) {
	for (Iterator it = listeners.iterator(); it.hasNext(); )
	    ((PiListener) it.next()).piNewVC(fname, vcname, valid);
    }

    public void firePiCompiled(boolean correct) {
	for (Iterator it = listeners.iterator(); it.hasNext(); )
	    ((PiListener) it.next()).piCompiled(correct);
    }

    public void firePiReturnValue(String rv) {
	for (Iterator it = listeners.iterator(); it.hasNext(); )
	    ((PiListener) it.next()).piReturnValue(rv);
    }
	
    public void firePiError(String rv) {
	for (Iterator it = listeners.iterator(); it.hasNext(); )
	    ((PiListener) it.next()).piError(rv);
    }

    public void firePiLines(Integer[] lines) {
	for (Iterator it = listeners.iterator(); it.hasNext(); )
	    ((PiListener) it.next()).piLines(lines);
    }

    public void setProve(boolean pr) {
	proveVCs = pr;
    }

    public void setRuntimeAssertions(boolean ra) {
	generateRuntimeAsserts = ra;
    }

    public void setGenerateInvariants(boolean ig) {
	generateInvariants = ig;
    }

    public void start() {
	if (piThread == null) {
	    piThread = new Thread(this, "Pi");
	    piThread.start();
	}
    }

    public void run() {
	switch (action) {
	case PI_COMPILE: 
	    run_compile();
	    break;
	case PI_EVAL:
	    run_eval();
	    break;
	}
    }

    private boolean isStart(String ln, String e) {
	return (ln.indexOf("<" + e + ">") != -1);
    }
    private boolean isEnd(String ln, String e) {
	return (ln.indexOf("</" + e + ">") != -1);
    }

    private void newFunction(String fname) {
	synchronized (functions) {
	    functions.add(fname);
	}
    }

    private void newVCName(String fname, String vcname) {
	synchronized (f2vcns) {
	    HashSet vcns = (HashSet) f2vcns.get(fname);
	    if (vcns == null) vcns = new HashSet();
	    vcns.add(vcname);
	    f2vcns.put(fname, vcns);
	}
    }

    final static int INVALID = 0;
    final static int PENDING = 1;
    final static int VALID = 2;

    private void newVC(String fname, String vcname, String spath, 
		       String vc, int status) {
	synchronized (vcn2vc) {
	    vc = spath 
		+ "\n" 
		+ vc 
		+ "\n\n"
		+ (status == 2 ? "[VALID]" : 
		   status == 0 ? "[NOT VALID]" :
		   "[PENDING]");
	    vcn2vc.put(vcname, vc);
	    vcn2v.put(vcname, new Integer (status));
	}
	firePiNewVC(fname, vcname, status == 2);
    }

    private void run_compile() {
	Thread myThread = Thread.currentThread();

	BufferedReader in = 
	    new BufferedReader(new InputStreamReader(piProc.getInputStream()));

	int state = 0;
	boolean correct = false, validating = false;
	String last = "";
	try {
	    String fname = "", vcname = "", vc = "", spath = "";
	    while (piThread == myThread && last != null) {
		last = in.readLine();

		if (last == null) {
		    if (state != 6) continue;
		}
		else firePiOutputLine(last);

		switch (state) {
		case 0:
		    if (last.indexOf("<function>") != -1) state = 1;
		    break;

		case 1:
		    if (last.indexOf("</function>") != -1) state = 2;
		    else {
			newFunction(last);
			fname = last;
		    }
		    break;

		case 2:
		    if (last.indexOf("<function>") != -1) state = 1;
		    else if (last.indexOf("<vcname>") != -1) state = 3;
		    else if (last.indexOf("<PROGRAM VALID>") != -1) {
			correct = true;
			state = 7;
		    }
		    else if (last.indexOf("<PROGRAM INVALID>") != -1) {
			correct = false;
			state = 7;
		    }
		    break;

		case 3:
		    if (last.indexOf("</vcname>") != -1) state = 8;
		    else {
			vcname = last.trim();
			newVCName(fname, vcname);
		    }
		    break;

		case 8:
		    if (isStart(last, "path")) {
			state = 9;
			spath = "";
		    }
		    break;

		case 9:
		    if (isEnd(last, "path")) state = 4;
		    else spath = spath + last.trim() + "\n";
		    break;

		case 4:
		    if (last.indexOf("<vc>") != -1) {
			validating = true;
			state = 5;
			vc = "";
		    }
		    break;

		case 5:
		    if (last.indexOf("</vc>") != -1) {
			/* announce early */
			newVC(fname, vcname, spath, vc, PENDING);
			state = 6;
		    }
		    else vc += last + "\n";
		    break;

		case 6:
		    if (last == null) {
			newVC(fname, vcname, spath, vc, INVALID);
			state = 0;			
		    }
		    else if (last.indexOf("<function>") != -1) {
			newVC(fname, vcname, spath, vc, INVALID);
			state = 1;
		    }
		    else if (last.indexOf("<vcname>") != -1) {
			newVC(fname, vcname, spath, vc, INVALID);
			state = 4;
		    }
		    else if (last.indexOf("<VALID>") != -1) {
			newVC(fname, vcname, spath, vc, VALID);
			state = 2;
		    }
		    else if (last.indexOf("<NOT VALID>") != -1) {
			newVC(fname, vcname, spath, vc, INVALID);
			state = 2;
		    }
		    break;

		case 7:
		}
	    }
	    if (last != null) correct = false;
	} 
	catch (Exception e) {
	    if (last != null) correct = false;
	    e.printStackTrace(System.out);
	}

	try {
	    in.close();
	} 
	catch (IOException e) {
	    e.printStackTrace(System.out);
	}

	if (piThread != null) piProc.destroy();
	piProc = null;
	if (validating && state == 7) firePiCompiled(correct);
	piThread = null;

	if (state != 7) System.out.println("Probable pvc failure.");
    }

    private void debugNew(String fn, String line) {
	debugger.newStep(fn.trim(), Integer.parseInt(line.trim()));
    }

    private void debugVV(String var, String val) {
	debugger.setValue(var.trim(), val.trim());
    }

    private void debugRV(String rv) {
	debugger.setRV(rv.trim());
    }

    private void debugError(String msg) {
	debugger.setError(msg);
    }

    private void run_eval() {
	Thread myThread = Thread.currentThread();

	BufferedReader in = 
	    new BufferedReader(new InputStreamReader(piProc.getInputStream()));

	String last = "";
	try {
	    String fname = "", var = "", val = "";
	    int state = 0;

	    while (piThread == myThread && last != null) {
		last = in.readLine();

		if (last == null) continue;
		else firePiOutputLine(last);

		switch (state) {
		case 0:
		    if (isStart(last, "function")) state = 1;
		    else if (isStart(last, "error")) state = 9;
		    break;

		case 1:
		    if (isEnd(last, "function")) state = 2;
		    else fname = last;
		    break;

		case 2:
		    if (isStart(last, "line")) state = 3;
		    break;

		case 3:
		    if (isEnd(last, "line")) state = 4;
		    else debugNew(fname, last);
		    break;

		case 4:
		    if (isStart(last, "var")) state = 5;
		    else if (isStart(last, "function")) state = 1;
		    else if (isStart(last, "rv")) state = 8;
		    else if (isStart(last, "error")) state = 9;
		    break;

		case 5:
		    if (isEnd(last, "var")) state = 6;
		    else var = last;
		    break;

		case 6:
		    if (isStart(last, "value")) {
			val = "";
			state = 7;
		    }
		    break;

		case 7:
		    if (isEnd(last, "value")) {
			debugVV(var, val);
			state = 4;
		    }
		    else val += (val == "" ? "" : "\n") + last;
		    break;

		case 8:
		    if (isEnd(last, "rv")) state = 0;
		    else debugRV(last);
		    break;

		case 9:
		    if (isEnd(last, "error")) state = 0;
		    else debugError(last);
		    break;
		}
	    }
	} 
	catch (Exception e) {
	    e.printStackTrace(System.out);
	}

	try {
	    in.close();
	} 
	catch (IOException e) {
	    e.printStackTrace(System.out);
	}

	if (piThread != null) piProc.destroy();
	piProc = null;
	piThread = null;
    }

    public boolean compile(String code, String path) {
	if (piProc != null) return false;

	functions.clear();
	f2vcns.clear();
	vcn2vc.clear();
	vcn2v.clear();

	String cmd = 
	    piexec
	    + " -vc " 
	    + (generateRuntimeAsserts ? "-rta " : "")
	    + (proveVCs ? "-prove " : "")
	    + cvcexec
	    + " -dir " + path
	    + " -timeout " + cvctimeout;

	System.out.println(cmd);

	try {
	    piProc = Runtime.getRuntime().exec(cmd);
	    BufferedWriter out = 
		new BufferedWriter(
                    new OutputStreamWriter(piProc.getOutputStream()));
	    out.write(code);
	    out.close();
	}
	catch (IOException e) {
	    e.printStackTrace(System.out);
	    return false;
	}

	action = PI_COMPILE;
	start();

	return true;
    }

    public boolean eval(String code, String expr, String dir) {
	if (piProc != null) return false;

	String cmd =
	    piexec
	    + " -rta"
	    + " -eval " + expr
	    + " -dir " + dir;

	try {
	    piProc = Runtime.getRuntime().exec(cmd);
	    BufferedWriter out = 
		new BufferedWriter(
                    new OutputStreamWriter(piProc.getOutputStream()));
	    out.write(code);
	    out.close();
	}
	catch (IOException e) {
	    e.printStackTrace(System.out);
	    return false;
	}

	debugger.reset();
	action = PI_EVAL;
	start();

	return true;
    }	    

    public void stop() {
	if (piProc != null) piProc.destroy();
	try {
	    Runtime.getRuntime().exec("killall cvcl");
	} 
	catch (IOException e) {
	    e.printStackTrace(System.out);
	}
	piThread = null;
    }

    public Collection getFunctions() {
	synchronized (functions) {
	    return (Collection) functions.clone();
	}
    }

    public Collection getVCNames(String fname) {
	synchronized (f2vcns) {
	    Collection vcns = (Collection) f2vcns.get(fname);
	    if (vcns == null) {
		return new HashSet();
	    }
	    else {
		return vcns;
	    }
	}
    }

    public String getVC(String vcname) {
	synchronized (vcn2vc) {
	    String vc = (String) vcn2vc.get(vcname);
	    if (vc == null) return "";
	    else return vc;
	}
    }

    public boolean getValid(String vcname) {
	synchronized(vcn2vc) {
	    Integer b = (Integer) vcn2v.get(vcname);
	    if (b == null) return false;
	    else return (b.intValue() == VALID);
	}
    }

    public int getStatus(String vcname) {
	synchronized(vcn2vc) {
	    Integer b = (Integer) vcn2v.get(vcname);
	    if (b == null) return PENDING;
	    else return b.intValue();
	}
    }
}
