(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Expr
open Interface

module T = Type

module A = Array
module B = Buffer
module L = List
module H = Hashtbl

exception Error of string

exception AssertionFailure of string
exception RuntimeError of string

exception EarlyExit of bool

(*
 * Values of expressions.
 *)
type value =
  | BoolV of bool
  | IntV of int
  | FloatV of float
  (* bool value is copy-on-write flag *)
  | StructV of bool * (string, value option) H.t
  | ArrayV of bool * value option array

(*
 * Set copy-on-write flag.
 *)
let mkcow = function
  | StructV (_, flds) -> StructV (true, flds)
  | ArrayV (_, arr) -> ArrayV (true, arr)
  | x -> x

let string_of_value v = 
  let rec sov = function
    | BoolV b -> if b then "true" else "false"
    | IntV i -> string_of_int i
    | FloatV f -> string_of_float f
    | StructV (_, flds) ->
	let sflds = ref [] in
	H.iter (fun k v -> sflds := (k ^ ": " ^ (sovo v)) :: !sflds) flds;
	"{ " ^ (String.concat ",\n  " (L.rev !sflds)) ^ "\n}"
    | ArrayV (_, arr) ->
	let selems = ref [] in
	A.iter (fun v -> selems := (sovo v) :: !selems) arr;
	"[" ^ (String.concat "; " (L.rev !selems)) ^ "]"
  and sovo = function
    | None -> "UNINIT"
    | Some v -> sov v
  in
  sov v

(*
 * Structural equality between two values.  Cyclic structures are
 * impossible.
 *)
let rec equalV v1 v2 =
  match v1, v2 with
  | BoolV b1, BoolV b2 -> b1 = b2
  | IntV i1, IntV i2 -> i1 = i2
  | FloatV f1, FloatV f2 -> f1 = f2
  | StructV (_, s1), StructV (_, s2) -> begin
      try
	H.iter
	  (fun k v -> 
	    match v, H.find s2 k with
	    | Some v1, Some v2 ->
		if not (equalV v1 v2) then raise (EarlyExit false)
	    | _ -> ())
	  s1;
	true
      with
      | EarlyExit b -> b
  end
  | ArrayV (_, a1), ArrayV (_, a2) -> begin
      let eq () =
	try
	  A.iteri
	    (fun i v ->
	      match v, a2.(i) with
	      | Some v1, Some v2 ->
		  if not (equalV v1 v2) then raise (EarlyExit false)
	      | _ -> ())
	    a1;
	  true
	with
	| EarlyExit b -> b
      in
      A.length a1 = A.length a2 && eq ()
  end
  | _ -> assert false

exception RV of value

(*
 * Location in memory.
 *)
type location =
  | ScopeVar of string
  | ArrayLoc of value option array * int
  | StructLoc of (string, value option) H.t * string

let new_loc typ =
  let rec nl = function
    | T.Struct (_, flds) ->
	let sloc = H.create 13 in
	L.iter (fun (typ, fld) -> H.add sloc fld (nl typ)) flds;
	Some (StructV (false, sloc))
    | _ -> None
  in
  nl (T.expand typ)

let new_scope fn =
  let sc = H.create 13 in
  L.iter
    (fun (typ, vnm) -> H.add sc vnm (new_loc typ))
    ((parameters fn) @ (locals fn));
  sc

let soe x = string_of_expr x

let print fnm cfg sc =
  let buf = B.create 1024 in
  let hdr =
    "<function>\n" ^ fnm ^ "\n</function>\n"
    ^ "<line>\n" ^ (string_of_int (linenum cfg)) ^ "\n</line>\n"
  in
  B.add_string buf hdr;
  H.iter
    (fun vnm v ->
      B.add_string buf ("<var>\n" ^ vnm ^ "\n</var>\n");
      let pv v = B.add_string buf ("<value>\n" ^ v ^ "\n</value>\n") in
      match v with
      | None -> pv "UNINIT"
      | Some v -> pv (string_of_value v))
    sc;
  print_endline (B.contents buf)

(*
 * Returns the location of an lval, ready for assignment.
 *)
let rec lval prgm sc = function
  | Var s -> ScopeVar s
  | ArrayAcc (arr, i) as x -> begin
      match cow prgm sc arr with
      | ArrayV (_, arr) -> begin
	  match eval prgm sc i with
	  | IntV i ->
	      if i < 0 or i >= A.length arr then
	      raise (RuntimeError ("Array out of bounds: " ^ (soe x))) else
	      ArrayLoc (arr, i)
	  | _ -> assert false
      end
      | _ -> assert false
  end
  | StructAcc (st, fld) as x -> begin
      match cow prgm sc st with
      | StructV (_, flds) -> StructLoc (flds, fld)
      | _ -> assert false
  end
  | _ -> assert false

(*
 * Called by lval, it copies a value if it's about to be
 * written to.
 *)
and cow prgm sc expr =
  let copy = function
    | None -> None
    | Some v ->
	let v' = 
	  match v with
	  | StructV (false, flds) -> StructV (true, flds)
	  | ArrayV (false, arr) -> ArrayV (true, arr)
	  | x -> x
	in
	Some v'
  in
  let cow = function
    | None -> raise (RuntimeError ("Uninitialized: " ^ (soe expr)))
    | Some v -> begin
	match v with
	| StructV (true, flds) -> 
	    let cflds = H.copy flds in
	    H.iter (fun k v -> H.replace cflds k (copy v)) cflds;
	    (true, StructV (false, cflds))
	| ArrayV (true, arr) -> 
	    let carr = A.make (A.length arr) None in
	    A.iteri (fun i v -> carr.(i) <- copy v) arr;
	    (true, ArrayV (false, carr))
	| x -> (false, x)
    end
  in
  match lval prgm sc expr with
  | ScopeVar s ->
      let chng, v = cow (H.find sc s) in
      if chng then H.replace sc s (Some v);
      v
  | ArrayLoc (arr, i) ->
      let chng, v = cow arr.(i) in
      if chng then arr.(i) <- Some v;
      v
  | StructLoc (flds, fld) ->
      let chng, v = cow (H.find flds fld) in
      if chng then H.replace flds fld (Some v);
      v

(*
 * Evaluates an Expr.t, given the prgm, the scope, and the expr.
 *)
and eval prgm sc expr =
  let rec eval sign = function
    | Bool b -> BoolV b
    | Int i -> IntV i
    | Float f -> FloatV f
    | Uninit -> raise (Error "uninit")
    | CArray elems -> ArrayV (false, A.map (fun e -> Some (eval sign e)) elems)
    | CStruct flds ->
	let flds' = H.create 13 in
	L.iter 
	  (fun (fld, v) -> 
	    match v with
	    | Uninit -> H.add flds' fld None
	    | v -> H.add flds' fld (Some (eval sign v)))
	  flds;
	StructV (false, flds')
    | Var s as x -> begin
	try 
	  match H.find sc s with
	  | None -> raise (RuntimeError ("Uninitialized: " ^ (soe x)))
	  | Some v -> v
	with
	| Not_found ->
	    let _, expr = define_of_name prgm s in
	    eval sign expr
    end
    | ArrayAlloc (typ, sz) as x -> begin
	match eval sign sz with
	| IntV sz ->
	    if sz < 0 then 
	    raise (RuntimeError ("Negative array length: " ^ (soe x)));
	    let arr = A.make sz None in
	    for i = 0 to sz - 1 do
	      arr.(i) <- new_loc typ
	    done;
	    ArrayV (false, arr)
	| _ -> assert false
    end
    | ArrayAcc (arr, i) as x -> begin
	match eval sign arr with
	| ArrayV (_, arr) -> begin
	    match eval sign i with
	    | IntV i -> begin
		if i < 0 or i >= A.length arr then
		raise (RuntimeError ("Array out of bounds: " ^ (soe x))) else
                match arr.(i) with
		| None -> raise (RuntimeError ("Uninitialized: " ^ (soe x)))
		| Some v -> v
	    end
	    | _ -> assert false
	end
	| _ -> assert false
    end
    | Length arr -> begin
	match eval sign arr with
	| ArrayV (_, arr) -> IntV (A.length arr)
	| _ -> assert false
    end
    | StructAcc (st, fld) as x -> begin
	match eval sign st with
	| StructV (_, flds) -> begin
	    match H.find flds fld with
	    | None -> raise (RuntimeError ("Uninitialized: " ^ (soe x)))
	    | Some v -> v
	end
	| _ -> assert false
    end
    | FnCall (fnm, args) as x -> begin
        (* interpreted predicates will generate error *)
	try
	  run prgm fnm (L.map (fun arg -> eval sign arg) args)
	with
	| Interface.Error _
        | Error "uninit" -> BoolV sign
    end
    | Collect _ -> BoolV true
    | Plus (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> IntV (i1 + i2)
	| FloatV f1, FloatV f2 -> FloatV (f1 +. f2)
	| _ -> assert false
    end
    | Minus (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> IntV (i1 - i2)
	| FloatV f1, FloatV f2 -> FloatV (f1 -. f2)
	| _ -> assert false
    end
    | Times (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> IntV (i1 * i2)
	| FloatV f1, FloatV f2 -> FloatV (f1 *. f2)
	| _ -> assert false
    end
    | Div (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| FloatV f1, FloatV f2 -> FloatV (f1 /. f2)
	| _ -> assert false
    end
    | IDiv (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> IntV (i1 / i2)
	| _ -> assert false
    end
    | Mod (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> IntV (i1 mod i2)
	| _ -> assert false
    end
    | UMinus e -> begin
	match eval sign e with
	| IntV i -> IntV (-i)
	| FloatV f -> FloatV (-1. *. f)
	| _ -> assert false
    end
    | LT (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> BoolV (i1 < i2)
	| FloatV f1, FloatV f2 -> BoolV (f1 < f2)
	| _ -> assert false
    end
    | LE (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> BoolV (i1 <= i2)
	| FloatV f1, FloatV f2 -> BoolV (f1 <= f2)
	| _ -> assert false
    end
    | NE (e1, e2) -> BoolV (not (equalV (eval sign e1) (eval sign e2)))
    | EQ (e1, e2) -> BoolV (equalV (eval sign e1) (eval sign e2))
    | GE (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> BoolV (i1 >= i2)
	| FloatV f1, FloatV f2 -> BoolV (f1 >= f2)
	| _ -> assert false
    end
    | GT (e1, e2) -> begin
	match eval sign e1, eval sign e2 with
	| IntV i1, IntV i2 -> BoolV (i1 > i2)
	| FloatV f1, FloatV f2 -> BoolV (f1 > f2)
	| _ -> assert false
    end
    | BExpr be -> eval sign be
    | And (e1, e2) -> begin
	match eval sign e1 with
	| BoolV true -> eval sign e2
	| BoolV false as x -> x
	| _ -> assert false
    end
    | Or (e1, e2) -> begin
	match eval sign e1 with
	| BoolV false -> eval sign e2
	| BoolV true as x -> x
	| _ -> assert false
    end
    | Not e -> begin
	match eval (not sign) e with
	| BoolV b -> BoolV (not b)
	| _ -> assert false
    end
  in
  eval true expr

(*
 * Assignment.
 *)
and assign prgm sc lv rv =
  let lv = lval prgm sc lv in
  let rv = match rv with
  | Uninit -> None
  | rv -> Some (mkcow (eval prgm sc rv))
  in
  match lv with
  | ScopeVar vnm -> H.replace sc vnm rv
  | ArrayLoc (arr, i) -> arr.(i) <- rv
  | StructLoc (st, fld) -> H.replace st fld rv

(*
 * Runs the program.
 *)
and run prgm fnm args =
  let fn = function_of_name prgm fnm in
  let sc = new_scope fn in
  (* initialize parameters *)
  L.iter2
    (fun (ptyp, pnm) arg -> 
      H.replace sc pnm (Some (mkcow arg)))
    (parameters fn) args;
  (* execute function block *)
  let pp = ref (body fn) in
  try
    while true do
      print fnm !pp sc;
      match statement !pp with
      | Skip next -> pp := next
      | Asgn (asgns, next) ->
	  L.iter (fun (lv, rv) -> assign prgm sc lv rv) asgns;
	  pp := next
      | Branch (grd, tnext, fnext) -> begin
	  match eval prgm sc grd with
	  | BoolV b -> pp := if b then tnext else fnext
	  | _ -> assert false
      end
      | Assert (asrt, next) -> begin
          (* TODO: Implement interpreted predicates.  try/with temp. *)
	  match eval prgm sc asrt with
	  | BoolV b ->
	      if not b then
	      raise (AssertionFailure ((string_of_expr asrt) 
				       ^ " near " 
				       ^ (string_of_int (linenum !pp))));
	      pp := next
	  | _ -> assert false
      end
      | Return rv -> raise (RV (eval prgm sc rv))
    done;
    (* never reached *)
    IntV (-1)
  with
  | RV rv -> rv

(*
 * Executes the given cmd, which should be a function call,
 * over prgm.
 *)
let execute prgm cmd =
  match Parsr.main Lexer.norm (Lexing.from_string cmd) with
  | Pars.Program _ -> raise (Lexc.SemanticError "Command expected.")
  | Pars.Command (FnCall (fnm, args)) -> begin
      let fn = function_of_name prgm fnm in
      let emptysc = H.create 0 in
      let argv (typ, _) arg = 
	assert_typ typ arg;
	eval prgm emptysc arg 
      in
      run prgm fnm (L.map2 argv (parameters fn) args)
  end
  | _ -> raise (Lexc.SemanticError "Function call expected.")
