(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module E = Expr
module T = Type

module H = Hashtbl
module L = List

let autoasserts = ref false

type id = int

type cfg =
  | Asgn of (E.t * E.t) list * id ref
  | Branch of E.t * id ref * id ref
  | Assert of E.t * id ref
  | Return of E.t option
  | Continue
  | Break

let fns = ref None

let function_source f = fns := Some f

let fninfo fnm =
  match !fns with
  | None -> assert false
  | Some fns -> fns fnm

let ids = ref (-1)
let cfgmap : (id, cfg) H.t = H.create 53
let cutpoints = ref []

let register_cutpoint cfg = cutpoints := cfg :: !cutpoints
let get_cutpoints () = !cutpoints

let linenummap : (id, int) H.t = H.create 53

let cfg_of_id id = H.find cfgmap id

let linenum_of_id id = H.find linenummap id

let id_of_cfg cfg linenum = 
  incr ids;
  let id = !ids in
  H.add cfgmap id cfg;
  H.add linenummap id linenum;
  id

let swap_cfg id cfg =
  try
    let old = H.find cfgmap id in
    H.replace cfgmap id cfg;
  with
  | Not_found -> H.add cfgmap id cfg

let dead = id_of_cfg (Asgn ([], ref 0)) (-1)

let next cfg =
  let next = function
  | Asgn (_, n) -> [!n]
  | Branch (_, n1, n2) -> [!n1; !n2]
  | Assert (_, n) -> [!n]
  | _ -> []
  in
  L.filter (fun cfg -> cfg != dead) (next cfg)

type blk = { 
    (* mutable to allow change in name of vars *)
    mutable vars : (Type.typ * string) list;
    cfg : id;
    dangling : id ref list;
  }

let linenum blk = string_of_int (H.find linenummap blk.cfg)
let linenum_id id = string_of_int (H.find linenummap id) 
	
module CSet = Set.Make (
  struct
    type t = id
    let compare t1 t2 = compare t1 t2
  end)

let string_of_cfg id =
  match cfg_of_id id with
  | Asgn (asgns, next) ->
      (String.concat " "
	 (L.map 
	    (fun (lv, rv) -> 
	      (E.string_of_expr lv) ^ " = " ^ (E.string_of_expr rv) ^ ";")
	    asgns))
  | Branch (grd, next1, next2) ->
      "if (" ^ (E.string_of_expr grd) ^ ") goto L" 
      ^ (string_of_int !next1) ^ " else goto L" 
      ^ (string_of_int !next2) ^ ";\n"
  | Assert (asrt, next) ->
      "assert(" ^ (E.string_of_expr asrt) ^ ");"
  | Return (Some rv) ->
      "return (" ^ (E.string_of_expr rv) ^ ");\n"
  | Return None -> "return;\n"
  | _ -> assert false


let rstring_of_cfg id =
  let seen = ref CSet.empty in
  let rec soc id =
    if CSet.mem id !seen then "" else begin
      seen := CSet.add id !seen;
      "L" ^ (string_of_int id) ^ ":"
      ^ (if id < 10 then "    " else if id < 100 then "   " else "  ")
      ^ match cfg_of_id id with
      | Asgn (asgns, next) -> 
	  (string_of_cfg id) ^ " goto L" ^ (string_of_int !next) 
	  ^ ";\n" ^ (soc !next)
      | Branch (grd, next1, next2) ->
	  (string_of_cfg id) ^ (soc !next1) ^ (soc !next2)
      | Assert (asrt, next) -> 
	  (string_of_cfg id) ^ " goto L" ^ (string_of_int !next) 
	  ^ ";\n" ^ (soc !next)
      | _ -> string_of_cfg id
    end
  in
  soc id

let string_of_blk blk =
  let svars = 
    "    " ^ (String.concat "\n    " 
		(L.map 
		   (fun (typ, v) -> (T.string_of_typ typ) ^ " " ^ v ^ ";") 
		   blk.vars))
    ^ "\n"
  in
  let scfg = rstring_of_cfg blk.cfg in
  "{\n" ^ svars ^ scfg ^ "}"

let empty_stmt () =
  let out = ref dead in
  { vars = [];
    cfg = id_of_cfg (Asgn ([], out)) (Global.get_linenum ());
    dangling = [out];
  }

let empty_blk () = empty_stmt ()

(* for now *)
let scope vars blk =
  blk.vars <- vars @ blk.vars

(* for now *)
let combine_vars s1 s2 = s1.vars @ s2.vars

let blk vars s = scope vars s; s

let concat_stmts s1 s2 =
  let empty stmt =
    stmt.vars = [] &&
    match cfg_of_id stmt.cfg with
    | Asgn ([], _) -> true
    | _ -> false
  in
  if empty s1 then s2 else
  if empty s2 then s1 else begin
    L.iter (fun d -> d := s2.cfg) s1.dangling;
    { vars = combine_vars s1 s2;
      cfg = s1.cfg;
      dangling = s2.dangling;
    }
  end

let rec auto_asserts expr stmt =
  let stmts = ref stmt in
  let add_asrt asrt =
    let assert_stmt asrt =
      let out = ref dead in
      { vars = [];
	cfg = id_of_cfg (Assert (asrt, out)) (Global.get_linenum ());
	dangling = [out];
      }
    in
    stmts := concat_stmts (assert_stmt asrt) !stmts
  in
  let rec aa = function
    | E.ArrayAcc (arr, i) ->
        if !autoasserts then
	add_asrt (E.band
		    (E.ge i (E.const_int 0))
		    (E.lt i (E.length arr)))
    | E.Div  (e1, e2) -> 
	if !autoasserts then
	add_asrt (E.ne e2 (E.const_float 0.))
    | E.IDiv (e1, e2) -> 
	if !autoasserts then
	add_asrt (E.ne e2 (E.const_int 0))
    | E.Mod (e1, e2) -> 
	if !autoasserts then
	add_asrt (E.ne e2 (E.const_int 0))
    | E.ArrayAlloc (_, e) -> 
	if !autoasserts then
	add_asrt (E.ge e (E.const_int 0))
    | E.Length e
    | E.StructAcc (e, _)
    | E.UMinus e
    | E.BExpr e
    | E.Not e -> aa e
    | E.Plus (e1, e2)
    | E.Minus (e1, e2)
    | E.Times (e1, e2)
    | E.LT (e1, e2)
    | E.LE (e1, e2)
    | E.NE (e1, e2)
    | E.EQ (e1, e2)
    | E.GE (e1, e2)
    | E.GT (e1, e2)
    | E.And (e1, e2)
    | E.Or (e1, e2) -> aa e1; aa e2
    | E.FnCall (fnm, args) -> L.iter aa args
    | _ -> ()
  in
  aa expr;
  !stmts
       
let assert_stmt asrt =
  let out = ref dead in
  { vars = [];
    cfg = id_of_cfg (Assert (asrt, out)) (Global.saved ());
    dangling = [out];
  }

let assign_stmt lval rval =
  let out = ref dead in
  auto_asserts lval
    (auto_asserts rval
       { vars = [];
	 cfg = id_of_cfg (Asgn ([(lval, rval)], out)) (Global.get_linenum ());
	 dangling = [out];
       })

let if_stmt grd th els =
  auto_asserts grd
    { vars = combine_vars th els;
      cfg = id_of_cfg (Branch (grd, ref th.cfg, ref els.cfg)) (Global.saved ());
      dangling = th.dangling @ els.dangling;
    }

let continue_stmt () =
  { vars = [];
    cfg = id_of_cfg Continue (Global.get_linenum ());
    dangling = [];
  }

let break_stmt () =
  { vars = [];
    cfg = id_of_cfg Break (Global.get_linenum ());
    dangling = [];
  }

let return_stmt = function
  | None -> { vars = [];
	      cfg = id_of_cfg (Return None) (Global.get_linenum ());
	      dangling = [];
	    }
  | Some rv ->
      auto_asserts rv
	{ vars = [];
	  cfg = id_of_cfg (Return (Some rv)) (Global.get_linenum ());
	  dangling = [];
	}

let fix_ctrl entry exit =
  let rec fix seen id =
    if not (CSet.mem id !seen) then begin
      seen := CSet.add id !seen;
      match cfg_of_id id with
      | Continue ->  swap_cfg id (Asgn ([], ref entry))
      | Break -> swap_cfg id (Asgn ([], exit))
      | x -> L.iter (fun x -> fix seen x) (next x)
    end
  in
  fix (ref CSet.empty) entry

(*
 * loopi grd blk loopi ...
 *)
let while_stmt loopi grd blk =
  (* loop exit *)
  let exit = ref dead in
  (* handle loopi first *)
  let loopi_asrt = assert_stmt loopi in
  (* branch to blk' or exit *)
  let stmt = 
    { vars = blk.vars;
      cfg = id_of_cfg (Branch (grd, ref blk.cfg, exit)) (Global.saved ());
      dangling = [exit];
    }
  in
  register_cutpoint stmt.cfg;
  let branch = auto_asserts grd stmt in
  (* entry is the loop invariant into the branch *)
  let loop = concat_stmts loopi_asrt branch in
  (* dangles lead back to entry *)
  L.iter (fun d -> d := loop.cfg) blk.dangling;
  (* fix Continue and Break statements within loop *)
  fix_ctrl loop.cfg exit;
  loop

(*
 * init loopi grd blk incr loopi ...
 *)
let for_stmt loopi init grd incr blk =
  concat_stmts init (while_stmt loopi grd (concat_stmts blk incr))

(*
 * fix Return stmts: one exit
 *)
let mk_function defs fnm typ params pre post blk =
  if blk.dangling <> [] then 
    raise (Lexc.SemanticError 
	     ("Function " ^ fnm ^ " lacks return near " ^ (linenum blk)));
  (* special variable: rv *)
  let rvnm = "rv" in
  let rv = E.var rvnm in
  (* replace Return stmts with goto to out stmt
   *)
  let rec fix seen id out =
    if not (CSet.mem id !seen) then begin
      seen := CSet.add id !seen;
      match cfg_of_id id with
      | Assert (asrt, next) ->
	  fix seen !next out
      | Return None -> swap_cfg id (Asgn ([], ref out))
      | Return (Some rv') -> swap_cfg id (Asgn ([(rv, rv')], ref out))
      | x -> L.iter (fun x -> fix seen x out) (next x)
    end
  in
  let fix blk out =
    fix (ref CSet.empty) blk.cfg out.cfg
  in
  (* post + return; map0 maps v -> v0 *)
  let out map0 ret = 
    let asrt = assert_stmt (E.sub map0 post) in
    concat_stmts asrt ret;
  in
  (* construct new blk with post and single Return statement *)
  let mkblk map0 =
    let out = out map0 (return_stmt (Some rv)) in
    fix blk out;
    blk.vars <- (typ, rvnm) :: blk.vars;
    blk
  in
  (* initial-value holding vars *)
  let p0vars = ref [] in
  let map0 = H.create 13 in
  let preblk =
    (* pre vars should only include params and defines *)
    E.VSet.iter
      (fun v ->
	if not ((L.exists (fun (typ, v') -> v = v') params)
		  or (L.exists (fun d -> v = d) defs)) then
	raise (Lexc.SemanticError 
		 ("Pre of function includes variable " ^ v 
		  ^ "near " ^ (linenum blk))))
      (E.vars pre);
    (* post vars should only include "rv", params, and defines *)
    E.VSet.iter
      (fun v ->
	if not (v = "rv" 
		  or (L.exists (fun (typ, v') -> v = v') params)
		  or (L.exists (fun d -> v = d) defs)) then
	raise (Lexc.SemanticError 
		 ("Post of function includes variable " ^ v 
		  ^ " near " ^ (linenum blk))))
      (E.vars post);
    (* insert assignments v_0 = v *)
    L.fold_left
      (fun stmt (typ, vnm) ->
	(* new variable to hold initial value *)
	let v0 = vnm ^ "_0" in
	H.add map0 vnm (E.var v0);
        (* if v0 does not appear in scope, add it *)
        if not (L.exists (fun (_, vnm) -> vnm = v0) blk.vars) then
	p0vars := (typ, v0) :: !p0vars;
	(* prepend assignment v0 = v *)
	concat_stmts stmt (assign_stmt (E.var v0) (E.var vnm)))
      (empty_stmt ()) params
  in
  (* preblk + blk *)
  let body = concat_stmts preblk (mkblk map0) in
  let body = concat_stmts (assert_stmt pre) body in
  (* add initial-value holders as vars *)
  scope !p0vars body;
  body

let typ blk vnm =
  let typ, _ = 
    try L.find (fun (typ, vnm') -> vnm = vnm') blk.vars with
    | Not_found -> 
	raise (Lexc.SemanticError 
		 ("Variable " ^ vnm ^ " undeclared near " ^ (linenum blk)))
  in
  typ

type quantifer =
  | Empty
  | Forall
  | Exists

let type_check fnm blk =
  (* extra constraints on formation of assertions *)
  let asrtn_form asrt id =
    let rec form q vars = function
      | E.FnCall ("forall", [ix; l; r; phi]) ->
	  if q = Exists then 
	    raise (Lexc.SemanticError 
		     ("Quantifier alternation near " ^ (linenum_id id)));
	  form Forall (ix :: vars) phi
      | E.FnCall ("exists", [ix; l; r; phi]) ->
	  if q = Forall then 
	    raise (Lexc.SemanticError 
		     ("Quantifier alternation near " ^ (linenum_id id)));
	  form Exists (ix :: vars) phi
      | E.BExpr t -> form q vars t
      | E.Not t -> begin
	  match q with
	  | Empty -> form q vars t
	  | Forall -> form Exists vars t
	  | Exists -> form Forall vars t
      end
      | E.And (t1, t2)
      | E.Or (t1, t2) -> form q vars t1; form q vars t2
      | E.Var s as x -> 
	  if L.mem x vars then
	      raise (Lexc.SemanticError 
		       ("Quantified index out of array read near " 
			^ (linenum_id id)))
      | E.ArrayAcc (arr, E.Var s) -> form q vars arr
      | E.ArrayAcc (arr, i) -> form q vars arr; form q vars i
      | E.FnCall (fn, ts) -> L.iter (fun t -> form q vars t) ts
      | E.Plus (t1, t2)
      | E.Minus (t1, t2)
      | E.Times (t1, t2)
      | E.Div (t1, t2)
      | E.IDiv (t1, t2)
      | E.Mod (t1, t2)
      | E.LT (t1, t2)
      | E.LE (t1, t2)
      | E.EQ (t1, t2)
      | E.NE (t1, t2)
      | E.GE (t1, t2)
      | E.GT (t1, t2) -> form q vars t1; form q vars t2
      | E.UMinus t -> form q vars t
      | x -> ()
    in
    form Empty [] asrt
  in
  let seen = ref CSet.empty in
  let rec tc id =
    if not (CSet.mem id !seen) then begin
      seen := CSet.add id !seen;
      let assert_typ typ expr =
	try E.assert_typ typ expr with
	| x -> print_endline (string_of_cfg id); raise x
      in
      match cfg_of_id id with
      | Asgn (asgns, next) ->
	  L.iter (fun (t1, t2) -> assert_typ (E.typ_of t1) t2) asgns;
	  tc !next
      | Branch (grd, next1, next2) ->
	  assert_typ T.Bool grd;
	  tc !next1;
	  tc !next2;
      | Assert (asrt, next) ->
	  assert_typ T.Bool asrt;
	  asrtn_form asrt id;
	  tc !next
      | Return (Some rv) -> begin
	  match fnm with
	  | None -> ()
	  | Some fnm ->
	      let _, _, rtyp, _ = fninfo fnm in
	      assert_typ rtyp rv
      end
      | _ -> ()
    end
  in
  tc blk.cfg
