(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module T = Type

module H = Hashtbl
module L = List

exception Error of string

type t =
  | Bool of bool
  | Int of int
  | Float of float
  | Uninit
  | CArray of t array
  | CStruct of (string * t) list
  | Var of string
  | ArrayAlloc of T.typ * t
  | ArrayAcc of t * t
  | Length of t
  | StructAcc of t * string
  | FnCall of string * t list
  | Collect of string * t list
  | Plus of t * t
  | Minus of t * t
  | Times of t * t
  | Div of t * t
  | IDiv of t * t
  | Mod of t * t
  | UMinus of t
  | LT of t * t
  | LE of t * t
  | NE of t * t
  | EQ of t * t
  | GE of t * t
  | GT of t * t
  | BExpr of t
  | And of t * t
  | Or of t * t
  | Not of t


module VSet = Set.Make (
  struct
    type t = string
    let compare t1 t2 = compare t1 t2
  end)


let typs = ref None

let typ_source ts = typs := Some ts

let typ vnm =
  match !typs with
  | None -> None
  | Some typs -> Some (typs vnm)


let fns = ref None

let function_source fs = fns := Some fs

let fninfo fnm =
  match !fns with
  | None -> raise (Error "function_source unregistered")
  | Some fninfo -> fninfo fnm


let const_true = Bool true

let const_false = Bool false

let const_int n = Int n

let const_float n = Float n


let bxor t1 t2 =
  Or (And (t1, Not t2), And (Not t1, t2))

let biff t1 t2 =
  And (Or (t1, Not t2), Or (Not t1, t2))

let bimplies t1 t2 =
  Or (Not t1, t2)

let band t1 t2 = And (t1, t2)

let bor t1 t2 = Or (t1, t2)

let bnot t = Not t

let bexpr t = BExpr t


let lt t1 t2 = LT (t1, t2)

let le t1 t2 = LE (t1, t2)

let ne t1 t2 = NE (t1, t2)

let eq t1 t2 = EQ (t1, t2)

let ge t1 t2 = GE (t1, t2)

let gt t1 t2 = GT (t1, t2)


let fn_call fn args = FnCall (fn, args)

let plus t1 t2 = Plus (t1, t2)

let minus t1 t2 = Minus (t1, t2)

let times t1 t2 = Times (t1, t2)

let div t1 t2 = Div (t1, t2)

let idiv t1 t2 = IDiv (t1, t2)

let modo t1 t2 = Mod (t1, t2)

let uminus t = UMinus t

let var vnm = Var vnm


let array_alloc typ t = ArrayAlloc (typ, t)

let length t = Length t

let array_access arr i = ArrayAcc (arr, i)

let struct_access str fld = StructAcc (str, fld)


let const_array elems = CArray (Array.of_list elems)

let const_struct flds = CStruct flds

let uninit = Uninit


let collect nm args = Collect (nm, args)


let vars t =
  let vs = ref VSet.empty in
  let rec vars = function
    | Var s -> vs := VSet.add s !vs
    | ArrayAcc (t1, t2)
    | Plus (t1, t2)
    | Minus (t1, t2)
    | Times (t1, t2)
    | Div (t1, t2)
    | IDiv (t1, t2)
    | Mod (t1, t2)
    | LT (t1, t2)
    | LE (t1, t2)
    | NE (t1, t2)
    | EQ (t1, t2)
    | GE (t1, t2)
    | GT (t1, t2)
    | And (t1, t2)
    | Or (t1, t2) -> vars t1; vars t2
    | ArrayAlloc (_, t)
    | Length t
    | StructAcc (t, _)
    | UMinus t
    | BExpr t
    | Not t -> vars t
    | FnCall (_, ts)
    | Collect (_, ts) -> L.iter vars ts
    | _ -> ()
  in
  vars t;
  !vs

let sub vmap t =
  let rec rnm = function
    | Var s as x -> begin
	try H.find vmap s with
	| Not_found -> x
    end
    | ArrayAcc (t1, t2) -> ArrayAcc (rnm t1, rnm t2)
    | Plus (t1, t2) -> Plus (rnm t1, rnm t2)
    | Minus (t1, t2) -> Minus (rnm t1, rnm t2)
    | Times (t1, t2) -> Times (rnm t1, rnm t2)
    | Div (t1, t2) -> Div (rnm t1, rnm t2)
    | IDiv (t1, t2) -> IDiv (rnm t1, rnm t2)
    | Mod (t1, t2) -> Mod (rnm t1, rnm t2)
    | LT (t1, t2) -> LT (rnm t1, rnm t2)
    | LE (t1, t2) -> LE (rnm t1, rnm t2)
    | NE (t1, t2) -> NE (rnm t1, rnm t2)
    | EQ (t1, t2) -> EQ (rnm t1, rnm t2)
    | GE (t1, t2) -> GE (rnm t1, rnm t2)
    | GT (t1, t2) -> GT (rnm t1, rnm t2)
    | And (t1, t2) -> And (rnm t1, rnm t2)
    | Or (t1, t2) -> Or (rnm t1, rnm t2)
    | ArrayAlloc (typ, t) -> ArrayAlloc (typ, rnm t)
    | Length t -> Length (rnm t)
    | StructAcc (t, fld) -> StructAcc (rnm t, fld)
    | UMinus t -> UMinus (rnm t)
    | BExpr t -> BExpr (rnm t)
    | Not t -> Not (rnm t)
    | FnCall (fnm, ts) -> FnCall (fnm, L.map rnm ts)
    | Collect (fnm, ts) -> Collect (fnm, L.map rnm ts)
    | x -> x
  in
  rnm t

let string_of_expr t =
  let rec soe = function
    | Bool b -> if b then "true" else "false"
    | Int n -> string_of_int n
    | Float n -> string_of_float n
    | Uninit -> "?"
    | CArray elems -> 
	let selems = ref [] in
	Array.iter (fun e -> selems := (soe e) :: !selems) elems;
	"[" ^ (String.concat "; " (L.rev !selems)) ^ "]"
    | CStruct flds -> 
	"{" ^ (String.concat ", " 
		 (L.map (fun (fld, v) -> fld ^ " = " ^ (soe v)) flds))
	^ "}"
    | Var v -> v
    | ArrayAlloc (typ, t) -> 
	"new " ^ (T.string_of_typ typ) ^ "[" ^ (soe t) ^ "]"
        (*
	let ts = T.string_of_typ typ in
	"{(" ^ ts ^ " *) malloc(" ^ (psoe t) 
	^ " * sizeof(" ^ ts ^ ")), "
	^ (soe t) ^ "}"
        *)
    | ArrayAcc (t1, t2) -> (soe t1) ^ "[" ^ (soe t2) ^ "]"
    | Length t -> "length(" ^ (soe t) ^ ")"
    | StructAcc (t, fld) -> (soe t) ^ "." ^ fld
    | FnCall (fn, args) -> 
	fn ^ "(" ^ (String.concat ", " (L.map soe args)) ^ ")"
    | Collect (fn, args) -> 
	"~" ^ fn ^ "(" ^ (String.concat ", " (L.map soe args)) ^ ")"
    | Plus (t1, t2) -> (psoe t1) ^ " + " ^ (psoe t2)
    | Minus (t1, t2) -> (psoe t1) ^ " - " ^ (psoe t2)
    | Times (t1, t2) -> (psoe t1) ^ " * " ^ (psoe t2)
    | Div (t1, t2) -> (psoe t1) ^ " / " ^ (psoe t2)
    | IDiv (t1, t2) -> (psoe t1) ^ " div " ^ (psoe t2)
    | Mod (t1, t2) -> (psoe t1) ^ " % " ^ (psoe t2)
    | UMinus t -> "-" ^ (psoe t)
    | LT (t1, t2) -> (soe t1) ^ " < " ^ (soe t2)
    | LE (t1, t2) -> (soe t1) ^ " <= " ^ (soe t2)
    | NE (t1, t2) -> (soe t1) ^ " != " ^ (soe t2)
    | EQ (t1, t2) -> (soe t1) ^ " == " ^ (soe t2)
    | GE (t1, t2) -> (soe t1) ^ " >= " ^ (soe t2)
    | GT (t1, t2) -> (soe t1) ^ " > " ^ (soe t2)
    | BExpr t -> soe t
    | And (t1, t2) -> (psoe t1) ^ " && " ^ (psoe t2)
    | Or (t1, t2) -> (psoe t1) ^ " || " ^ (psoe t2)
    | Not t -> "!" ^ (psoe t)
  and psoe t = 
    match t with
    | Bool _
    | Int _
    | Float _
    | Var _
    | ArrayAcc _
    | StructAcc _
    | Length _ -> soe t
    | _ -> "(" ^ (soe t) ^ ")"
  in
  soe t

let rec typ_of = function
  | Bool _ -> T.Bool
  | Int _ -> T.Int
  | Float _ -> T.Float
  | Uninit
  | CArray _
  | CStruct _ -> 
      raise (Error ("Cannot get typ_of uninit or constant struct or array"))
  | Var s -> begin
      match typ s with
      | None -> raise (Error ("typ_source unregistered"))
      | Some typ -> typ
  end
  | ArrayAlloc (typ, t) -> 
      assert_typ T.Int t;
      T.Array typ
  | ArrayAcc (t1, t2) -> begin
      assert_typ T.Int t2;
      match T.expand (typ_of t1) with
      | T.Array typ -> typ
      | _ -> raise (Lexc.TypeError ("Array expected"))
  end
  | Length t -> T.Int
  | StructAcc (t, fld) -> begin
      match T.expand (typ_of t) with
      | T.Struct (nm, flds) -> begin
	  try
	    let t, _ = L.find (fun (_, f) -> f = fld) flds in
	    t
	  with
	  | Not_found -> raise (Lexc.SemanticError (nm ^ " lacks field " ^ fld))
      end
      | x -> raise (Lexc.TypeError ((string_of_expr t) ^ " is not a struct"))
  end
  | FnCall (fnm, ts) ->
      let ftyp, pts = fninfo fnm in
      L.iter2
	(fun pt arg -> assert_typ pt arg)
	pts ts;
      ftyp
  | Collect _ -> T.Bool
  | Plus (t1, t2)
  | Minus (t1, t2)
  | Times (t1, t2) ->
      let typ = typ_of t1 in
      begin
	match typ with
	| T.Int | T.Float -> ()
	| _ -> raise (Lexc.TypeError "Number expected (+/-/*//)")
      end;
      assert_typ typ t2;
      typ
  | Div (t1, t2) ->
      assert_typ T.Float t1;
      assert_typ T.Float t2;
      T.Float
  | IDiv (t1, t2) -> 
      assert_typ T.Int t1;
      assert_typ T.Int t2;
      T.Int
  | Mod (t1, t2) -> 
      assert_typ T.Int t1;
      assert_typ T.Int t2;
      T.Int
  | UMinus t -> begin
      match typ_of t with
      | T.Int -> T.Int
      | T.Float -> T.Float
      | _ -> raise (Lexc.TypeError "Number expected (-)")
  end
  | LT (t1, t2)
  | LE (t1, t2)
  | GE (t1, t2)
  | GT (t1, t2) ->
      let typ = typ_of t1 in
      begin
	match typ with
	| T.Int | T.Float -> ()
	| x -> raise (Lexc.TypeError "Number expected (</<=/!=/==/>=/>")
      end;
      assert_typ typ t2;
      T.Bool
  | NE (t1, t2)
  | EQ (t1, t2) ->
      let typ = typ_of t1 in
      assert_typ typ t2;
      T.Bool
  | BExpr t ->
      assert_typ T.Bool t;
      T.Bool
  | And (t1, t2)
  | Or (t1, t2) ->
      assert_typ T.Bool t1;
      assert_typ T.Bool t2;
      T.Bool
  | Not t ->
      assert_typ T.Bool t;
      T.Bool
and assert_typ typ expr =
  let terr typ typ' expr =
    raise (Lexc.TypeError 
	     ((string_of_expr expr) ^ " has type " 
	      ^ (T.string_of_typ typ') ^ " but should have type " 
	      ^ (T.string_of_typ typ)))
  in
  match expr with
  | CArray elems -> begin
      match T.expand typ with
      | T.Array typ -> Array.iter (fun e -> assert_typ typ e) elems
      | _ -> terr typ (T.Array (T.UIT "Undetermined")) expr
  end
  | CStruct flds -> begin
      match T.expand typ with
      | T.Struct (nm, tflds) -> begin
          (* Same number of fields. *)
	  if L.length tflds != L.length flds then 
	  terr typ (T.Struct ("Undetermined", [])) expr;
          (* Same fields, and same field types. *)
	  L.iter
	    (fun (typ, fnm) ->
	      try
		let fnm', expr = L.find (fun (fnm', _) -> fnm = fnm') flds in
		match expr with
		| Uninit -> ()
		| expr -> assert_typ typ expr
	      with
	      | Not_found -> 
		  raise (Lexc.TypeError ((string_of_expr expr) 
					 ^ " missing field " ^ fnm)))
	    tflds
      end
      | _ -> terr typ (T.Struct ("Undetermined", [])) expr
  end 
  | Uninit -> ()
  | t ->
      let typ' = typ_of t in
      if T.expand typ <> T.expand typ' then
      terr typ typ' t
