(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module C = Cfg
module E = Expr
module P = Program
module T = Type

module A = Array
module B = Buffer
module L = List
module H = Hashtbl
module Q = Queue
module S = String

exception Error of string

type cfg = int

type stmt = 
  | Skip of cfg
  | Asgn of (Expr.t * Expr.t) list * cfg
  | Branch of Expr.t * cfg * cfg
  | Assert of Expr.t * cfg
  | Return of Expr.t

module CfgSet = Set.Make (
  struct
    type t = cfg
    let compare t1 t2 = compare t1 t2
  end)

type cfg_extra = {
    preds : CfgSet.t ref;
    mutable inv : Expr.t;
  }

type fn = {
    mutable pre : E.t;
    post : E.t;
    rtyp : T.typ;
    fnm : string;
    params : (T.typ * string) list;
    locals : (T.typ * string) list;
    body : cfg;
  }

module FnSet = Set.Make (
  struct
    type t = fn
    let compare t1 t2 = compare t1 t2
  end)

type prgm = {
    p : P.t;
    types : T.TSet.t;
    defines : (T.typ * string * Expr.t) list;
    functions : FnSet.t;
  }

(* --- cfg --- *)

let id cfg = cfg

let linenum cfg = C.linenum_of_id cfg

let maxcfgid = 1024

let cfgs : (stmt * cfg_extra) option array = A.make maxcfgid None

let cutpoints = ref CfgSet.empty

let register_cutpoint cfg = cutpoints := CfgSet.add cfg !cutpoints

let is_cutpoint cfg = CfgSet.mem cfg !cutpoints

let declare_cfg cfg stmt =
  assert (cfg >= 0 && cfg < maxcfgid);
  assert (cfgs.(cfg) = None);
  cfgs.(cfg) <- Some (stmt, { preds = ref CfgSet.empty; inv = E.const_true })

let lookup i =
  assert (i >= 0 && i < maxcfgid);
  cfgs.(i)

let strengthen cfg expr =
  match lookup cfg with
  | None -> raise (Error "strengthen: cfg does not exist")
  | Some (stmt, extra) ->
      extra.inv <-
	if extra.inv = E.const_true then 
	  expr else 
	  E.band extra.inv expr

let finalize () =
  let preds cfg stmt =
    let add succ pred =
      match lookup succ with
      | None -> assert false
      | Some (_, extra) -> extra.preds := CfgSet.add pred !(extra.preds)
    in
    match stmt with
    | Skip next
    | Asgn (_, next)
    | Assert (_, next) -> add next cfg
    | Branch (_, next1, next2) ->
	add next1 cfg;
	add next2 cfg
    | _ -> ()
  in
  A.iteri
    (fun cfg v ->
      match v with
      | None -> ()
      | Some (stmt, _) -> preds cfg stmt)
    cfgs

let statement cfg =
  match lookup cfg with
  | None -> raise (Error "Cfg does not exist.")
  | Some (stmt, _) -> stmt

let predecessors cfg =
  match lookup cfg with
  | None -> raise (Error "Cfg does not exist.")
  | Some (_, extra) -> !(extra.preds)

let successors cfg =
  match statement cfg with
  | Skip next
  | Asgn (_, next)
  | Assert (_, next) -> CfgSet.singleton next
  | Branch (_, next1, next2) ->
      CfgSet.add next1 (CfgSet.singleton next2)
  | _ -> CfgSet.empty

let simple_invariants fns =
  let inv fn =
    (* strengthen each Assert with precondition *)
    let str body lens pre0 =
      (* strengthen first cfg (the precondition) with lens *)
      strengthen body lens;
      fn.pre <- E.band fn.pre lens;
      (* strengthen rest with pre0 *)
      let seen = ref CfgSet.empty in
      let rec str cfg pre =
	if not (CfgSet.mem cfg !seen) then begin
	  seen := CfgSet.add cfg !seen;
	  begin
	    match statement cfg with
	    | Assert _ -> strengthen cfg pre
	    | _ -> ()
	  end;
	  CfgSet.iter (fun succ -> str succ pre) (successors cfg)
	end
      in
      CfgSet.iter (fun succ -> str succ pre0) (successors body)
    in
    (* strengthen precondition with array params having len >= 0 *)
    let avars = ref [] in
    let rec sarray typ v =
      match typ with
      | T.Struct (_, flds) -> 
	  L.iter (fun (ftyp, fld) -> sarray ftyp (v ^ "." ^ fld)) flds
      | T.Array _ -> avars := v :: !avars
      | _ -> ()
    in
    L.iter (fun (typ, v) -> sarray typ v) fn.params;
    let lens =
      L.map (fun v -> E.ge (E.length (E.var v)) (E.const_int 0)) !avars
    in
    let lens = match lens with
    | [] -> E.const_true
    | [x] -> x
    | h :: t -> L.fold_left E.band h t
    in
    (* form precondition, with v -> v_0 *)
    let inv = E.band fn.pre lens in
    let map0 = H.create 13 in
    L.iter
      (fun v -> H.add map0 v (E.var (v ^ "_0")))
      (L.map (fun (_, v) -> v) fn.params);

    (* 09/27/2006: deduce v = v_0 when easy *)
    let eqinv =
      let unmod_params = ref fn.params in
      let mod_arrays = ref [] in
      let remove v =
	let rec rm = function
	  | (typ, v') :: t -> if v' = v then t else (typ, v') :: (rm t)
	  | [] -> []
	in
	unmod_params := rm !unmod_params
      in

      let seen = ref CfgSet.empty in
      let rec eqinv cfg =
	if not (CfgSet.mem cfg !seen) then begin
	  seen := CfgSet.add cfg !seen;
	  begin
	    match statement cfg with
	    | Asgn (asgns, _) ->
		let rec getvar = function
		  | E.Var v -> v
		  | E.ArrayAcc (a, _) -> getvar a
		  | E.StructAcc (a, _) -> getvar a
		  | _ -> assert false
		in
		L.iter 
		  (fun (x, _) -> 
		    remove (getvar x);
		    match x with
		    | E.ArrayAcc (E.Var v, _) -> 
			if not (L.mem v !mod_arrays) 
			    && (L.exists (fun (_, v') -> v' = v) fn.params)
			then
			  mod_arrays := v :: !mod_arrays
		    | _ -> ())
		  asgns
	    | _ -> ()
	  end;
	  CfgSet.iter (fun succ -> eqinv succ) (successors cfg)
	end
      in
      CfgSet.iter (fun succ -> eqinv succ) (successors fn.body);

      let eqs =
	(L.map
	   (fun (typ, v) ->
	     let eq v = E.eq (E.var v) (E.var (v ^ "_0")) in
	     match typ with
	     | T.Array _ -> 
		 E.band 
		   (eq v) 
		   (E.eq (E.length (E.var v)) (E.length (E.var (v ^ "_0"))))
	     | _ -> eq v)
	   !unmod_params)
	  @
	    (L.map
	       (fun v -> 
		 E.eq (E.length (E.var v)) (E.length (E.var (v ^ "_0"))))
	       !mod_arrays)
      in
      match eqs with
      | h :: [] -> h
      | h :: t -> L.fold_left E.band h t
      | [] -> E.const_true
    in
    (* end 09/27/2006 *)

    (* finally *)
    str fn.body lens (E.band (E.sub map0 inv) eqinv)
  in
  FnSet.iter inv fns

(* --- prgm --- *)

module IdSet = Set.Make (
  struct
    type t = C.id
    let compare t1 t2 = compare t1 t2
  end)

let prgm_of_prgm pr =
  let cfg_of_cfg id =
    let seen = ref IdSet.empty in
    let rec cfg_of_cfg id =
      if not (IdSet.mem id !seen) then begin
	seen := IdSet.add id !seen;
	let stmt =
	  match C.cfg_of_id id with
	  | C.Asgn ([], next) -> Skip (cfg_of_cfg !next)
	  | C.Asgn (asgns, next) -> Asgn (asgns, cfg_of_cfg !next)
	  | C.Branch (grd, next1, next2) -> 
	      Branch (grd, cfg_of_cfg !next1, cfg_of_cfg !next2)
	  | C.Assert (asrt, next) -> Assert (asrt, cfg_of_cfg !next)
	  | C.Return (Some rv) -> Return rv
	  | _ -> assert false
	in
	declare_cfg id stmt
      end;
      id
    in
    cfg_of_cfg id
  in
  let fn_of_fn = function
    | (fn, Some blk) -> begin
	{ pre = fn.P.fpre;
	  post = fn.P.fpost;
	  rtyp = fn.P.ftyp;
	  fnm = fn.P.fnm;
	  params = fn.P.fparams;
	  locals = blk.C.vars;
	  body = cfg_of_cfg blk.C.cfg;
	}
    end
    | _ -> assert false
  in
  let fns = ref FnSet.empty in
  H.iter 
    (fun fnm fn -> 
      if not (Interpreted.interpreted fnm) then
      fns := FnSet.add (fn_of_fn fn) !fns)
    pr.P.fns;
  finalize ();
  L.iter (fun cfg -> register_cutpoint cfg) (C.get_cutpoints ());
  simple_invariants !fns;
  { p = pr;
    types = !(pr.P.typs);
    defines = !(pr.P.defs);
    functions = !fns
  }


type parse =
  | Program of prgm
  | Command of Expr.t

let gparse = function
  | Pars.Program pr -> 
      Interpreted.register pr;
      P.type_check pr;
      Program (prgm_of_prgm pr)
  | Pars.Command cmd ->
      Command cmd

let read ch =
  let code = Q.create () in
  begin
    try
      while true do
	Q.add (input_line ch) code
      done
    with
    | End_of_file -> ()
  end;
  code

let read_str str =
  let code = Q.create () in
  L.iter 
    (fun ln -> Q.add ln code)
    (Str.split (Str.regexp "\n") str);
  code

let preprocess code dir =
  let included = ref [] in
  let buf = B.create 1024 in
  while not (Q.is_empty code) do
    let ln = Q.take code in
    if S.length ln > 0 && ln.[0] = '#' then begin
      match Str.split (Str.regexp "[\" \t]+") ln with
      | ["#include"; fl] -> begin
	  let fl = dir ^ "/" ^ fl in
	  if L.mem fl !included then
	    raise (Lexc.ParseError ("Circular inclusion: " ^ fl))
	  else
	    included := fl :: !included;
	  try
	    let ch = open_in fl in
	    Q.transfer (read ch) code;
	    close_in ch
	  with
	  | Sys_error msg -> 
	      raise (Lexc.ParseError 
		       ("Could not open: " ^ fl ^ " (" ^ msg ^ ")"))
      end
      | _ -> raise (Lexc.ParseError ("Bad preprocess command: " ^ ln))
    end else B.add_string buf ln;
    B.add_char buf '\n'
  done;
  B.contents buf

let parse inch dir =
  let s = preprocess (read inch) dir in
  gparse (Parsr.main Lexer.norm (Lexing.from_string s))

let parse_string s dir =
  let s = preprocess (read_str s) dir in
  gparse (Parsr.main Lexer.norm (Lexing.from_string s))

let parse_file fl dir = 
  let inch = open_in fl in
  let prgm = parse inch dir in
  close_in inch;
  prgm
  
let fast_parse inch =
  gparse (Parsr.main Lexer.norm (Lexing.from_channel inch))

let types prgm = prgm.types

let defines prgm = prgm.defines

let define_of_name prgm nm =
  try
    let typ, _, expr = L.find (fun (_, nm', _) -> nm = nm') prgm.defines in
    (typ, expr)
  with
  | Not_found -> raise (Error ("Define " ^ nm ^ " undefined."))

let functions prgm = prgm.functions

let function_of_name prgm fnm =
  let find = FnSet.filter (fun fn -> fn.fnm = fnm) prgm.functions in
  try FnSet.choose find with
  | Not_found -> raise (Error ("Function " ^ fnm ^ " undeclared."))

let string_of_prgm prgm = P.string_of_program prgm.p

(* --- fn --- *)

let name fn = fn.fnm

let return_type fn = fn.rtyp

let parameters fn = fn.params

let return_value fn = (fn.rtyp, "rv")

let precondition fn = fn.pre

let postcondition fn = fn.post

let locals fn = fn.locals

let body fn = fn.body

let type_of_var prgm fn vnm =
  if vnm = "rv" then fn.rtyp else
  try 
    let typ, _ = 
      L.find (fun (_, vnm') -> vnm = vnm') (fn.locals @ fn.params) 
    in
    typ
  with
  | Not_found -> begin
      try
	let typ, _, _ =
	  L.find (fun (_, dnm, _) -> vnm = dnm) prgm.defines
	in
	typ
      with
      | Not_found -> 
	  raise (Error ("Variable " ^ vnm ^ " undeclared in scope " 
			^ fn.fnm ^ "."))
  end

let type_of_expr tvars prgm fn expr =
  let src prgm fn vnm =
    try
      let typ, _ =
	L.find (fun (_, vnm') -> vnm = vnm') tvars
      in
      typ
    with
    | Not_found -> type_of_var prgm fn vnm
  in
  E.typ_source (src prgm fn);
  E.typ_of expr

let invariants cfg =
  match lookup cfg with
  | None -> raise (Error "invariants: cfg does not exist")
  | Some (stmt, extra) -> extra.inv

(* control variables *)

let auto_asserts b = C.autoasserts := b
