(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * Interface
 * ---------
 * Main interface to language.  See also
 *  o type.mli for module Type
 *  o expr.mli for module Expr
 *)

(*
 * Program
 * -------
 * A program consists of
 *  o a set of types (see Type.mli)
 *  o a set of global constant definitions
 *  o a set of functions
 *)
type prgm

(*
 * Function
 * --------
 * A function consists of
 *  o a precondition and a postcondition
 *  o a return type
 *  o a set of typed parameters
 *  o an implementation as a cfg
 *)
type fn

(*
 * Control Flow Graph
 * ------------------
 * A cfg is a node in a control flow graph (CFG).  It has a stmt,
 * a set of predecessors, and a set of successors.
 *)
type cfg

(*
 * Statement
 *)
type stmt =
  (* Empty statement; cfg is successor. *)
  | Skip of cfg
  (* List of assignments, in order; cfg is successor. *)
  | Asgn of (Expr.t * Expr.t) list * cfg
  (* Branch with a guard and two successors (true and false branches). *)
  | Branch of Expr.t * cfg * cfg
  (* Assertion that expression is true; cfg is successor. *)
  | Assert of Expr.t * cfg
  (* Return with return-value. *)
  | Return of Expr.t

(*
 * Set of functions (fn).
 *)
module FnSet : Set.S with type elt = fn

(*
 * Set of CFG nodes (cfg).
 *)
module CfgSet : Set.S with type elt = cfg


(*
 * Parsing may return a prgm or a command, consisting of a
 * function call.
 *)
type parse =
  | Program of prgm
  | Command of Expr.t

val parse : in_channel -> string -> parse

val parse_string : string -> string -> parse

val parse_file : string -> string -> parse


(*
 * prgm methods
 *)
val types : prgm -> Type.TSet.t

val defines : prgm -> (Type.typ * string * Expr.t) list

val define_of_name : prgm -> string -> (Type.typ * Expr.t)

val functions : prgm -> FnSet.t

val function_of_name : prgm -> string -> fn

val string_of_prgm : prgm -> string


(*
 * fn methods
 *)
val name : fn -> string

val return_type : fn -> Type.typ

val parameters : fn -> (Type.typ * string) list

val return_value : fn -> (Type.typ * string)

val precondition : fn -> Expr.t

val postcondition : fn -> Expr.t

val locals : fn -> (Type.typ * string) list

val body : fn -> cfg


(*
 * Type of named variable and expression.
 *)
val type_of_var : prgm -> fn -> string -> Type.typ

(*
 * First parameter allows temporary variables not in fn scope.
 *)
val type_of_expr : (Type.typ * string) list -> prgm -> fn -> Expr.t -> Type.typ

(*
 * cfg methods
 *)
val id : cfg -> int

val linenum : cfg -> int

val statement : cfg -> stmt

val predecessors : cfg -> CfgSet.t

val successors : cfg -> CfgSet.t

val invariants : cfg -> Expr.t

val is_cutpoint : cfg -> bool

(*
 * For invariant generation.
 *)
val strengthen : cfg -> Expr.t -> unit


(*
 * Control parameters.
 *)

(* If true, runtime assertions are added. *)
val auto_asserts : bool -> unit


exception Error of string
