(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

{
open Parsr

exception Error

module SSet = Set.Make (
  struct
    type t = string
    let compare t1 t2 = compare t1 t2
  end)

let saved = ref SSet.empty

let kwds = Hashtbl.create 53
let _ = 
  List.iter
    (fun (kwd, tok) -> Hashtbl.add kwds kwd tok)
    [ "execute",    EXECUTE;
      "include",    INCLUDE;

      "struct",     STRUCT;
      "typedef",    TYPEDEF;
      "define",     DEFINE;
      "interp",     INTERP;

      "=",          ASSIGN;
      "?",          UNINIT;
      "~",          COLLECT;

      "declare",    DECLARE;

      "if",         IF;
      "else",       ELSE;
      "for",        FOR;
      "while",      WHILE;
      "break",      BREAK;
      "continue",   CONTINUE;
      "return",     RETURN;

      "new",        NEW;
      "length",     LENGTH;

      "forall",     FORALL;
      "exists",     EXISTS;
      "<!>",        XOR;
      "<->",        IFF;
      "->",         IMPLIES;
      "!",          NOT;
      "&&",         AND;
      "||",         OR;
      "true",       TRUE;
      "false",      FALSE;

      "+",          PLUS;
      "-",          MINUS;
      "*",          TIMES;
      "/",          DIV;
      "div",        IDIV;
      "%",          MOD;

      "<",          LT;
      "<=",         LE;
      "==",         EQ;
      "!=",         NE;
      ">=",         GE;
      ">",          GT;
    ];
  List.iter
    (fun kwd -> saved := SSet.add kwd !saved)
    [ "if"; "for"; "while" ]
}

let whitespc = [' ' '\t']
let letter = ['A'-'Z' 'a'-'z']
let number = ['0'-'9']
let alphanum = ['A'-'Z' 'a'-'z' '0'-'9' '_']

rule norm = parse
| whitespc          { Global.last "whitespc"; norm lexbuf }
| '\n'              { Global.last "\\n"; Global.incr_linenum (); norm lexbuf }

| "/*"              { Global.last "/*"; comment 1 lexbuf }

| number+ as n      { Global.last n; INT (int_of_string n) }
| number+ '.' number* (('e' | 'E' | "e+" | "E+" | "e-" | "E-") number+)? as n
                    { Global.last n; FLOAT (float_of_string n) }

| ','               { Global.last ","; COMMA }
| ';'               { Global.last ";"; SEMICOLON }
| '.'               { Global.last "."; PERIOD }
| '('               { Global.last "("; LPAREN }
| ')'               { Global.last ")"; RPAREN }
| '['               { Global.last "["; LSPAREN }
| ']'               { Global.last "]"; RSPAREN }
| '{'               { Global.last "{"; LCPAREN }
| '}'               { Global.last "}"; RCPAREN }

| "@"               { Global.last "@"; Global.save (); ASSERT }
| "@pre"            { Global.last "@pre"; Global.save (); PRE }
| "@post"           { Global.last "@post"; Global.save (); POST }

| ['<' '>' '!' '&' '|' '-' '+' '*' '/' '%' '=' '?' '~']+ as id
                    { Global.last id;
		      try 
			let kwd = Hashtbl.find kwds id in
			if SSet.mem id !saved then Global.save ();
			kwd
		      with
		      | Not_found -> 
			  print_endline ("Bad token: " ^ id);
			  raise Error
                    }

| letter alphanum* as id  
                    { Global.last id;
		      try 
			let kwd = Hashtbl.find kwds id in
			if SSet.mem id !saved then Global.save ();
			kwd
		      with
                      | Not_found -> IDENT id 
                    }

| eof               { Global.last "EOF"; EOF }
| _ as x            { Global.last (String.make 1 x);
		      print_endline ("Bad token: " ^ (String.make 1 x));
		      raise Error
		    }

and comment depth = parse
| "/*"              { comment (depth + 1) lexbuf }
| "*/"              { if depth = 1 then norm lexbuf else 
                      comment (depth - 1) lexbuf
		    }

| '\n'              { Global.incr_linenum (); comment depth lexbuf }
| eof               { print_endline ("Unexpected EOF");
		      raise Error }
| _                 { comment depth lexbuf }
