/*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

%{
module C = Cfg
module E = Expr
module I = Interpreted
module P = Program
module T = Type

module H = Hashtbl
module L = List

exception Error of string

let parse_error msg =
  print_endline 
    ("<error>\nLine <" ^ (string_of_int (Global.get_linenum ())) 
     ^ ">: " ^ msg ^ "\n</error>");
  print_endline ("Last token read: " ^ (Global.get_last ()));
  parse_error msg

let prgm = P.new_program ()
%}

%token <string> IDENT
%token <int> INT
%token <float> FLOAT
%token TRUE
%token FALSE

%token COMMA SEMICOLON PERIOD
%token LPAREN RPAREN LSPAREN RSPAREN LCPAREN RCPAREN DECLARE

%token EXECUTE INCLUDE

%token STRUCT TYPEDEF DEFINE INTERP
%token ASSIGN UNINIT
%token IF ELSE FOR WHILE BREAK CONTINUE RETURN NEW
%token LENGTH

%token ASSERT PRE POST COLLECT

%token FORALL EXISTS
%token XOR IFF IMPLIES
%token NOT AND OR

%token PLUS MINUS TIMES DIV IDIV MOD
%token LT LE EQ NE GE GT

%token EOF

%right COMMA SEMICOLON 
%nonassoc STRUCT TYPEDEF DEFINE INTERP IF ELSE FOR WHILE BREAK CONTINUE RETURN NEW
%nonassoc ASSERT PRE POST DO UNINIT COLLECT
%right ASSIGN
%nonassoc XOR IFF
%right IMPLIES
%right OR
%right AND
%nonassoc NOT
%nonassoc LT LE EQ NE GE GT
%left PLUS MINUS FPLUS FMINUS
%left TIMES DIV IDIV MOD FTIMES FDIV
%nonassoc UMINUS
%nonassoc LENGTH
%left LSPAREN RSPAREN PERIOD

%start main
%type <Pars.parse> main

%%

main:
| decls EOF                 { let rv = Pars.Program prgm in
			      Global.reset_linenum ();
			      rv
			    }
| EXECUTE fncall EOF        { let rv = Pars.Command $2 in
                              Global.reset_linenum ();
                              rv
                            }
;

decls:
| typedecl decls            {}
| define decls              {}
| functiondecl decls        {}
| interp decls              {}
| /* empty */               {}


typedecl:
| typ SEMICOLON             { P.declare_typ prgm $1 }
| TYPEDEF typ IDENT SEMICOLON
    { P.declare_typ prgm $2;
      P.declare_typ_alias prgm $2 $3
    }

typ:
| STRUCT IDENT LCPAREN fielddecls RCPAREN
    { let typ = T.struct_typ $2 $4 in
      P.declare_typ_alias prgm typ ("struct " ^ $2);
      typ
    }
| typ LSPAREN RSPAREN       { T.array_typ $1 }
| STRUCT IDENT              { T.typ ("struct " ^ $2) }
| IDENT                     { T.typ $1 }

fielddecls:
| fielddecl fielddecls      { $1 :: $2 }
| fielddecl                 { [$1] }

fielddecl:
| typ IDENT SEMICOLON       { ($1, $2) }

define:
| DEFINE typ IDENT ASSIGN rval SEMICOLON
                            { P.declare_def prgm $2 $3 $5 }

functiondecl:
| functionhdr block         { P.declare_fn_impl prgm $1 $2 }

functionhdr:
| PRE ssent POST ssent typ IDENT LPAREN parameters RPAREN
                            { P.declare_fn prgm $2 $4 $5 $6 $8; $6 }
| PRE ssent POST ssent typ IDENT LPAREN RPAREN
                            { P.declare_fn prgm $2 $4 $5 $6 []; $6 }

parameters:
| parameter COMMA parameters
                            { $1 :: $3 }
| parameter                 { [$1] }

parameter:
| typ IDENT                 { ($1, $2) }

interp:
| INTERP typ IDENT LPAREN parameters RPAREN SEMICOLON
                            { I.declare $2 $3 $5 }
/* special case, as last parameter is sentence */
| INTERP typ FORALL LPAREN parameters RPAREN SEMICOLON
                            { I.declare $2 "forall" $5 }
| INTERP typ EXISTS LPAREN parameters RPAREN SEMICOLON
                            { I.declare $2 "exists" $5 }

/* organized to reduce number of empty_stmts */
block:
| LCPAREN RCPAREN           { C.empty_blk () }
| stmt                      { C.blk [] $1 }
| LCPAREN stmts RCPAREN     { C.blk [] $2 }
| DECLARE vardecls LCPAREN RCPAREN
                            { C.empty_blk () }
| DECLARE vardecls LCPAREN stmts RCPAREN
                            { C.blk $2 $4 }

vardecls:
| vardecl vardecls          { $1 @ $2 }
| /* empty */               { [] }

vardecl:
| typ varlist SEMICOLON     { L.map (fun v -> ($1, v)) $2 }

varlist:
| IDENT COMMA varlist       { $1 :: $3 }
| IDENT                     { [$1] }

stmts:
| stmt stmts                { C.concat_stmts $1 $2 }
| stmt                      { $1 }

stmt:
| assign SEMICOLON          { $1 }
| IF LPAREN sent RPAREN block els
                            { C.if_stmt $3 $5 $6 }
| FOR ASSERT ssent LPAREN assign SEMICOLON sent SEMICOLON assign RPAREN block
                            { C.for_stmt $3 $5 $7 $9 $11 }
| WHILE ASSERT ssent LPAREN sent RPAREN block
                            { C.while_stmt $3 $5 $7 }
| CONTINUE SEMICOLON        { C.continue_stmt () }
| BREAK SEMICOLON           { C.break_stmt () }
//| RETURN SEMICOLON          { C.return_stmt None }
| RETURN rval SEMICOLON     { C.return_stmt (Some $2) }
| ASSERT ssent SEMICOLON    { C.assert_stmt $2 }

els:
| ELSE block                { $2 }
| /* empty */               { C.empty_blk () }

assign:
| lval ASSIGN rval
                            { C.assign_stmt $1 $3 }
| lval ASSIGN NEW typ LSPAREN expr RSPAREN
                            { C.assign_stmt $1 (E.array_alloc $4 $6) }

sent:
| sent XOR sent             { E.bxor $1 $3 }
| sent IFF sent             { E.biff $1 $3 }
| sent IMPLIES sent         { E.bimplies $1 $3 }
| sent OR sent              { E.bor $1 $3 }
| sent AND sent             { E.band $1 $3 }
| NOT sent                  { E.bnot $2 }
| LPAREN sent RPAREN        { $2 }  /* source of shift/reduce conflict */
| atom                      { $1 }

atom:
| expr                      { E.bexpr $1 }
| expr LT expr              { E.lt $1 $3 }
| expr LE expr              { E.le $1 $3 }
| expr EQ expr              { E.eq $1 $3 }
| expr NE expr              { E.ne $1 $3 }
| expr GE expr              { E.ge $1 $3 }
| expr GT expr              { E.gt $1 $3 }

expr:
| TRUE                      { E.const_true }
| FALSE                     { E.const_false }
| INT                       { E.const_int $1 }
| FLOAT                     { E.const_float $1 }
| init                      { $1 }
| lval                      { $1 }
| LENGTH LPAREN lval RPAREN { E.length $3 }
| expr PLUS expr            { E.plus $1 $3 }
| expr MINUS expr           { E.minus $1 $3 }
| expr TIMES expr           { E.times $1 $3 }
| expr DIV expr             { E.div $1 $3 }
| expr IDIV expr            { E.idiv $1 $3 }
| expr MOD expr             { E.modo $1 $3 }
| MINUS expr %prec UMINUS   { E.uminus $2 }
| LPAREN expr RPAREN        { $2 }

init:
| LCPAREN fldinits RCPAREN  { E.const_struct $2 }
| LSPAREN arrinits RSPAREN  { E.const_array $2 }
| LSPAREN RSPAREN           { E.const_array [] }

fldinits:
| fldinit COMMA fldinits    { $1 :: $3 }
| fldinit                   { [$1] }

fldinit:
| IDENT ASSIGN rval         { ($1, $3) }

arrinits:
| expr SEMICOLON arrinits   { $1 :: $3 }
| expr                      { [$1] }

lval:
| IDENT                     { E.var $1 }
| lval LSPAREN expr RSPAREN { E.array_access $1 $3 }
| lval PERIOD IDENT         { E.struct_access $1 $3 }

rval:
| expr                      { $1 }
| fncall                    { $1 }
| UNINIT                    { E.uninit }

fncall:
| IDENT LPAREN arguments RPAREN
                            { E.fn_call $1 $3 }
| IDENT LPAREN RPAREN       { E.fn_call $1 [] }

arguments:
| argument COMMA arguments  { $1 :: $3 }
| argument                  { [$1] }

argument:
| expr                      { $1 }

/* allows embedded function calls */
ssent:
| ssent XOR ssent           { E.bxor $1 $3 }
| ssent IFF ssent           { E.biff $1 $3 }
| ssent IMPLIES ssent       { E.bimplies $1 $3 }
| ssent OR ssent            { E.bor $1 $3 }
| ssent AND ssent           { E.band $1 $3 }
| NOT ssent                 { E.bnot $2 }
| LPAREN ssent RPAREN       { $2 }  /* source of shift/reduce conflict */
| satom                     { $1 }

satom:
| sexpr                     { E.bexpr $1 }
| sexpr LT sexpr            { E.lt $1 $3 }
| sexpr LE sexpr            { E.le $1 $3 }
| sexpr EQ sexpr            { E.eq $1 $3 }
| sexpr NE sexpr            { E.ne $1 $3 }
| sexpr GE sexpr            { E.ge $1 $3 }
| sexpr GT sexpr            { E.gt $1 $3 }

sexpr:
| TRUE                      { E.const_true }
| FALSE                     { E.const_false }
| INT                       { E.const_int $1 }
| FLOAT                     { E.const_float $1 }
| init                      { $1 }
| lval                      { $1 }
| LENGTH LPAREN lval RPAREN { E.length $3 }
| sfncall                   { $1 }
| COLLECT collect           { $2 }
| sexpr PLUS sexpr          { E.plus $1 $3 }
| sexpr MINUS sexpr         { E.minus $1 $3 }
| sexpr TIMES sexpr         { E.times $1 $3 }
| sexpr DIV sexpr           { E.div $1 $3 }
| sexpr IDIV sexpr          { E.idiv $1 $3 }
| sexpr MOD sexpr           { E.modo $1 $3 }
| MINUS sexpr %prec UMINUS  { E.uminus $2 }
| LPAREN sexpr RPAREN       { $2 }

sfncall:
| IDENT LPAREN sarguments RPAREN
                            { E.fn_call $1 $3 }
| IDENT LPAREN RPAREN       { E.fn_call $1 [] }
/* special case, as last parameter is sentence */
| FORALL LPAREN sexpr COMMA sexpr COMMA sexpr COMMA ssent RPAREN
                            { E.fn_call "forall" [$3; $5; $7; $9] }
| EXISTS LPAREN sexpr COMMA sexpr COMMA sexpr COMMA ssent RPAREN
                            { E.fn_call "exists" [$3; $5; $7; $9] }

sarguments:
| sexpr COMMA sarguments    { $1 :: $3 }
| sexpr                     { [$1] }

collect:
| IDENT LPAREN sarguments RPAREN
                            { E.collect $1 $3 }
