(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module C = Cfg
module E = Expr
module T = Type

module H = Hashtbl
module L = List

type fn = {
    fpre : E.t;
    fpost : E.t;
    ftyp : T.typ;
    fnm : string;
    fparams : (T.typ * string) list;
  }

module FSet = Set.Make (
  struct
    type t = fn
    let compare t1 t2 = compare t1 t2
  end)

type t = {
    typs : T.TSet.t ref;
    typ_aliases : (string, T.typ) H.t;
    defs : (T.typ * string * E.t) list ref;
    fns : (string, (fn * C.blk option)) H.t;
  }

let isrc = ref None
let interp_source isrc' = isrc := Some isrc'
let interpreted fnm =
  match !isrc with
  | None -> assert false
  | Some isrc -> isrc fnm
    
let alias prgm als =
  try Some (H.find prgm.typ_aliases als) with
  | Not_found -> None

let typ prgm fnm vnm =
  let fn, impl = 
    try H.find prgm.fns fnm with
    | Not_found -> 
	raise (Lexc.SemanticError ("Function " ^ fnm ^ " undeclared"))
  in
  try
    let typ, _, _ = L.find (fun (typ, dnm, _) -> vnm = dnm) !(prgm.defs) in
    typ
  with
  | Not_found -> begin
      try 
	let typ, _ = L.find (fun (typ, vnm') -> vnm = vnm') fn.fparams in
	typ
      with
      | Not_found -> begin
	  match impl with
	  | None -> 
	      raise (Lexc.SemanticError ("Function " ^ fnm ^ " unimplemented"))
	  | Some impl -> C.typ impl vnm
      end
  end

let fns prgm fnm =
  try
    let hdr, _ = H.find prgm.fns fnm in
    (hdr.fpre, hdr.fpost, hdr.ftyp, hdr.fparams)
  with
  | Not_found ->
      raise (Lexc.SemanticError ("Function " ^ fnm ^ " undeclared"))

let simple_fns prgm fnm =
  try
    let hdr, _ = H.find prgm.fns fnm in
    (hdr.ftyp, L.map (fun (t, _) -> t) hdr.fparams)
  with
  | Not_found ->
      raise (Lexc.SemanticError ("Function " ^ fnm ^ " undeclared"))

let new_program () =
  let prgm =
    { typs = ref T.TSet.empty;
      typ_aliases = H.create 13;
      defs = ref [];
      fns = H.create 13;
    }
  in
  T.alias_source (alias prgm);
  E.function_source (simple_fns prgm);
  C.function_source (fns prgm);
  prgm

let declare_typ prgm typ =
  prgm.typs := T.TSet.add (T.expand typ) !(prgm.typs)

let declare_typ_alias prgm typ als =
  if H.mem prgm.typ_aliases als then 
  raise (Lexc.SemanticError ("Alias " ^ als ^ " already declared")) else
  H.add prgm.typ_aliases als (T.expand typ)

let declare_def prgm typ dnm expr =
  prgm.defs := (typ, dnm, expr) :: !(prgm.defs)

let declare_fn prgm pre post rtyp fnm params =
  try
    let fn, _ = H.find prgm.fns fnm in
    if fn.fpre <> pre
	or fn.fpost <> post
	or not (T.equal_typ fn.ftyp rtyp)
	or fn.fparams <> params then
      raise (Lexc.SemanticError ("Function " ^ fnm ^ " already declared"))
  with
  | Not_found -> 
      let fn =
	{ fpre = pre;
	  fpost = post;
	  ftyp = rtyp;
	  fnm = fnm;
	  fparams = params;
	}
      in
      H.add prgm.fns fnm (fn, None)

let declare_fn_impl prgm fnm blk =
  E.typ_source (typ prgm fnm);
  let fnhdr =
    try 
      match H.find prgm.fns fnm with
      | (_, Some _) -> 
	  raise (Lexc.SemanticError 
		   ("Function " ^ fnm ^ " already implemented"))
      | (fnhdr, None) -> fnhdr
    with
    | Not_found -> 
	raise (Lexc.SemanticError ("Function " ^ fnm ^ " undeclared"))
  in
  let fn = 
    C.mk_function (L.map (fun (_, defnm, _) -> defnm) !(prgm.defs))
      fnm fnhdr.ftyp fnhdr.fparams fnhdr.fpre fnhdr.fpost blk
  in
  H.replace prgm.fns fnm (fnhdr, Some fn)

let string_of_program prgm =
  let styps =
    let styps = ref [] in
    T.TSet.iter 
      (fun t -> styps := (T.string_of_typ t) :: !styps) 
      !(prgm.typs);
    String.concat ";\n\n" !styps
  in
  let saliases =
    let sal = ref [] in
    (* don't print "struct x" aliases *)
    let is_struct_al s =
      String.length s > 7 && String.sub s 0 7 = "struct "
    in
    H.iter 
      (fun al typ -> 
	if not (is_struct_al al) then
	sal := 
	  ("typedef " ^ (T.short_string_of_typ typ) ^ " " ^ al ^ ";") :: !sal)
      prgm.typ_aliases;
    String.concat "\n\n" !sal
  in
  let sdefs =
    String.concat "\n"
      (L.map
	 (fun (typ, dnm, expr) ->
	   (T.short_string_of_typ typ) ^ " " ^ dnm ^ " = " 
	   ^ (E.string_of_expr expr) ^ ";\n")
	 !(prgm.defs))
  in
  let shdr fn =
    let sp = 
      String.concat ", " 
	(L.map (fun (typ, v) -> (T.string_of_typ typ) ^ " " ^ v) fn.fparams)
    in
    (T.string_of_typ fn.ftyp) ^ " " ^ fn.fnm ^ "(" ^ sp ^ ")"
  in
  let fnhdrs =
    let sfnh = ref [] in
    H.iter
      (fun fnm (hdr, _) -> 
	if not (interpreted fnm) then
	sfnh := ((shdr hdr) ^ ";") :: !sfnh)
      prgm.fns;
    String.concat "\n\n" !sfnh
  in
  let fnbdy =
    let sfnb = ref [] in
    H.iter
      (fun fnm (hdr, blk) ->
	match blk with
	| None -> ()
	| Some blk -> 
	    if not (interpreted fnm) then begin
	      E.typ_source (typ prgm fnm);
	      sfnb := (shdr hdr ^ " " ^ (C.string_of_blk blk)) :: !sfnb
	    end)
      prgm.fns;
    String.concat "\n\n" !sfnb
  in
  "/* --- Types --- */\n\n" ^ styps
  ^ "\n\n/* --- Aliases --- */\n\n" ^ saliases
  ^ "\n\n/* --- Definitions --- */\n\n" ^ sdefs
  ^ "\n\n/* --- Function declarations --- */\n\n" ^ fnhdrs
  ^ "\n\n/* --- Function implementations --- */\n\n" ^ fnbdy

let type_check prgm =
  L.iter
    (fun (typ, _, expr) ->
      E.assert_typ typ expr)
    !(prgm.defs);
  H.iter
    (fun fnm (hdr, blk) ->
      match blk with
      | None -> raise (Lexc.SemanticError (fnm ^ " unimplemented"))
      | Some blk ->
	  E.typ_source (typ prgm fnm);
	  C.type_check (Some fnm) blk)
    prgm.fns

