(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * TODO: factor this code with vcg.ml
 *)

module E = Expr
module I = Interface
module T = Type

module A = Array
module H = Hashtbl
module L = List

type expr =
  | Bool of bool
  | Int of int
  | Float of float
  | Anything
  | CArray of expr array
  | CStruct of (string * expr) list
  | Var of string
  | Array of (expr * expr) list * expr
  | Length of expr
  | ArrayAcc of expr * expr
  | Struct of (string * expr) list * expr
  | StructAcc of expr * string
  | FnCall of string * expr list
  | Plus of expr * expr
  | Minus of expr * expr
  | Times of expr * expr
  | Div of expr * expr
  | IDiv of expr * expr
  | Mod of expr * expr
  | UMinus of expr
  | LT of expr * expr
  | LE of expr * expr
  | NE of expr * expr
  | EQ of expr * expr
  | GE of expr * expr
  | GT of expr * expr
  | BExpr of expr
  | And of expr * expr
  | Or of expr * expr
  | Not of expr
  | Forall of Type.typ * string * expr

let rec base_array = function
  | Array (_, init) -> base_array init
  | x -> x

let string_of_expr expr =
  let rec soe = function
    | Bool b -> if b then "true" else "false"
    | Int i -> string_of_int i
    | Float f -> string_of_float f
    | Anything -> "ANYTHING"
    | CArray arr ->
        let elems = ref [] in
        A.iter (fun e -> elems := (soe e) :: !elems) arr;
        "[" ^ (String.concat "; " (L.rev !elems)) ^ "]"
    | CStruct flds ->
        "{" ^ (String.concat ", " 
                 (L.map (fun (fld, e) -> fld ^ " = " ^ (soe e)) flds)) 
        ^ "}"
    | Var s -> s
    | Array (upd, init) ->
        let rec update = function
          | (i, v) :: t -> (update t) ^ "[" ^ (soe i) ^ " = " ^ (soe v) ^ "]"
          | [] -> ""
        in
        (soe init) ^ (update upd)
    | Length e -> "length(" ^ (soe (base_array e)) ^ ")"
    | ArrayAcc (e, i) -> (soe e) ^ "[" ^ (soe i) ^ "]"
    | Struct (upd, init) ->
        let rec update = function
          | (fld, v) :: t -> (update t) ^ "{" ^ fld ^ " = " ^ (soe v) ^ "}"
          | [] -> ""
        in
        (soe init) ^ (update upd)
    | StructAcc (st, fld) -> (soe st) ^ "." ^ fld
    | FnCall (fnm, args) ->
	fnm ^ "(" ^ (String.concat ", " (L.map soe args)) ^ ")"
    | Plus (e1, e2) -> (psoe e1) ^ " + " ^ (psoe e2)
    | Minus (e1, e2) -> (psoe e1) ^ " - " ^ (psoe e2)
    | Times (e1, e2) -> (psoe e1) ^ " * " ^ (psoe e2)
    | Div (e1, e2) -> (psoe e1) ^ " / " ^ (psoe e2)
    | IDiv (e1, e2) -> (psoe e1) ^ " div " ^ (psoe e2)
    | Mod (e1, e2) -> (psoe e1) ^ " % " ^ (psoe e2)
    | UMinus e -> "-" ^ (psoe e)
    | LT (e1, e2) -> (soe e1) ^ " < " ^ (soe e2)
    | LE (e1, e2) -> (soe e1) ^ " <= " ^ (soe e2)
    | NE (e1, e2) -> (soe e1) ^ " != " ^ (soe e2)
    | EQ (e1, e2) -> (soe e1) ^ " = " ^ (soe e2)
    | GE (e1, e2) -> (soe e1) ^ " >= " ^ (soe e2)
    | GT (e1, e2) -> (soe e1) ^ " > " ^ (soe e2)
    | BExpr e -> soe e
    | And (e1, e2) -> (psoe e1) ^ " && " ^ (psoe e2)
    | Or (e1, e2) -> (psoe e1) ^ " || " ^ (psoe e2)
    | Not e -> "!" ^ (psoe e)
    | Forall (typ, v, e) -> 
        "(forall " ^ v ^ " : " ^ (T.string_of_typ typ) ^ "\n" ^ (soe e) ^ ")"
  and psoe expr =
    match expr with
    | Bool _
    | Int _
    | Float _
    | Var _
    | ArrayAcc _
    | StructAcc _
    | Length _ -> soe expr
    | _ -> "(" ^ (soe expr) ^ ")"
  in
  soe expr

module SSet = Set.Make (
  struct
    type t = string
    let compare t1 t2 = compare t1 t2
  end)

let vars expr =
  let vs = ref SSet.empty in
  let rec vars = function
    | Var s -> vs := SSet.add s !vs
    | Array (upd, init) ->
	L.iter (fun (e1, e2) -> vars e1; vars e2) upd;
	vars init
    | Length e -> vars (base_array e)
    | Struct (upd, init) ->
	L.iter (fun (_, e) -> vars e) upd;
	vars init
    | FnCall (_, args) -> L.iter vars args
    | ArrayAcc (e1, e2)
    | Plus (e1, e2)
    | Minus (e1, e2)
    | Times (e1, e2)
    | Div (e1, e2)
    | IDiv (e1, e2)
    | Mod (e1, e2)
    | LT (e1, e2)
    | LE (e1, e2)
    | NE (e1, e2)
    | EQ (e1, e2)
    | GE (e1, e2)
    | GT (e1, e2)
    | And (e1, e2)
    | Or (e1, e2) -> vars e1; vars e2
    | StructAcc (e, _)
    | UMinus e
    | BExpr e
    | Not e -> vars e
    | _ -> ()
  in
  vars expr;
  !vs

let rec eoe = function
  | E.Bool b -> Bool b
  | E.Int i -> Int i
  | E.Float f -> Float f
  | E.Uninit -> Anything
  | E.CArray arr -> CArray (A.map eoe arr)
  | E.CStruct flds -> CStruct (L.map (fun (fld, e) -> (fld, eoe e)) flds)
  | E.Var s -> Var s
  | E.ArrayAcc (arr, i) -> ArrayAcc (eoe arr, eoe i)
  | E.Length e -> Length (eoe e)
  | E.StructAcc (st, fld) -> StructAcc (eoe st, fld)
  | E.FnCall (fnm, args) -> FnCall (fnm, L.map eoe args)
  | E.Collect (fnm, args) -> FnCall (fnm, L.map eoe args)
  | E.Plus (e1, e2) -> Plus (eoe e1, eoe e2)
  | E.Minus (e1, e2) -> Minus (eoe e1, eoe e2)
  | E.Times (e1, e2) -> Times (eoe e1, eoe e2)
  | E.Div (e1, e2) -> Div (eoe e1, eoe e2)
  | E.IDiv (e1, e2) -> IDiv (eoe e1, eoe e2)
  | E.Mod (e1, e2) -> Mod (eoe e1, eoe e2)
  | E.UMinus e -> UMinus (eoe e)
  | E.LT (e1, e2) -> LT (eoe e1, eoe e2)
  | E.LE (e1, e2) -> LE (eoe e1, eoe e2)
  | E.NE (e1, e2) -> NE (eoe e1, eoe e2)
  | E.EQ (e1, e2) -> EQ (eoe e1, eoe e2)
  | E.GE (e1, e2) -> GE (eoe e1, eoe e2)
  | E.GT (e1, e2) -> GT (eoe e1, eoe e2)
  | E.BExpr e -> BExpr (eoe e)
  | E.And (e1, e2) -> And (eoe e1, eoe e2)
  | E.Or (e1, e2) -> Or (eoe e1, eoe e2)
  | E.Not e -> Not (eoe e)
  | _ -> assert false

let type_of_expr tvars prgm fn expr =
  let rec ieoe = function
    | Bool b -> E.Bool b
    | Int i -> E.Int i
    | Float f -> E.Float f
    | Anything -> E.Uninit
    | CArray arr -> E.CArray (A.map ieoe arr)
    | CStruct flds -> E.CStruct (L.map (fun (fld, e) -> (fld, ieoe e)) flds)
    | Var s -> E.Var s
    | Array (upd, init) -> (* not exact! *) ieoe init
    | Length e -> E.Length (ieoe e)
    | ArrayAcc (arr, i) -> E.ArrayAcc (ieoe arr, ieoe i)
    | Struct (upd, init) -> (* !not exact! *) ieoe init
    | StructAcc (st, fld) -> E.StructAcc (ieoe st, fld)
    | Plus (e1, e2) -> E.Plus (ieoe e1, ieoe e2)
    | Minus (e1, e2) -> E.Minus (ieoe e1, ieoe e2)
    | Times (e1, e2) -> E.Times (ieoe e1, ieoe e2)
    | Div (e1, e2) -> E.Div (ieoe e1, ieoe e2)
    | IDiv (e1, e2) -> E.IDiv (ieoe e1, ieoe e2)
    | Mod (e1, e2) -> E.Mod (ieoe e1, ieoe e2)
    | UMinus e -> E.UMinus (ieoe e)
    | LT (e1, e2) -> E.LT (ieoe e1, ieoe e2)
    | LE (e1, e2) -> E.LE (ieoe e1, ieoe e2)
    | NE (e1, e2) -> E.NE (ieoe e1, ieoe e2)
    | EQ (e1, e2) -> E.EQ (ieoe e1, ieoe e2)
    | GE (e1, e2) -> E.GE (ieoe e1, ieoe e2)
    | GT (e1, e2) -> E.GT (ieoe e1, ieoe e2)
    | BExpr e -> E.BExpr (ieoe e)
    | FnCall (fnm, args) -> begin
	try
	  ignore (I.function_of_name prgm fnm);
	  E.FnCall (fnm, L.map ieoe args)
	with
	| _ -> E.Collect (fnm, L.map ieoe args)
    end
    | And (e1, e2) -> E.And (ieoe e1, ieoe e2)
    | Or (e1, e2) -> E.Or (ieoe e1, ieoe e2)
    | Not e -> E.Not (ieoe e)
    | _ -> assert false
  in
  match expr with
  | Forall _ -> T.Bool
  | x -> I.type_of_expr tvars prgm fn (ieoe x)

let is_loop cfg not_thru =
  let rec dfs first seen curr =
    if not first && curr = cfg then true else
    if curr = not_thru or I.CfgSet.mem curr !seen then false else begin
      seen := I.CfgSet.add curr !seen;
      I.CfgSet.exists (fun s -> dfs false seen s) (I.successors curr)
    end
  in
  dfs true (ref I.CfgSet.empty) cfg

let paths notify fn =
  (* generates paths starting from cfg *)
  let paths cfg =
    (* next holds path ends *)
    let next = ref I.CfgSet.empty in
    let rec paths first curr =
      let is_loop curr =
        (* Only Assert statements can be loop merge-points. *)
	match I.statement curr with
        | I.Assert _ -> is_loop curr cfg
        | _ -> false
      in
      let next_is_return curr =
        match I.statement curr with
        | I.Assert (_, next) -> begin
            match I.statement next with
            | I.Return _ -> true
            | _ -> false
        end
        | _ -> false
      in
      if not first && curr = cfg then begin
        (* reached cfg, no need to add cfg to next *)
        [[curr]]
      end else
      if is_loop curr then begin
        (* curr is a loop (not through cfg), so an end *)
        next := I.CfgSet.add curr !next;
        [[curr]] 
      end else
      if next_is_return curr then begin
        (* avoid ending path with Return *)
	[[curr]]
      end else begin
        (* recursive case *)
        let ps = ref [] in
        I.CfgSet.iter 
	  (fun cfg -> ps := (paths false cfg) @ !ps) 
	  (I.successors curr);
	match I.statement curr with
	| I.Assert _ when curr != cfg ->
	    (* add new VC path for intermediate assert *)
            (* but also add to others so paths are complete *)
            [curr] :: (L.map (fun p -> curr :: p) !ps);
	| _ ->
            (* add curr to existing VC paths *)
            L.map (fun p -> curr :: p) !ps
      end
    in
    (* return paths from cfg, next starts of paths *)
    let paths = paths true cfg in
    (paths, !next)
  in
  (* remembers starting points *)
  let did = ref I.CfgSet.empty in
  (* computes all paths starting from cfgs in starts *)
  let rec allpaths starts =
    let next = ref I.CfgSet.empty in
    I.CfgSet.iter
      (fun cfg ->
        let ps, ns = paths cfg in
        (* notify of path *)
        L.iter (fun p -> notify fn p) ps;
        next := I.CfgSet.union ns !next)
      starts;
    did := I.CfgSet.union starts !did;
    let next = I.CfgSet.diff !next !did in
    if not (I.CfgSet.is_empty next) then allpaths next
  in
  allpaths (I.CfgSet.singleton (I.body fn))

type mem =
  | MScope of (string * mem) list
  | MBasic of expr
  | MStruct of (string * mem) list * expr
  | MArray of (expr * mem) list * expr

type acclist =
  | AVar of string
  | AField of string
  | AIndex of expr

let rec acclist = function
  | Var s -> [AVar s]
  | StructAcc (st, field) -> (acclist st) @ [AField field]
  | ArrayAcc (arr, i) -> (acclist arr) @ [AIndex i]
  | _ -> assert false

let unacclist acc =
  let rec unacc expr = function
    | (AField fld) :: t -> unacc (StructAcc (expr, fld)) t
    | (AIndex i) :: t -> unacc (ArrayAcc (expr, i)) t
    | [] -> expr
    | _ -> assert false
  in
  match acc with
  | (AVar s) :: t -> unacc (Var s) t
  | _ -> assert false

let rec mem_of_expr = function
  | Struct (upd, init) ->
      MStruct (L.map (fun (fld, e) -> (fld, mem_of_expr e)) upd, init)
  | Array (upd, init) ->
      MArray (L.map (fun (i, e) -> (i, mem_of_expr e)) upd, init)
  | x -> MBasic x

let rec expr_of_mem = function
  | MBasic expr -> expr
  | MStruct (upd, init) -> 
      Struct (L.map (fun (fld, u) -> (fld, expr_of_mem u)) upd, init)
  | MArray (upd, init) ->
      Array (L.map (fun (i, u) -> (i, expr_of_mem u)) upd, init)
  | _ -> assert false

let replace hash k v = (k, v) :: (L.remove_assoc k hash)

let read mem expr =
  let new_mem expr = MBasic expr in
  let rec read mem accl =
    match mem, accl with
    (* convert MBasic to specific *)
    | MBasic bexpr, ((AField _) :: t as x) -> read (MStruct ([], bexpr)) x
    | MBasic bexpr, ((AIndex _) :: t as x) -> read (MArray ([], bexpr)) x
	  
    (* bases cases: drilled all the way down *)
    | x, [] -> (x, expr_of_mem x)

    (* recursive cases: try reading, on failure update mem and try again *)
    | MScope sc, ((AVar s) :: t as x) -> begin
        try 
          let mem, expr = read (L.assoc s sc) t in
          (MScope (replace sc s mem), expr)
        with
        | Not_found -> 
            let expr = Var s in
            read (MScope (replace sc s (new_mem expr))) x
    end
    | (MStruct (upd, init) as m), ((AField fld) :: t as x) -> begin
        try 
          let mem, expr = read (L.assoc fld upd) t in
          (MStruct (replace upd fld mem, init), expr)
        with
        | Not_found -> 
            let expr = StructAcc (init, fld) in
            read (MStruct (replace upd fld (new_mem expr), init)) x
    end
    | (MArray (upd, init) as m), ((AIndex i) :: t as x) -> begin
        try 
          let mem, expr = read (L.assoc i upd) t in
          (MArray (replace upd i mem, init), expr)
        with
        | Not_found -> 
            let expr = ArrayAcc (expr_of_mem m, i) in
            read (MArray (replace upd i (new_mem expr), init)) x
    end
    | _ -> assert false
  in
  read mem (acclist expr)

let value mem expr =
  let _, expr = read mem expr in
  expr

let write mem lv rv =
  let rec write mem accl =
    match mem, accl with
    (* bases cases: drilled all the way down *)
    | MScope sc, [AVar s] -> MScope (replace sc s (mem_of_expr rv))
    | MStruct (upd, init), [AField fld] -> 
        MStruct (replace upd fld (mem_of_expr rv), init)
    | MArray (upd, init), [AIndex i] -> 
        MArray (replace upd i (mem_of_expr rv), init)
	  
    (* recursive cases *)
    | MScope sc, (AVar s) :: t -> 
        MScope (replace sc s (write (L.assoc s sc) t))
    | MStruct (upd, init), (AField fld) :: t ->
        MStruct (replace upd fld (write (L.assoc fld upd) t), init)
    | MArray (upd, init), (AIndex i) :: t ->
        MArray (replace upd i (write (L.assoc i upd) t), init)
	  
    | _ -> assert false
  in
  (* read lv to initialize mem *)
  let mem', _ = read mem lv in
  write mem' (acclist lv)

let propagate prgm fn path =

  let mem = ref (MScope []) in
  let read expr = 
    let mem', rv = read !mem expr in
    mem := mem';
    rv
  in
  let write lv rv =
    let mem' = write !mem lv rv in
    mem := mem'
  in

  let rec apply = function
    | Var _ as x -> read x
    | Array _ -> assert false
    | Length e -> Length (apply e)
    | ArrayAcc (arr, i) -> read (ArrayAcc (arr, apply i))
    | Struct _ -> assert false
    | StructAcc _ as x -> read x
    | Plus (e1, e2) -> Plus (apply e1, apply e2)
    | Minus (e1, e2) -> Minus (apply e1, apply e2)
    | Times (e1, e2) -> Times (apply e1, apply e2)
    | Div (e1, e2) -> Div (apply e1, apply e2)
    | IDiv (e1, e2) -> IDiv (apply e1, apply e2)
    | Mod (e1, e2) -> Mod (apply e1, apply e2)
    | UMinus e -> UMinus (apply e)
    | LT (e1, e2) -> LT (apply e1, apply e2)
    | LE (e1, e2) -> LE (apply e1, apply e2)
    | NE (e1, e2) -> NE (apply e1, apply e2)
    | EQ (e1, e2) -> EQ (apply e1, apply e2)
    | GE (e1, e2) -> GE (apply e1, apply e2)
    | GT (e1, e2) -> GT (apply e1, apply e2)
    | BExpr e -> BExpr (apply e)
    | FnCall (fnm, args) -> FnCall (fnm, L.map apply args)
    | And (e1, e2) -> And (apply e1, apply e2)
    | Or (e1, e2) -> Or (apply e1, apply e2)
    | Not e -> Not (apply e)
    | Forall _ -> assert false
    | x -> x
  in

  let update lv rv = write lv (apply rv) in

  let req = ref [] in
  let require sent = req := sent :: !req in

  let fvars = ref [] in
  let fvid = ref 0 in
  let fresh typ =
    incr fvid;
    let fvnm = "t_" ^ (string_of_int !fvid) in
    fvars := (typ, fvnm) :: !fvars;
    fvnm
  in

  let required () = 
    let rec band = function
      | h :: [] -> BExpr h
      | h :: t -> And (BExpr h, band t)
      | [] -> BExpr (Bool true)
    in
    (* form expr just to get relevant vars *)
    let evars = vars (band (!req)) in
    (* vars with types *)
    let tvars expr =
      match !mem with
      | MScope sc -> 
	  let sv = ref [] in
	  L.iter
	    (fun (v, _) ->
	      if SSet.mem v evars
		  && not (L.exists (fun (_, v') -> v = v') !fvars) then
		sv := (I.type_of_var prgm fn v, v) :: !sv)
            sc;
          !sv @ (L.filter (fun (_, v) -> SSet.mem v evars) !fvars)
      | _ -> assert false
    in
    (* assertions about prgm's defines *)
    let defines =
      let def nm = function
	| CArray arr -> 
	    A.to_list (A.mapi (fun i v -> EQ (ArrayAcc(Var nm, Int i), v)) arr)
	| CStruct flds ->
	    L.map (fun (fld, v) -> EQ (StructAcc (Var nm, fld), v)) flds
	| _ -> assert false
      in
      L.concat
	(L.map 
	   (fun (_, nm, expr) -> def nm (eoe expr))
	   (L.filter (fun (_, nm, _) -> SSet.mem nm evars) (I.defines prgm)))
    in
    let e = band (defines @ !req) in
    L.fold_left (fun e (typ, v) -> Forall (typ, v, e)) e (tvars e)
  in

  let rec walk first = function
    | [] -> (required (), !mem)
    | cfg :: tpath -> 
        begin
          match I.statement cfg with
          | I.Skip _ -> ()
          | I.Asgn (asgns, _) ->
              let asgn = function
                | (lv, E.FnCall (fnm, args)) ->
                    let fnc = I.function_of_name prgm fnm in
                    (* params -> args *)
                    let vmap = H.create 13 in
                    L.iter2
                      (fun (_, p) arg -> H.add vmap p arg)
                      (I.parameters fnc) args;
		    (* map rv -> fresh var *)
                    let rtyp, rv = I.return_value fnc in
                    let tvar = fresh rtyp in
                    H.add vmap rv (E.Var tvar);
                    (* postcondition becomes transition relation *)
                    let rltn = E.sub vmap (I.postcondition fnc) in
                    (* update lv with fresh variable *)
                    update (eoe lv) (Var tvar);
                    (* assert that rltn holds, constraining tvar *)
                    require (apply (eoe rltn))
		| (lv, E.ArrayAlloc (typ, sz)) ->
		    let avar = fresh (T.array_typ typ) in
		    let lv = eoe lv in
		    update lv (Var avar);
		    require (EQ (Length lv, eoe sz))
                | (lv, rv) -> update (eoe lv) (eoe rv)
              in
              L.iter asgn asgns
          | I.Branch (grd, tb, fb) -> begin
              (* determine whether grd or not grd *)
              let grd = match tpath with
              | h :: t -> if tb = h then grd else E.bnot grd
              | _ -> assert false
              in
              require (apply (eoe grd))
          end
          | I.Return _ -> assert false
	  | _ -> ()
        end;
        walk false tpath
  in
  walk true path
