(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module L = List

type typ =
  | Int
  | Float
  | Bool
  | Alias of string
  | UIT of string
  | Struct of string * (typ * string) list
  | Array of typ

module TSet = Set.Make (
  struct
    type t = typ
    let compare t1 t2 = compare t1 t2
  end)

let als = ref None

let alias_source a = als := Some a

let alias nm =
  match !als with
  | None -> None
  | Some als -> als nm

let typ = function
  | "int" -> Int
  | "float" -> Float
  | "bool" -> Bool
  | x -> begin
      match alias x with
      | None -> UIT x
      | Some t -> Alias x
  end

let struct_typ nm flds = 
  let rnm = "struct " ^ nm in
  let rec rds = function
    | UIT tnm as x -> if tnm = rnm then Alias rnm else x
    | Struct (nm, flds) ->
	Struct (nm, L.map (fun (typ, fld) -> (rds typ, fld)) flds)
    | Array t -> Array (rds t)
    | x -> x
  in
  Struct (nm, L.map (fun (typ, fld) -> (rds typ, fld)) flds)

let array_typ typ = Array typ

module SSet = Set.Make (
  struct
    type t = string
    let compare t1 t2 = compare t1 t2
  end)

let expand typ =
  let except = ref SSet.empty in
  let rec expand = function
    | Alias nm when not (SSet.mem nm !except) -> begin
	match alias nm with
	| None -> 
	    (* "struct x" may not yet be declared *)
	    if SSet.mem nm !except then Alias nm else begin
	      print_endline nm; 
	      assert false
	    end
	| Some t -> expand t
    end
    | Struct (nm, flds) ->
	except := SSet.add ("struct " ^ nm) !except;
	Struct (nm, L.map (fun (t, fld) -> (expand t, fld)) flds)
    | Array t -> Array (expand t)
    | x -> x
  in
  expand typ

let rec is_array = function
  | Array _ -> true
  | Alias x -> begin
      match alias x with
      | None -> false
      | Some t -> is_array t
  end
  | _ -> false

let equal_typ t1 t2 = 
  expand t1 = expand t2

let string_of_typ typ =
  let rec sot = function
    | Int -> "int"
    | Float -> "float"
    | Bool -> "bool"
    | Alias nm -> nm
    | UIT nm -> "(UIT:" ^ nm ^ ")"
    | Struct (nm, flds) ->
	"struct " ^ nm ^ " {\n  "
	^ (String.concat ";\n  " 
	     (L.map (fun (t, fld) -> (sot t) ^ " " ^ fld) flds))
	^ ";\n}"
    | Array t -> (sot t) ^ "[]"
      (* "struct { " ^ (sot t) ^ "[] e; int l; }" *)
  in
  sot typ


let step_string_of_typ typ =
  let rec sot = function
    | Int -> "int"
    | Float -> "float"
    | Bool -> "bool"
    | Alias nm -> nm
    | UIT nm -> "(UIT:" ^ nm ^ ")"
    | Struct (nm, flds) ->
	"struct " ^ nm ^ " {\n  "
	^ (String.concat ";\n  " 
	     (L.map (fun (t, fld) -> (sot t) ^ " " ^ fld) flds))
	^ ";\n}"
    | Array t -> " array [0..N] of " ^ (sot t)
      (* "struct { " ^ (sot t) ^ "[] e; int l; }" *)
  in
  sot typ

let rec short_string_of_typ = function
  | Int -> "int"
  | Float -> "float"
  | Bool -> "bool"
  | Alias nm -> nm
  | UIT nm -> nm
  | Struct (nm, flds) -> "struct " ^ nm
  | Array t -> (short_string_of_typ t) ^ "[]"
    (* "struct { " ^ (short_string_of_typ t) ^ "[] e; int l; }" *)
