(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module I = Interface
module L = List

exception NotValid

let _ =
  let print_error msg =
    print_endline ("<error>\n" ^ msg ^ "\n</error>")
  in
  try
    let prgm = ref None in
    let cmd = ref None in
    let dir = ref "." in
    let run ch =
      match I.parse ch !dir with
      | I.Program p -> prgm := Some p
      | I.Command cmd -> print_endline ("Program expected.")
    in
    let rd = ref false in
    let evalfn expr = cmd := Some ("execute " ^ expr) in
    let vcg = ref false in
    let dot = ref false in
    let prove = ref false in
    let abstract = ref false in
    let smt_strings = ref false in
    let timeout = ref 3600 in
    let specs = 
      [ ("-eval", Arg.String evalfn, "evaluate expression");
	("-vc", Arg.Set vcg, "print VCs");
	("-rta", Arg.Unit (fun () -> I.auto_asserts true),
	 "produce runtime assertions");
	("-prove", Arg.Set prove, "send VCs to DP");
	("-smt", Arg.Set smt_strings, "smt strings");
	("-cvcl", Arg.Set_string Cvc.cvc_exec, 
	 "use given executable of CVC Lite");
	("-yices", Arg.Set_string Yices.yices_exec, 
	 "use given executable of Yices");
	("-timeout", Arg.Set_int timeout, "timeout for DP");
	("-dir", Arg.Set_string dir, "working directory");
	("-dot", Arg.Set dot, "dot output");
	("-abstract", Arg.Set abstract, "abstract");
	("-sanity", Arg.Set Vcg.sanity, "check generated invariants");
      ]
    in
    Arg.parse specs (fun fl -> rd := true; run (open_in fl)) "vc <file>";
    if !Cvc.cvc_exec != "" then Cvc.cvc_timeout := !timeout;
    if !Yices.yices_exec != "" then Yices.yices_timeout := !timeout;
    if not !rd then run stdin;
    match !prgm, !cmd with
    | Some prgm, None -> 
	print_endline (I.string_of_prgm prgm);
	if !vcg then begin
	  let correct = ref true in
	  Vcg.vcs prgm 
	    (fun nm path prgm fn e -> 
	      let vcs id e =
		let ids = if id <> "" then " (" ^ id ^ ")" else "" in
		print_endline 
		  ("<function>\n" ^ (I.name fn) ^ "\n</function>\n"
		   ^ "<vcname>\n" ^ nm ^ ids ^ "\n</vcname>\n"
		   ^ "<path>\n" ^ path ^ "\n</path>\n"
		   ^ "<vc>\n" ^ (Vcg.string_of_expr e) ^ "\n</vc>\n")
	      in
	      vcs "" e;
	      if !prove then begin
		let valid = ref true in
		let split_valid svc =
		  let vld, svc' = Idp.valid prgm fn svc in
		  if not vld then begin
		    vcs "" svc';
		    print_endline "<NOT VALID>\n";
		    correct := false;
		    valid := false;
		    raise NotValid
		  end
		in
		begin
		  (* only need one invalidity *)
		  try
		    L.iter split_valid (Idp.split e);
		  with
		  | NotValid -> ()
		end;
		if !valid then begin
		  vcs "" e;
		  print_endline "<VALID>\n"
		end
	      end else if !smt_strings then begin
		let smt = Idp.smt_of prgm fn e in
		print_endline ("<SMT>\n" ^ smt ^ "</SMT>");
		let fn, fl = Filename.open_temp_file "piVC_" ".smt" in
		output_string fl smt;
		close_out fl
	      end);
	  print_endline ("<PROGRAM " 
			 ^ (if !correct then "VALID>" else "INVALID>"))
	end;
	if !dot then Todot.dot_of_prgm prgm
    | Some prgm, Some cmd -> 
	print_endline 
	  ("<rv>\n" ^
	   (Interpreter.string_of_value (Interpreter.execute prgm cmd))
	   ^ "\n</rv>")
    | _ -> assert false
  with
  | Lexc.ParseError msg -> print_error ("<Parse Error> " ^ msg)
  | Lexc.SemanticError msg -> print_error ("<Semantic Error> " ^ msg)
  | Lexc.TypeError msg -> print_error ("<Type Error> " ^ msg)
  | Interpreter.AssertionFailure msg -> 
      print_error ("<Assertion Failure> " ^ msg)
  | Interpreter.RuntimeError msg ->
      print_error ("<Runtime Error> " ^ msg)
