(*
 * Copyright 2005, 2006, 2007 Aaron R. Bradley and Zohar Manna
 *
 * This file is part of PiVC.
 * 
 * PiVC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 * 
 * PiVC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module E = Expr
module I = Interface
module T = Type

module A = Array
module H = Hashtbl
module L = List

let sanity = ref false

type expr =
  | Bool of bool
  | Int of int
  | Float of float
  | Anything
  | CArray of expr array
  | CStruct of (string * expr) list
  | Var of string
  | Array of (expr * expr) list * expr
  | Length of expr
  | ArrayAcc of expr * expr
  | Struct of (string * expr) list * expr
  | StructAcc of expr * string
  | FnCall of string * expr list
  | Plus of expr * expr
  | Minus of expr * expr
  | Times of expr * expr
  | Div of expr * expr
  | IDiv of expr * expr
  | Mod of expr * expr
  | UMinus of expr
  | LT of expr * expr
  | LE of expr * expr
  | NE of expr * expr
  | EQ of expr * expr
  | GE of expr * expr
  | GT of expr * expr
  | BExpr of expr
  | And of expr * expr
  | Or of expr * expr
  | Not of expr
  | Forall of Type.typ * string * expr

(* 10/11/2006 *)
module ESet = Set.Make (
  struct
    type t = expr
    let compare e1 e2 = compare e1 e2
  end)
exception SFalse

let rec base_array = function
  | Array (_, init) -> base_array init
  | x -> x


let step_string_of_expr expr =
  let rec ssoe = function
   | And(e1, e2) -> "(" ^ (ssoe e1) ^ " /\\ " ^ (ssoe e2) ^ ")"
   | Or (e1, e2) -> "(" ^ (ssoe e1) ^ " \\/ " ^ (ssoe e2) ^ ")"
   | Not e -> "(!" ^ (ssoe e) ^ ")"
   | Forall (typ, v, e) -> "(Forall " ^ v ^ " : " ^ (T.step_string_of_typ typ) ^
           " . " ^ "(" ^ (ssoe e) ^ "))"
   | BExpr e -> ssoe e
   | e -> soe e
 and soe = function
   | Bool b -> if b then "true" else "false"
   | Int i -> string_of_int i
   | Float f -> string_of_float f
   | Anything -> "ANYTHING"
   | CArray arr -> 
        let elems = ref [] in
        A.iter (fun e ->elems := (soe e) :: !elems) arr;
        "[" ^ (String.concat "; " (L.rev !elems)) ^ "]"
   | CStruct flds ->
        "{" ^ (String.concat ", "
                 (L.map (fun (fld, e) -> fld ^ " = " ^ (soe e)) flds))
         ^ "}"
   | Var s -> s
   | Array (upd, init) ->
       let rec update = function
         | (i, v) :: t -> (update t) ^ "{" ^ (soe i)
           ^ " = " ^ (soe v) ^ "}"
         | [] -> ""
       in
         (soe init) ^ (update upd)
   | Length e -> "lngth(" ^ (soe e) ^ ")"
   | ArrayAcc (e, i) -> (soe e) ^ "[" ^ (soe i) ^ "]"
   | Struct (upd, init) ->
       let rec update = function
         | (fld, v) :: t -> (update t) ^ "{" ^ fld ^ " = " ^ (soe v) ^ "}"
         | [] -> ""
       in
         (soe init) ^ (update upd)
   | StructAcc (st, fld) -> (soe st) ^ "." ^ fld
   | FnCall (fnm, args) ->
         fnm ^ "(" ^ (String.concat ", " (L.map (fun e -> soe e) args)) ^ ")"
   | Plus  (e1, e2) -> (soe e1) ^ " + " ^ (soe e2)
   | Minus (e1, e2) -> (soe e1) ^ " - " ^ (soe e2)
   | Times (e1, e2) -> (soe e1) ^ " * " ^ (soe e2)
   | Div   (e1, e2) -> (soe e1) ^ " / " ^ (soe e2)
   | IDiv  (e1, e2) -> (soe e1) ^ " div " ^ (soe e2)
   | Mod   (e1, e2) -> (soe e1) ^ " mod " ^ (soe e2)
   | UMinus e -> " - " ^ (soe e)
   | LT (e1, e2) -> (soe e1) ^ " <  " ^ (soe e2)
   | LE (e1, e2) -> (soe e1) ^ " <= " ^ (soe e2)
   | NE (e1, e2) -> (soe e1) ^ " != " ^ (soe e2)
   | EQ (e1, e2) -> (soe e1) ^ " = "  ^ (soe e2)
   | GE (e1, e2) -> (soe e1) ^ " >= " ^ (soe e2)
   | GT (e1, e2) -> (soe e1) ^ " > "  ^ (soe e2)
   | BExpr e -> soe e
   | And(e1, e2) -> (soe e1) ^ " /\\ " ^ (soe e2)
   | Or (e1, e2) -> (soe e1) ^ " \\/ " ^ (soe e2)
   | Not e -> "!(" ^ (soe e) ^ ")"
   | _ -> "unknown"
 in
 ssoe expr


let string_of_expr expr =
  let rec pdepth x =
    if (x <= 0) then "" else ("    " ^ (pdepth (x - 1)))
  in
  let rec bsoe alt sign depth = function
    | And (e1, e2) -> 
	if not sign then (bsoe alt true depth (Or (Not e1, Not e2))) else
	let depth = if alt then depth else depth + 1 in
	(bsoe true sign depth e1) 
	^ "\n" ^ (pdepth depth) ^ "&&\n" 
	^ (bsoe true sign depth e2)
    | Or (Not e1, e2) -> 
	if not sign then (bsoe alt true depth (And (e1, Not e2))) else
	let depth = depth + 1 in
	(bsoe false sign depth e1) 
	^ "\n" ^ (pdepth depth) ^ "->\n"
	^ (bsoe false sign depth e2)
    | Or (e1, e2) -> 
	if not sign then (bsoe alt true depth (And (Not e1, Not e2))) else
	let depth = if alt then depth + 1 else depth in
	(bsoe false sign depth e1) 
	^ "\n" ^ (pdepth depth) ^ "||\n" 
	^ (bsoe false sign depth e2)
    | Not e -> bsoe alt (not sign) depth e
    | Forall (typ, v, e) -> 
        assert sign;
        "(forall " ^ v ^ " : " ^ (T.string_of_typ typ) ^ "\n" 
	^ (bsoe true true depth e) ^ ")"
    | BExpr e -> bsoe alt sign depth e
    | e -> (pdepth (depth + 1)) ^ (soe sign e)
  and soe sign = function
    | Bool b when sign -> if b then "true" else "false"
    | Bool b when not sign -> if b then "false" else "true"
    | Int i -> string_of_int i
    | Float f -> string_of_float f
    | Anything -> "ANYTHING"
    | CArray arr ->
        let elems = ref [] in
        A.iter (fun e -> elems := (soe sign e) :: !elems) arr;
        "[" ^ (String.concat "; " (L.rev !elems)) ^ "]"
    | CStruct flds ->
        "{" ^ (String.concat ", " 
                 (L.map (fun (fld, e) -> fld ^ " = " ^ (soe sign e)) flds)) 
        ^ "}"
    | Var s -> s
    | Array (upd, init) ->
        let rec update = function
          | (i, v) :: t -> (update t) ^ "{" ^ (soe sign i) 
	      ^ " = " ^ (soe sign v) ^ "}"
          | [] -> ""
        in
        (soe sign init) ^ (update upd)
    | Length e -> "length(" ^ (soe sign e) ^ ")"
    | ArrayAcc (e, i) -> 
	(soe sign e) ^ "[" ^ (soe sign i) ^ "]"
    | Struct (upd, init) ->
        let rec update = function
          | (fld, v) :: t -> (update t) ^ "{" ^ fld ^ " = " 
	      ^ (soe sign v) ^ "}"
          | [] -> ""
        in
        (soe sign init) ^ (update upd)
    | StructAcc (st, fld) -> (soe sign st) ^ "." ^ fld
    | FnCall (fnm, args) ->
        (if not sign then "!" else "")
	^ fnm ^ "(" 
	^ (String.concat ", " (L.map (fun e -> soe true e) args)) ^ ")"
    | Plus (e1, e2) -> assert sign; (soe sign e1) ^ " + " ^ (soe sign e2)
    | Minus (e1, e2) -> assert sign; (soe sign e1) ^ " - " ^ (soe sign e2)
    | Times (e1, e2) -> assert sign; (soe sign e1) ^ " * " ^ (soe sign e2)
    | Div (e1, e2) -> assert sign; (soe sign e1) ^ " / " ^ (soe sign e2)
    | IDiv (e1, e2) -> assert sign; (soe sign e1) ^ " div " ^ (soe sign e2)
    | Mod (e1, e2) -> assert sign; (soe sign e1) ^ " % " ^ (soe sign e2)
    | UMinus e -> assert sign; "-" ^ (soe sign e)
    | LT (e1, e2) -> 
	if not sign then soe true (GE (e1, e2)) else
	(soe sign e1) ^ " < " ^ (soe sign e2)
    | LE (e1, e2) -> 
	if not sign then soe true (GT (e1, e2)) else
	(soe sign e1) ^ " <= " ^ (soe sign e2)
    | NE (e1, e2) -> 
	if not sign then soe true (EQ (e1, e2)) else
	(soe sign e1) ^ " != " ^ (soe sign e2)
    | EQ (e1, e2) -> 
	if not sign then soe true (NE (e1, e2)) else
	(soe sign e1) ^ " = " ^ (soe sign e2)
    | GE (e1, e2) -> 
	if not sign then soe true (LT (e1, e2)) else
	(soe sign e1) ^ " >= " ^ (soe sign e2)
    | GT (e1, e2) -> 
	if not sign then soe true (LE (e1, e2)) else
	(soe sign e1) ^ " > " ^ (soe sign e2)
    | BExpr e -> soe sign e
    | And (e1, e2) -> (soe sign e1) ^ " && " ^ (soe sign e2)
    | Or (e1, e2) -> (soe sign e1) ^ " || " ^ (soe sign e2)
    | Not e -> "!(" ^ (soe sign e) ^ ")"
    | _ -> assert false
  in
  bsoe true true 0 expr

module SSet = Set.Make (
  struct
    type t = string
    let compare t1 t2 = compare t1 t2
  end)

let vars expr =
  let vs = ref SSet.empty in
  let rec vars = function
    | Var s -> vs := SSet.add s !vs
    | Array (upd, init) ->
	L.iter (fun (e1, e2) -> vars e1; vars e2) upd;
	vars init
    | Length e -> vars e
    | Struct (upd, init) ->
	L.iter (fun (_, e) -> vars e) upd;
	vars init
    | FnCall (_, args) -> L.iter vars args
    | ArrayAcc (e1, e2)
    | Plus (e1, e2)
    | Minus (e1, e2)
    | Times (e1, e2)
    | Div (e1, e2)
    | IDiv (e1, e2)
    | Mod (e1, e2)
    | LT (e1, e2)
    | LE (e1, e2)
    | NE (e1, e2)
    | EQ (e1, e2)
    | GE (e1, e2)
    | GT (e1, e2)
    | And (e1, e2)
    | Or (e1, e2) -> vars e1; vars e2
    | StructAcc (e, _)
    | UMinus e
    | BExpr e
    | Not e -> vars e
    | _ -> ()
  in
  vars expr;
  !vs

let rec eoe = function
  | E.Bool b -> Bool b
  | E.Int i -> Int i
  | E.Float f -> Float f
  | E.Uninit -> Anything
  | E.CArray arr -> CArray (A.map eoe arr)
  | E.CStruct flds -> 
      let flds = L.map (fun (fld, e) -> (fld, eoe e)) flds in
      let cs = CStruct flds in
      Struct (flds, cs)
  | E.Var s -> Var s
  | E.ArrayAcc (arr, i) -> ArrayAcc (eoe arr, eoe i)
  | E.Length e -> Length (eoe e)
  | E.StructAcc (st, fld) -> StructAcc (eoe st, fld)
  | E.FnCall (fnm, args) -> FnCall (fnm, L.map eoe args)
  | E.Collect (fnm, args) -> FnCall (fnm, L.map eoe args)
  | E.Plus (e1, e2) -> Plus (eoe e1, eoe e2)
  | E.Minus (e1, e2) -> Minus (eoe e1, eoe e2)
  | E.Times (e1, e2) -> Times (eoe e1, eoe e2)
  | E.Div (e1, e2) -> Div (eoe e1, eoe e2)
  | E.IDiv (e1, e2) -> IDiv (eoe e1, eoe e2)
  | E.Mod (e1, e2) -> Mod (eoe e1, eoe e2)
  | E.UMinus e -> UMinus (eoe e)
  | E.LT (e1, e2) -> LT (eoe e1, eoe e2)
  | E.LE (e1, e2) -> LE (eoe e1, eoe e2)
  | E.NE (e1, e2) -> NE (eoe e1, eoe e2)
  | E.EQ (e1, e2) -> EQ (eoe e1, eoe e2)
  | E.GE (e1, e2) -> GE (eoe e1, eoe e2)
  | E.GT (e1, e2) -> GT (eoe e1, eoe e2)
  | E.BExpr e -> BExpr (eoe e)
  | E.And (e1, e2) -> And (eoe e1, eoe e2)
  | E.Or (e1, e2) -> Or (eoe e1, eoe e2)
  | E.Not e -> Not (eoe e)
  | _ -> assert false

let type_of_expr tvars prgm fn expr =
  let rec ieoe = function
    | Bool b -> E.Bool b
    | Int i -> E.Int i
    | Float f -> E.Float f
    | Anything -> E.Uninit
    | CArray arr -> E.CArray (A.map ieoe arr)
    | CStruct flds -> E.CStruct (L.map (fun (fld, e) -> (fld, ieoe e)) flds)
    | Var s -> E.Var s
    | Array (upd, init) -> (* not exact! *) ieoe init
    | Length e -> E.Length (ieoe e)
    | ArrayAcc (arr, i) -> E.ArrayAcc (ieoe arr, ieoe i)
    | Struct (upd, init) -> (* !not exact! *) ieoe init
    | StructAcc (st, fld) -> E.StructAcc (ieoe st, fld)
    | Plus (e1, e2) -> E.Plus (ieoe e1, ieoe e2)
    | Minus (e1, e2) -> E.Minus (ieoe e1, ieoe e2)
    | Times (e1, e2) -> E.Times (ieoe e1, ieoe e2)
    | Div (e1, e2) -> E.Div (ieoe e1, ieoe e2)
    | IDiv (e1, e2) -> E.IDiv (ieoe e1, ieoe e2)
    | Mod (e1, e2) -> E.Mod (ieoe e1, ieoe e2)
    | UMinus e -> E.UMinus (ieoe e)
    | LT (e1, e2) -> E.LT (ieoe e1, ieoe e2)
    | LE (e1, e2) -> E.LE (ieoe e1, ieoe e2)
    | NE (e1, e2) -> E.NE (ieoe e1, ieoe e2)
    | EQ (e1, e2) -> E.EQ (ieoe e1, ieoe e2)
    | GE (e1, e2) -> E.GE (ieoe e1, ieoe e2)
    | GT (e1, e2) -> E.GT (ieoe e1, ieoe e2)
    | BExpr e -> E.BExpr (ieoe e)
    | FnCall (fnm, args) -> begin
	try
	  ignore (I.function_of_name prgm fnm);
	  E.FnCall (fnm, L.map ieoe args)
	with
	| _ -> E.Collect (fnm, L.map ieoe args)
    end
    | And (e1, e2) -> E.And (ieoe e1, ieoe e2)
    | Or (e1, e2) -> E.Or (ieoe e1, ieoe e2)
    | Not e -> E.Not (ieoe e)
    | _ -> assert false
  in
  match expr with
  | Forall _ -> T.Bool
  | x -> I.type_of_expr tvars prgm fn (ieoe x)

let is_loop cfg not_thru =
  let rec dfs first seen curr =
    if not first && curr = cfg then true else
    if curr = not_thru or I.CfgSet.mem curr !seen then false else begin
      seen := I.CfgSet.add curr !seen;
      I.CfgSet.exists (fun s -> dfs false seen s) (I.successors curr)
    end
  in
  dfs true (ref I.CfgSet.empty) cfg

let paths notify fn =
  (* generates paths starting from cfg and ending at seen cfgs or loops *)
  let paths seen cfg =
    (* next holds path ends *)
    let next = ref I.CfgSet.empty in
    let rec paths first curr =
      let is_loop curr =
        (* Only Assert statements can be loop merge-points. *)
	match I.statement curr with
        | I.Assert _ -> 
	    (not first && I.CfgSet.mem curr seen)
	      or is_loop curr cfg
        | _ -> false
      in
      let next_is_return curr =
        match I.statement curr with
        | I.Assert (_, next) -> begin
            match I.statement next with
            | I.Return _ -> true
            | _ -> false
        end
        | _ -> false
      in
      if not first && curr = cfg then begin
        (* reached cfg, no need to add cfg to next *)
        [[curr]]
      end else
      if is_loop curr then begin
        (* curr is a loop (not through cfg), so an end *)
        next := I.CfgSet.add curr !next;
        [[curr]] 
      end else
      if next_is_return curr then begin
        (* avoid ending path with Return *)
	[[curr]]
      end else begin
        (* recursive case *)
        let ps = ref [] in
        I.CfgSet.iter 
	  (fun cfg -> ps := (paths false cfg) @ !ps) 
	  (I.successors curr);
	match I.statement curr with
	| I.Assert _ when curr != cfg ->
	    (* add new VC path for intermediate assert *)
            (* but also add to others so paths are complete *)
            [curr] :: (L.map (fun p -> curr :: p) !ps);
	| I.Asgn ([(_, E.FnCall _)], _) ->
            (* add new VC for precondition of fn *)
	    [curr] :: (L.map (fun p -> curr :: p) !ps);
	| _ ->
            (* add curr to existing VC paths *)
            L.map (fun p -> curr :: p) !ps
      end
    in
    (* return paths from cfg, next starts of paths *)
    let paths = paths true cfg in
    (paths, !next)
  in
  (* remembers starting points *)
  let seen = ref I.CfgSet.empty in
  (* computes all paths starting from cfgs in starts *)
  let rec allpaths starts =
    let next = ref I.CfgSet.empty in
    seen := I.CfgSet.union starts !seen;
    I.CfgSet.iter
      (fun cfg ->
        let ps, ns = paths !seen cfg in
        (* notify of path *)
        L.iter (fun p -> notify fn p) ps;
        next := I.CfgSet.union ns !next)
      starts;
    let next = I.CfgSet.diff !next !seen in
    if not (I.CfgSet.is_empty next) then allpaths next
  in
  allpaths (I.CfgSet.singleton (I.body fn))

type mem =
  | MScope of (string * mem) list
  | MBasic of expr
  | MStruct of (string * mem) list * expr
  | MArray of (expr * mem) list * expr

type acclist =
  | AVar of string
  | AField of string
  | AIndex of expr

let vc notify prgm fn path =
  let rec acclist = function
    | Var s -> [AVar s]
    | StructAcc (st, field) -> (acclist st) @ [AField field]
    | ArrayAcc (arr, i) -> (acclist arr) @ [AIndex i]
    | _ -> assert false
  in

  let rec unacc expr = function
    | (AField fld) :: t -> unacc (StructAcc (expr, fld)) t
    | (AIndex i) :: t -> unacc (ArrayAcc (expr, i)) t
    | [] -> expr
    | _ -> assert false
  in
  let unacclist acc =
    match acc with
    | (AVar s) :: t -> unacc (Var s) t
    | _ -> assert false
  in

  let rec mem_of_expr = function
    | Struct (upd, init) ->
	MStruct (L.map (fun (fld, e) -> (fld, mem_of_expr e)) upd, init)
    | Array (upd, init) ->
	MArray (L.map (fun (i, e) -> (i, mem_of_expr e)) upd, init)
    | x -> MBasic x
  in
  let rec expr_of_mem = function
    | MBasic expr -> expr
    | MStruct (upd, init) -> 
	if upd = [] then init else
        Struct (L.map (fun (fld, u) -> (fld, expr_of_mem u)) upd, init)
    | MArray (upd, init) ->
	if upd = [] then init else
        Array (L.map (fun (i, u) -> (i, expr_of_mem u)) upd, init)
    | _ -> assert false
  in

  let replace hash k v = (k, v) :: (L.remove_assoc k hash) in

  let mem = ref (MScope []) in

  let raw_read expr modify =
    let new_mem expr = MBasic expr in
    let rec read mem accl =
      match mem, accl with
      (* convert MBasic to specific *)
      | MBasic bexpr, ((AField _) :: t as x) -> read (MStruct ([], bexpr)) x
      | MBasic bexpr, ((AIndex _) :: t as x) -> read (MArray ([], bexpr)) x

      (* base case: drilled all the way down *)
      | x, [] -> (x, expr_of_mem x)

      (* recursive cases: 
       * if modify, try reading, on failure update mem and try again;
       * else, try reading, on failure return new mem *)
      | MScope sc, ((AVar s) :: t as x) -> begin
          try 
            let mem, expr = read (L.assoc s sc) t in
            (MScope (replace sc s mem), expr)
          with
          | Not_found -> 
	      (* always modify scope *)
	      let expr = Var s in
	      read (MScope (replace sc s (new_mem expr))) x
      end
      | (MStruct (upd, init) as m), ((AField fld) :: t as x) -> begin
          try 
            let mem, expr = read (L.assoc fld upd) t in
            (MStruct (replace upd fld mem, init), expr)
          with
          | Not_found -> 
	      if modify then
		let expr = StructAcc (init, fld) in
		read (MStruct (replace upd fld (new_mem expr), init)) x
	      else
		(m, unacc init x)
      end
      | (MArray (upd, init) as m), ((AIndex i) :: t as x) -> begin
          try 
            let mem, expr = read (L.assoc i upd) t in
            (MArray (replace upd i mem, init), expr)
          with
          | Not_found -> 
	      if modify then
		let expr = ArrayAcc (expr_of_mem m, i) in
		read (MArray (replace upd i (new_mem expr), init)) x
	      else
		(m, unacc (expr_of_mem m) x)
      end
      | _ -> assert false
    in
    let mem', expr = read !mem (acclist expr) in
    mem := mem';
    expr
  in
  let read expr = raw_read expr false in
  let write lv rv =
    let rec write mem accl =
      match mem, accl with
      (* base cases: drilled all the way down *)
      | MScope sc, [AVar s] -> MScope (replace sc s (mem_of_expr rv))
      | MStruct (upd, init), [AField fld] -> 
          MStruct (replace upd fld (mem_of_expr rv), init)
      | MArray (upd, init), [AIndex i] -> 
          MArray (replace upd i (mem_of_expr rv), init)

      (* recursive cases *)
      | MScope sc, (AVar s) :: t -> 
          MScope (replace sc s (write (L.assoc s sc) t))
      | MStruct (upd, init), (AField fld) :: t ->
          MStruct (replace upd fld (write (L.assoc fld upd) t), init)
      | MArray (upd, init), (AIndex i) :: t ->
          MArray (replace upd i (write (L.assoc i upd) t), init)

      | _ -> assert false
    in
    (* read lv with modifications to initialize mem *)
    ignore (raw_read lv true);
    let mem' = write !mem (acclist lv) in
    mem := mem'
  in

  let rec apply = function
    | Var _ as x -> read x
    | Array _ -> assert false
    | Length e -> Length (base_array (apply e))
    | ArrayAcc (arr, i) -> read (ArrayAcc (arr, apply i))
    | Struct _ -> assert false
    | StructAcc _ as x -> read x
    | Plus (e1, e2) -> Plus (apply e1, apply e2)
    | Minus (e1, e2) -> Minus (apply e1, apply e2)
    | Times (e1, e2) -> Times (apply e1, apply e2)
    | Div (e1, e2) -> Div (apply e1, apply e2)
    | IDiv (e1, e2) -> IDiv (apply e1, apply e2)
    | Mod (e1, e2) -> Mod (apply e1, apply e2)
    | UMinus e -> UMinus (apply e)
    | LT (e1, e2) -> LT (apply e1, apply e2)
    | LE (e1, e2) -> LE (apply e1, apply e2)
    | NE (e1, e2) -> NE (apply e1, apply e2)
    | EQ (e1, e2) -> EQ (apply e1, apply e2)
    | GE (e1, e2) -> GE (apply e1, apply e2)
    | GT (e1, e2) -> GT (apply e1, apply e2)
    | BExpr e -> BExpr (apply e)
    | FnCall (fnm, args) -> FnCall (fnm, L.map apply args)
    | And (e1, e2) -> And (apply e1, apply e2)
    | Or (e1, e2) -> Or (apply e1, apply e2)
    | Not e -> Not (apply e)
    | Forall _ -> assert false
    | x -> x
  in
  let rec lval = function
    | Array _ -> assert false
    | ArrayAcc (arr, i) -> ArrayAcc (lval arr, apply i)
    | Struct _ -> assert false
    | StructAcc (st, fld) -> StructAcc (lval st, fld)
    | x -> x
  in

  let update lv rv = write (lval lv) (apply rv) in

  let ante = ref [] in
  let cons = ref [] in
  let require sent = ante := sent :: !ante in
  let imply sent = cons := sent :: !cons in

  let fvars = ref [] in
  let fvid = ref 0 in
  let fresh typ =
    incr fvid;
    let fvnm = "t_" ^ (string_of_int !fvid) in
    fvars := (typ, fvnm) :: !fvars;
    fvnm
  in

  let vc () = 
    let rec band = function
      | h :: [] -> BExpr h
      | h :: t -> And (BExpr h, band t)
      | [] -> BExpr (Bool true)
    in

    (* 10/11/2006 *)
    let band les =
      let es = ref ESet.empty in
      L.iter 
	(fun e ->
	  let rec x = function
	    | BExpr e -> x e
	    | And (e1, e2) -> x e1; x e2
	    | Bool b -> if not b then raise SFalse
	    | EQ (e1, e2) when e1 = e2 -> ()
	    | EQ (e1, e2) as e -> es := ESet.add e !es
	    | e -> es := ESet.add e !es
	  in
	  x e)
	les;
      band (ESet.elements !es)
    in
    let band es =
      try band es with
      | SFalse -> BExpr (Bool false)
    in
    (* 10/11/2006 END *)

    (* form expr just to get relevant vars *)
    let evars = vars (band (!ante @ !cons)) in
    (* vars with types *)
    let tvars expr =
      match !mem with
      | MScope sc -> 
	  let sv = ref [] in
	  L.iter
	    (fun (v, _) ->
	      if SSet.mem v evars
		  && not (L.exists (fun (_, v') -> v = v') !fvars) then
		sv := (I.type_of_var prgm fn v, v) :: !sv)
            sc;
          !sv @ (L.filter (fun (_, v) -> SSet.mem v evars) !fvars)
      | _ -> assert false
    in
    (* assertions about prgm's defines *)
    let defines =
      let def nm = function
	| CArray arr -> 
	    A.to_list (A.mapi (fun i v -> EQ (ArrayAcc(Var nm, Int i), v)) arr)
	| CStruct flds ->
	    L.map (fun (fld, v) -> EQ (StructAcc (Var nm, fld), v)) flds
	| Anything -> []
	| x -> [EQ (Var nm, x)]
      in
      L.concat
	(L.map 
	   (fun (_, nm, expr) -> def nm (eoe expr))
	   (L.filter (fun (_, nm, _) -> SSet.mem nm evars) (I.defines prgm)))
    in
    (* drum-roll, please *)
    let e = Or (Not (band (defines @ !ante)), band !cons) in
    L.fold_left (fun e (typ, v) -> Forall (typ, v, e)) e (tvars e)
  in

  let spath = ref "" in
  let add_spath x = spath := !spath ^ x in
  let soe x = E.string_of_expr x in
  let rec walk first = function
    | [] -> vc ()
    | cfg :: tpath -> 
	begin
	  (* require any invariants *)
	  let inv = I.invariants cfg in
	  if inv != E.const_true then begin
	    require (apply (eoe inv))
	  end;
          match I.statement cfg with
          | I.Skip _ -> ()
          | I.Asgn (asgns, _) ->
              let asgn = function
		| (lv, E.FnCall (fnm, args)) ->
                    let fnc = I.function_of_name prgm fnm in
                    (* params -> args *)
                    let vmap = H.create 13 in
                    L.iter2
                      (fun (_, p) arg -> H.add vmap p arg)
                      (I.parameters fnc) args;
		    if tpath = [] then begin
                      (* last statement in path: assert precondition *)
		      let prec = E.sub vmap (I.precondition fnc) in
		      add_spath ("@ " ^ (soe prec) ^ ";\n");
		      imply (apply (eoe prec))
		    end else begin
		      add_spath 
			((soe lv) ^ " = " 
			 ^ fnm ^ "(" 
			 ^ (String.concat ", " (L.map soe args))
			 ^ ");\n");
		      (* map rv -> fresh var *)
                      let rtyp, rv = I.return_value fnc in
                      let tvar = fresh rtyp in
                      H.add vmap rv (E.Var tvar);
                      (* postcondition becomes transition relation *)
                      let rltn = E.sub vmap (I.postcondition fnc) in
                      (* assert that rltn holds, constraining tvar *)
                      require (apply (eoe rltn));
                      (* update lv with fresh variable *)
                      update (eoe lv) (Var tvar)
		    end
		| (lv, E.ArrayAlloc (typ, sz)) ->
		    add_spath
		      ((soe lv) ^ " = new "
		       ^ (T.string_of_typ typ) 
		       ^ "[" ^ (soe sz) ^ "];\n");
		    let lv = eoe lv in
		    require (EQ (Length (base_array lv), eoe sz))
		| (lv, rv) -> 
		    add_spath ((soe lv) ^ " = " ^ (soe rv) ^ ";\n");
		    update (eoe lv) (eoe rv)
              in
              L.iter asgn asgns
          | I.Branch (grd, tb, fb) -> begin
              (* determine whether grd or not grd *)
              let grd = match tpath with
              | h :: t -> 
		  if tb = h then begin
		    add_spath ("assume " ^ (soe grd) ^ ";\n");
		    grd
		  end else begin
		    add_spath ("assume " ^ (soe (E.bnot grd)) ^ ";\n");
		    E.bnot grd
		  end
              | _ -> assert false
              in
              require (apply (eoe grd))
          end
          | I.Assert (asrt, _) when first -> 
	      add_spath ("@ " ^ (soe asrt) ^ ";\n");
	      require (apply (eoe asrt))
	  | I.Assert _ when tpath != [] ->
              (* intermediate assert, covered by another VC *)
	      ()
	  | I.Assert (asrt, _) -> 
	      add_spath ("@ " ^ (soe asrt) ^ ";\n");
	      (* for sanity check *)
	      let expr = 
		if !sanity then E.band asrt (I.invariants cfg) else asrt
	      in
	      imply (apply (eoe expr))
          | I.Return _ -> assert false
	end;
        walk false tpath
  in

  let verbose = true in
  let nm =
    let rec vcname = function
      | h :: t -> 
	  if verbose then print_string ((string_of_int (I.id h)) ^ " ");
	  (string_of_int (I.linenum h)) ^ " " ^ (vcname t)
      | [] -> 
	  if verbose then print_endline "";
	  ""
    in
    vcname path
  in
  notify nm !spath prgm fn (walk true path)

let vcs prgm notify =
  I.FnSet.iter
    (fun fn -> paths (vc notify prgm) fn)
    (I.functions prgm)
